/*
 *  furniture_optimizer.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 10/25/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#include <stdio.h>

#include "delfem/quaternion.h"
#include "delfem/indexed_array.h"
#include "delfem/matvec/matdia_blkcrs.h"
#include "delfem/matvec/matdiafrac_blkcrs.h"
#include "delfem/matvec/vector_blk.h"
#include "delfem/matvec/bcflag_blk.h"
//#include "delfem/rigid/linearsystem_rigid.h"

//#include "furniture.h"
#include "rigid_pointset.h"
#include "furniture_utility.h"
#include "joint_connectivity.h"
#include "rigid_contact_ls.h"
#include "furniture_optimizer.h"
#include "gl_drawer_furniture.h"


const Com::CVector3D aFaceNormal[6] = {
  Com::CVector3D(+1, 0, 0),
  Com::CVector3D(-1, 0, 0),
  Com::CVector3D(+0,+1, 0),
  Com::CVector3D(-0,-1, 0),
  Com::CVector3D(+0, 0,+1),
  Com::CVector3D(-0, 0,-1) };

// t is a tmporary buffer size of 9
static inline void CalcInvMat3(double a[], double t[] )
{
	const double det = a[0]*a[4]*a[8] + a[3]*a[7]*a[2] + a[6]*a[1]*a[5]
  - a[0]*a[7]*a[5] - a[6]*a[4]*a[2] - a[3]*a[1]*a[8];
	const double inv_det = 1.0/det;
  
  for(int i=0;i<9;i++){ t[i] = a[i]; }
  
	a[0] = inv_det*(t[4]*t[8]-t[5]*t[7]);
	a[1] = inv_det*(t[2]*t[7]-t[1]*t[8]);
	a[2] = inv_det*(t[1]*t[5]-t[2]*t[4]);
  
	a[3] = inv_det*(t[5]*t[6]-t[3]*t[8]);
	a[4] = inv_det*(t[0]*t[8]-t[2]*t[6]);
	a[5] = inv_det*(t[2]*t[3]-t[0]*t[5]);
  
	a[6] = inv_det*(t[3]*t[7]-t[4]*t[6]);
	a[7] = inv_det*(t[1]*t[6]-t[0]*t[7]);
	a[8] = inv_det*(t[0]*t[4]-t[1]*t[3]);
}

static void CalcInvMat(double* a, const unsigned int n, int& info )
{
	double tmp1;
  
	info = 0;
	unsigned int i,j,k;
	for(i=0;i<n;i++){
		if( fabs(a[i*n+i]) < 1.0e-30 ){
			info = 1;
			return;
		}
		if( a[i*n+i] < 0.0 ){
			info--;
		}
		tmp1 = 1.0 / a[i*n+i];
		a[i*n+i] = 1.0;
		for(k=0;k<n;k++){
			a[i*n+k] *= tmp1;
		}
		for(j=0;j<n;j++){
			if( j!=i ){
				tmp1 = a[j*n+i];
				a[j*n+i] = 0.0;
				for(k=0;k<n;k++){
					a[j*n+k] -= tmp1*a[i*n+k];
				}
			}
		}
	}
}

void SetPattern_JointConnectivity
(MatVec::CMatDia_BlkCrs& m_mat,     
 MatVec::CVector_Blk& m_residual,
 const unsigned int nRB,
 const std::vector<CJoint_Fix>& aFix )
{
  //  unsigned int nRB = aRB.size();
  unsigned int nConst = aFix.size();
  const unsigned int nblk = nRB+nConst;  
  {
    Com::CIndexedArray crs;
    crs.InitializeSize(nblk);
    crs.index.resize(nblk+1);
    crs.index[0] = 0;
    for(unsigned int i=0;i<nRB;i++){ crs.index[i+1]=0; }
    for(unsigned int i=nRB;i<nRB+nConst;i++){ crs.index[i+1]=0; }
    for(unsigned int icst=0;icst<nConst;icst++){
      const unsigned int irb0 = aFix[icst].irb0;     
      const unsigned int irb1 = aFix[icst].irb1;
      //      crs.index[irb0+1] += 1;        
      //      crs.index[irb1+1] += 1;
      crs.index[irb0+1] += 2;        
      crs.index[irb1+1] += 2;      
      crs.index[icst+nRB+1] += 2;
    }
    for(unsigned int i=0;i<nblk;i++){ 
      crs.index[i+1] = crs.index[i+1] + crs.index[i];
    }
    const unsigned int ncrs = crs.index[nblk];
    crs.array.resize(ncrs);
    for(unsigned int icst=0;icst<nConst;icst++){
      unsigned int irb0 = aFix[icst].irb0;
      unsigned int irb1 = aFix[icst].irb1;
      {
        crs.array[ crs.index[irb0] ] = irb1;
        crs.index[irb0]++;        
        crs.array[ crs.index[irb0] ] = icst+nRB;
        crs.index[irb0]++;
      }
      {
        crs.array[ crs.index[irb1] ] = irb0;
        crs.index[irb1]++;
        crs.array[ crs.index[irb1] ] = icst+nRB;
        crs.index[irb1]++;
      }
      {
        crs.array[ crs.index[icst+nRB] ] = irb0;
        crs.index[icst+nRB]++;
        crs.array[ crs.index[icst+nRB] ] = irb1;
        crs.index[icst+nRB]++;
      }
    }
    for(int i=nblk;i>0;i--){ 
      crs.index[i] = crs.index[i-1];
    }
    crs.index[0] = 0;
    crs.Sort();
    assert( crs.index[nRB+nConst] == ncrs );
    m_mat.AddPattern(crs);      
  }    
}


void AddLynSysJointInsideForce // (Q1ex)*(b0+QBa0-b1-QBa1) 
(MatVec::CMatDia_BlkCrs& mat,     
 MatVec::CVector_Blk& res,
 unsigned int irb0, unsigned int irb1,
 const Com::CVector3D& d0, const Com::CVector3D& d1,
 const Com::CMatrix3& Q0,  const Com::CMatrix3& Q1,
 const Com::CVector3D& b0, const Com::CVector3D& b1,
 const Com::CVector3D& a0, const Com::CVector3D& a1,
 const Com::CVector3D& e, 
 double offset, 
 const Com::CVector3D& doffsetda0,
 const Com::CVector3D& doffsetda1, 
 double stiff)
{
  const Com::CMatrix3 B0(b0.x,b0.y,b0.z);  
  const Com::CMatrix3 B1(b1.x,b1.y,b1.z);    
  const Com::CMatrix3& QB0 = Q0.MatMat(B0);  
  const Com::CMatrix3& QB1 = Q1.MatMat(B1);    
  const Com::CVector3D& QBa0 = QB0.MatVec(a0);
  const Com::CVector3D& QBa1 = QB1.MatVec(a1);  
  const Com::CVector3D& Q1e = Q1.MatVec(e);
  const Com::CVector3D v = d0 + QBa0 - d1 - QBa1;  
  double co = Dot(Q1e,v);
  if( fabs(co)-offset < 0 ) return;
  const double flgof = ( co < 0 ) ? -1 : 1;
  double p = co-flgof*offset;
  //  double eng = stiff*0.5*p*p;
  //  std::cout << "active joint : " << eng << std::endl;
  const Com::CVector3D& dpdd0 = Q1e;
  const Com::CVector3D& dpds0 = -Com::CMatrix3(Q1e).MatVec(QBa0);
  const Com::CVector3D& dpda0 = +(QB0).MatVecTrans(Q1e)-flgof*doffsetda0;
  
  const Com::CVector3D& dpdd1 = -Q1e;
  const Com::CVector3D& dpds1 = Com::CMatrix3(Q1e).MatVec(QBa1) + Com::CMatrix3(Q1e).MatVec(v);
  const Com::CVector3D& dpda1 = -(QB1).MatVecTrans(Q1e)-flgof*doffsetda1;
  
  AddVector(res, irb0,0, p*dpdd0,-stiff);
  AddVector(res, irb0,3, p*dpds0,-stiff);
  AddVector(res, irb0,6, p*dpda0,-stiff);  
  
  AddVector(res, irb1,0, p*dpdd1,-stiff);
  AddVector(res, irb1,3, p*dpds1,-stiff);
  AddVector(res, irb1,6, p*dpda1,-stiff);    
  
  AddMatrix(mat, irb0,0, irb0,0, Com::CMatrix3(dpdd0,dpdd0), stiff,true);
  AddMatrix(mat, irb0,0, irb0,3, Com::CMatrix3(dpdd0,dpds0), stiff,true);  
  AddMatrix(mat, irb0,0, irb0,6, Com::CMatrix3(dpdd0,dpda0), stiff,true);  
  AddMatrix(mat, irb0,3, irb0,0, Com::CMatrix3(dpds0,dpdd0), stiff,true);
  AddMatrix(mat, irb0,3, irb0,3, Com::CMatrix3(dpds0,dpds0), stiff,true);  
  AddMatrix(mat, irb0,3, irb0,6, Com::CMatrix3(dpds0,dpda0), stiff,true);   
  AddMatrix(mat, irb0,6, irb0,0, Com::CMatrix3(dpda0,dpdd0), stiff,true);
  AddMatrix(mat, irb0,6, irb0,3, Com::CMatrix3(dpda0,dpds0), stiff,true);
  AddMatrix(mat, irb0,6, irb0,6, Com::CMatrix3(dpda0,dpda0), stiff,true);
  
  AddMatrix(mat, irb0,0, irb1,0, Com::CMatrix3(dpdd0,dpdd1), stiff,true);
  AddMatrix(mat, irb0,0, irb1,3, Com::CMatrix3(dpdd0,dpds1), stiff,true);  
  AddMatrix(mat, irb0,0, irb1,6, Com::CMatrix3(dpdd0,dpda1), stiff,true);     
  AddMatrix(mat, irb0,3, irb1,0, Com::CMatrix3(dpds0,dpdd1), stiff,true);
  AddMatrix(mat, irb0,3, irb1,3, Com::CMatrix3(dpds0,dpds1), stiff,true);  
  AddMatrix(mat, irb0,3, irb1,6, Com::CMatrix3(dpds0,dpda1), stiff,true);     
  AddMatrix(mat, irb0,6, irb1,0, Com::CMatrix3(dpda0,dpdd1), stiff,true);
  AddMatrix(mat, irb0,6, irb1,3, Com::CMatrix3(dpda0,dpds1), stiff,true);  
  AddMatrix(mat, irb0,6, irb1,6, Com::CMatrix3(dpda0,dpda1), stiff,true);   
  
  AddMatrix(mat, irb1,0, irb0,0, Com::CMatrix3(dpdd1,dpdd0), stiff,true);
  AddMatrix(mat, irb1,0, irb0,3, Com::CMatrix3(dpdd1,dpds0), stiff,true);  
  AddMatrix(mat, irb1,0, irb0,6, Com::CMatrix3(dpdd1,dpda0), stiff,true);     
  AddMatrix(mat, irb1,3, irb0,0, Com::CMatrix3(dpds1,dpdd0), stiff,true);
  AddMatrix(mat, irb1,3, irb0,3, Com::CMatrix3(dpds1,dpds0), stiff,true);  
  AddMatrix(mat, irb1,3, irb0,6, Com::CMatrix3(dpds1,dpda0), stiff,true);     
  AddMatrix(mat, irb1,6, irb0,0, Com::CMatrix3(dpda1,dpdd0), stiff,true);
  AddMatrix(mat, irb1,6, irb0,3, Com::CMatrix3(dpda1,dpds0), stiff,true);  
  AddMatrix(mat, irb1,6, irb0,6, Com::CMatrix3(dpda1,dpda0), stiff,true);  
  
  AddMatrix(mat, irb1,0, irb1,0, Com::CMatrix3(dpdd1,dpdd1), stiff,true);
  AddMatrix(mat, irb1,0, irb1,3, Com::CMatrix3(dpdd1,dpds1), stiff,true);  
  AddMatrix(mat, irb1,0, irb1,6, Com::CMatrix3(dpdd1,dpda1), stiff,true);   
  AddMatrix(mat, irb1,3, irb1,0, Com::CMatrix3(dpds1,dpdd1), stiff,true);
  AddMatrix(mat, irb1,3, irb1,3, Com::CMatrix3(dpds1,dpds1), stiff,true);  
  AddMatrix(mat, irb1,3, irb1,6, Com::CMatrix3(dpds1,dpda1), stiff,true);     
  AddMatrix(mat, irb1,6, irb1,0, Com::CMatrix3(dpda1,dpdd1), stiff,true);
  AddMatrix(mat, irb1,6, irb1,3, Com::CMatrix3(dpda1,dpds1), stiff,true);
  AddMatrix(mat, irb1,6, irb1,6, Com::CMatrix3(dpda1,dpda1), stiff,true);     
}

void AddLinSys_JointInside
(MatVec::CMatDia_BlkCrs& mat,     
 MatVec::CVector_Blk& res,
 unsigned int icst, const CJoint_Fix& cst,
 // const std::vector<CRigidBody3D>& aRB,
 const std::vector<CRigid3D_Ini>& aRI,
 unsigned int idir0, unsigned int idir1,
 double stiff)
{
  unsigned int irb0 = cst.irb0;
  unsigned int irb1 = cst.irb1;
  //  const CRigidBody3D& rb0 = aRB[irb0];  
  //  const CRigidBody3D& rb1 = aRB[irb1];
  const CRigid3D_Ini& ri0 = aRI[irb0];      
  const CRigid3D_Ini& ri1 = aRI[irb1];        
  const Com::CMatrix3& Q0 = ri0.GetRotMatrix_Ini();
  const Com::CMatrix3& Q1 = ri1.GetRotMatrix_Ini();
  const Com::CVector3D& d0 = ri0.GetIniPosCG();
  const Com::CVector3D& d1 = ri1.GetIniPosCG();
  const Com::CVector3D a0(ri0.xlen,ri0.ylen,ri0.zlen);
  const Com::CVector3D a1(ri1.xlen,ri1.ylen,ri1.zlen);  
  const Com::CVector3D& b0 = aFaceNormal[idir0]*0.5;
  const Com::CVector3D& b1 = aFaceNormal[idir1]*0.5;
  
  if(      (idir0==4||idir0==5) && (idir1!=4&&idir1!=5) ){
    AddLynSysJointInsideForce(mat,res,
                              irb1,irb0,
                              d1,d0,
                              Q1,Q0,
                              b1,b0,
                              a1,a0,
                              Com::CVector3D(1,0,0), 
                              ri0.xlen*0.5-ri1.zlen,
                              Com::CVector3D(0,0,-1),    
                              Com::CVector3D(0.5,0,0),                                              
                              stiff);
  }
  else if( (idir0!=4&&idir0!=5) && (idir1==4||idir1==5) ){
    AddLynSysJointInsideForce(mat,res,
                              irb0,irb1,
                              d0,d1,
                              Q0,Q1,
                              b0,b1,
                              a0,a1,
                              Com::CVector3D(1,0,0), 
                              ri1.xlen*0.5-ri1.zlen,
                              Com::CVector3D(0,0,-1),                                                           
                              Com::CVector3D(0.5,0,0),
                              stiff);
  }  
}


void AddLinSys_JointConnectivity
(MatVec::CMatDia_BlkCrs& m_mat,     
 MatVec::CVector_Blk& m_residual,
 unsigned int icst, const CJoint_Fix& cst,
 const std::vector<CRigid3D_Ini>& aRI,
 unsigned int idir0, unsigned int idir1)
{
  unsigned int nRB = aRI.size();
  unsigned int irb0 = cst.irb0;
  unsigned int irb1 = cst.irb1;
  const CRigid3D_Ini& ri0 = aRI[irb0]; 
  const CRigid3D_Ini& ri1 = aRI[irb1];       
  const Com::CMatrix3& Q0 = ri0.GetRotMatrix_Ini();
  const Com::CMatrix3& Q1 = ri1.GetRotMatrix_Ini();
  const Com::CVector3D& d0 = ri0.GetIniPosCG();
  const Com::CVector3D& d1 = ri1.GetIniPosCG();
  const Com::CVector3D a0(ri0.xlen,ri0.ylen,ri0.zlen);
  const Com::CVector3D a1(ri1.xlen,ri1.ylen,ri1.zlen);  
  const Com::CVector3D& b0 = aFaceNormal[idir0]*0.5;
  const Com::CVector3D& b1 = aFaceNormal[idir1]*0.5;
  const Com::CMatrix3 B0(b0.x,b0.y,b0.z);
  const Com::CMatrix3 B1(b1.x,b1.y,b1.z);  
  const Com::CMatrix3& QB0 = Q0.MatMat(B0);
  const Com::CMatrix3& QB1 = Q1.MatMat(B1);  
  const Com::CVector3D& QBa0 = QB0.MatVec(a0);
  const Com::CVector3D& QBa1 = QB1.MatVec(a1);  
  
  Com::CVector3D Qb;
  if(      (idir0==4||idir0==5) && (idir1!=4&&idir1!=5) ){
    Qb = Q0.MatVec(b0);
  }
  else if( (idir0!=4&&idir0!=5) && (idir1==4||idir1==5) ){
    Qb = Q1.MatVec(b1);
  }
  else{ assert(0); }
  
  
  {
    Add_Vector(m_mat, irb0,    0,  nRB+icst,0, Qb,     +1,true);
    Add_Vector(m_mat, nRB+icst,0,  irb0,    0, Qb,     +1,false);  
    Add_Vector(m_mat, irb1,    0,  nRB+icst,0, Qb,     -1,true);
    Add_Vector(m_mat, nRB+icst,0,  irb1,    0, Qb,     -1,false);      
  }
  
  {
    const Com::CVector3D& wQbQBa0 = Com::CMatrix3(Qb).MatVec(QBa0);
    const Com::CVector3D& wQbQBa1 = Com::CMatrix3(Qb).MatVec(QBa1);    
    Add_Vector(m_mat, irb0,    3,  nRB+icst,0, wQbQBa0,-1,true);
    Add_Vector(m_mat, nRB+icst,0,  irb0,    3, wQbQBa0,-1,false);  
    Add_Vector(m_mat, irb1,    3,  nRB+icst,0, wQbQBa1,+1,true);
    Add_Vector(m_mat, nRB+icst,0,  irb1,    3, wQbQBa1,+1,false);  
  }
  
  {
    const Com::CVector3D& QBtQb0 = QB0.MatVecTrans(Qb);
    const Com::CVector3D& QBtQb1 = QB1.MatVecTrans(Qb);    
    Add_Vector(m_mat, irb0,    6,  nRB+icst,0, QBtQb0,+1,true);
    Add_Vector(m_mat, nRB+icst,0,  irb0,    6, QBtQb0,+1,false);  
    Add_Vector(m_mat, irb1,    6,  nRB+icst,0, QBtQb1,-1,true);
    Add_Vector(m_mat, nRB+icst,0,  irb1,    6, QBtQb1,-1,false);      
  }
  
  const Com::CVector3D v = d0 + QBa0 - d1 - QBa1;
  if(      (idir0==4||idir0==5) && (idir1!=4&&idir1!=5) ){ // Qb = Q0B0
    const Com::CVector3D& wQb0v = Com::CMatrix3(Qb).MatVec(v);
    Add_Vector(m_mat, irb0,    3,  nRB+icst,0, wQb0v,+1,true);
    Add_Vector(m_mat, nRB+icst,0,  irb0,    3, wQb0v,+1,false);  
  }
  else if( (idir0!=4&&idir0!=5) && (idir1==4||idir1==5) ){ // QB = Q1B1
    const Com::CVector3D& wQb1v = Com::CMatrix3(Qb).MatVec(v);
    Add_Vector(m_mat, irb1,    3,  nRB+icst,0, wQb1v,+1,true);
    Add_Vector(m_mat, nRB+icst,0,  irb1,    3, wQb1v,+1,false);      
  }  
  
  const double phi = Com::Dot(Qb,v);
  //  std::cout << phi << "    " << d0 + QBa0 << "   " << d1 + QBa1 << std::endl;
  m_residual.AddValue(nRB+icst, 0, -phi);    
}

void OptimizeEvalFunc_Grad
(CJointedRigidBody& jrb,
 /////
 PARAM_TYPE itype_s, unsigned int irb_s, unsigned int idim_s,  
 double penalty_coeff_joint, double penalty_coeff_cont,
 double alpha,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc)
{   
  std::vector<CRigid3D_Ini>& aRI = jrb.aRI;
  std::vector<CJoint_Fix>& aFix = jrb.aFix;
  std::vector< std::pair<unsigned int, unsigned int> >& aFixDir = jrb.aFixDir;
  
  std::vector<CRigid3D_Def>& aRD = jrb.aRD;
  std::vector<CJoint_Fix_Force>& aJFixForce = jrb.aJFixForce;
  std::vector<CExternalForceOnRigidBody>& aExForce = jrb.aExForce;
  std::vector<CContactPoints_RigidBody3D>& aFP = jrb.aFP;
  
  Com::CVector3D& gravity = jrb.gravity;
  CContactProp& contProp = jrb.contProp;
  
//  double penalty_coeff_joint = jrb.penalty_coeff_joint;
//  double penalty_coeff_cont = jrb.penalty_coeff_cont;
  
  CSimulationSensitivity simSens;
  //  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensPR;
  //  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensLM;
  GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem(simSens,//aSensPR,aSensLM, 
                                                                    itype_s,irb_s,idim_s, itype_anchor_reloc,
                                                                    aRI,aRD,aFix,aJFixForce,aExForce,
                                                                    gravity,
                                                                    //                                                                    cont_base,cont_norm,stiff_n,stiff_f,
                                                                    contProp,
                                                                    aFP);
  
  //  double eng0 = GetEvaluationFunction(penalty_coeff_joint,penalty_coeff_cont);  
  double gradient = GetDerivativeEvaluationFunction(itype_s,irb_s,idim_s, 
                                                    simSens,
                                                    penalty_coeff_joint, penalty_coeff_cont,
                                                    aRI,aRD,aFix,aJFixForce,aFixDir,
                                                    contProp.cont_base,contProp.cont_norm,
                                                    //                                                    contProp,
                                                    aFP);
  double eps = -gradient*alpha;
  MoveInitialConfig_JointedRigidBodyContactSystem(itype_s,irb_s,idim_s, eps, aRI,aFix);
  RelocateAnchors_JointedRigidBodyContactSystem(aFP,
                                                itype_s,irb_s,idim_s, 
                                                itype_anchor_reloc,
                                                simSens.aPlankPR[irb_s], eps, 
                                                aRI,aRD,contProp.cont_norm); 
  MoveCurrentDeform(simSens,eps,aRD,aJFixForce);    
  //  double eng1 = GetEvaluationFunction(penalty_coeff_joint,penalty_coeff_cont);      
  /*
   std::cout << itype_s << " " << irb_s << " " << idim_s;
   std::cout << "   --> " << eng0 << " " << eng1;
   std::cout << "   --> " << gradient;
   std::cout << "   --> " << eng1-eng0 << " should be equal " << eps*gradient << std::endl;    
   */
}



void ItrOptimizeEvalFunc_Grad
(CJointedRigidBody& jrb, double alpha, 
 const std::vector<COptimizationDof>& aOptDof,                              
 double penalty_coeff_joint, double penalty_coeff_cont)
{
  std::vector<CRigid3D_Ini>& aRI = jrb.aRI;
  std::vector<CJoint_Fix>& aFix = jrb.aFix;
  std::vector< std::pair<unsigned int, unsigned int> >& aFixDir = jrb.aFixDir;
  
  std::vector<CRigid3D_Def>& aRD = jrb.aRD;
  std::vector<CJoint_Fix_Force>& aJFixForce = jrb.aJFixForce;
  std::vector<CExternalForceOnRigidBody>& aExForce = jrb.aExForce;
  std::vector<CContactPoints_RigidBody3D>& aFP = jrb.aFP;
  
  Com::CVector3D& gravity = jrb.gravity;
  CContactProp& contProp = jrb.contProp;
  
//  double penalty_coeff_joint = jrb.penalty_coeff_joint;
//  double penalty_coeff_cont = jrb.penalty_coeff_cont;
  
  std::cout << "######################" << std::endl;      
  std::cout << "ItrOptimizeEvalFunc_Grad" << std::endl;  
  FindInitialConfiguration_SameFrictionAnchor(jrb);  
  const unsigned int nOptDof = aOptDof.size();
  std::vector<CEvalGradSensPRLM> aEvalDevalSensPRLM;
  aEvalDevalSensPRLM.resize(nOptDof);
  const ANCHOR_RELOCATION_TYPE itype_anchor_reloc = MOVE_INITIAL;
  for(unsigned int itr=0;itr<10;itr++){ // this is needed optimization converge at some level until reevaluation of anchor
    for(unsigned int iop=0;iop<nOptDof;iop++){
      const COptimizationDof& od = aOptDof[iop];
      CEvalGradSensPRLM& egs = aEvalDevalSensPRLM[iop];
      GetGradEvaluationFunctionAndSens(egs.eval,egs.deval,egs.simSens,
                                       penalty_coeff_joint, penalty_coeff_cont,
                                       od.itype,od.iobj,od.idim, itype_anchor_reloc,
                                       aRI,aRD,aFix,aJFixForce,aFixDir,aExForce,
                                       gravity,
                                       contProp,
                                       aFP);      
    }
    for(unsigned int iop=0;iop<nOptDof;iop++){
      CEvalGradSensPRLM& egs = aEvalDevalSensPRLM[iop];      
      const COptimizationDof& od = aOptDof[iop];      
      double eps = -egs.deval*alpha;
      MoveInitialConfig_JointedRigidBodyContactSystem(od.itype,od.iobj,od.idim, eps, aRI,aFix);
      MoveCurrentDeform(egs.simSens,eps, aRD,aJFixForce);  
      RelocateAnchors_JointedRigidBodyContactSystem(aFP,
                                                    od.itype,od.iobj,od.idim, 
                                                    itype_anchor_reloc,
                                                    egs.simSens.aPlankPR[od.iobj], eps, 
                                                    aRI,aRD,contProp.cont_norm);       
    }
    for(int i=0;i<1000;i++){
      double resn, updn;
      StepTime_BackwardEular_UntactAnchor(jrb,0.01,resn,updn);	// solve rigid motion    
      if( updn < 1.0e-10 ){ break; }
    }  
    // projection is nesesally because optimization uses penalty method and doesn't satisfy constraint
    FindInitialConfiguration_SameFrictionAnchor(jrb);   
  }    
}        

void OptimizeEvalFunc_GradMultDirPrj
(CJointedRigidBody& jrb,double alpha,
 const std::vector<COptimizationDof>& aOptDof,                                     
 double penalty_coeff_joint, double penalty_coeff_cont)
{
  std::vector<CRigid3D_Ini>& aRI = jrb.aRI;
  std::vector<CJoint_Fix>& aFix = jrb.aFix;
  std::vector< std::pair<unsigned int, unsigned int> >& aFixDir = jrb.aFixDir;
  
  std::vector<CRigid3D_Def>& aRD = jrb.aRD;
  std::vector<CJoint_Fix_Force>& aJFixForce = jrb.aJFixForce;
  std::vector<CExternalForceOnRigidBody>& aExForce = jrb.aExForce;
  std::vector<CContactPoints_RigidBody3D>& aFP = jrb.aFP;
  
  Com::CVector3D& gravity = jrb.gravity;
  CContactProp& contProp = jrb.contProp;
  
//  double penalty_coeff_joint = jrb.penalty_coeff_joint;
//  double penalty_coeff_cont = jrb.penalty_coeff_cont;
  
//  double dt = 0.005;
  const unsigned int nOptDof = aOptDof.size();
  const ANCHOR_RELOCATION_TYPE itype_anchor_reloc = MOVE_INITIAL;
  std::vector<CEvalGradSensPRLM> aEvalDevalSensPRLM1;
  {
    aEvalDevalSensPRLM1.resize(nOptDof);      
    std::vector<CExternalForceOnRigidBody> aMyExForce1;     
    aMyExForce1.push_back( CExternalForceOnRigidBody(0,Com::CVector3D(0,0,0), Com::CVector3D(-100.0,-300,0) ) );
    SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                           gravity,
                           contProp,
                           aFP);
    /*
    for(unsigned int i=0;i<1000;i++){
      double resn, updn;
      StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(dt,
                                                                        aRI,aRD,aFix,aJFixForce,aMyExForce1,
                                                                        gravity,
                                                                        //                                                                        cont_base,cont_norm,stiff_n,stiff_f,
                                                                        contProp,
                                                                        aFP,
                                                                        resn, updn);
      if( updn < 1.0e-10 ) break;
    } 
     */
    for(unsigned int iop=0;iop<nOptDof;iop++){
      const COptimizationDof& od = aOptDof[iop];
      CEvalGradSensPRLM& egs = aEvalDevalSensPRLM1[iop];
      GetGradEvaluationFunctionAndSens(egs.eval,egs.deval,egs.simSens,
                                       penalty_coeff_joint, penalty_coeff_cont,
                                       od.itype,od.iobj,od.idim, itype_anchor_reloc,
                                       aRI,aRD,aFix,aJFixForce,aFixDir,aMyExForce1,
                                       gravity,
                                       //                                       cont_base,cont_norm,stiff_n,stiff_f,
                                       contProp,
                                       aFP);      
    }      
  }
  ////
  std::vector<CEvalGradSensPRLM> aEvalDevalSensPRLM2;
  {
    aEvalDevalSensPRLM2.resize(nOptDof);      
    std::vector<CExternalForceOnRigidBody> aMyExForce2;     
    aMyExForce2.push_back( CExternalForceOnRigidBody(0,Com::CVector3D(0,0,0), Com::CVector3D(+100.0,-300,0) ) );
    SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                           gravity,
                           contProp,
                           aFP);
    /*
    for(unsigned int i=0;i<1000;i++){
      double resn, updn;
      StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(dt,
                                                                        aRI,aRD,aFix,aJFixForce,aMyExForce2,
                                                                        gravity,
                                                                        //                                                                        cont_base,cont_norm,stiff_n,stiff_f,
                                                                        contProp,
                                                                        aFP,
                                                                        resn,updn);
      if( updn < 1.0e-10 ) break;            
    } 
     */     
    for(unsigned int iop=0;iop<nOptDof;iop++){
      const COptimizationDof& od = aOptDof[iop];
      CEvalGradSensPRLM& egs = aEvalDevalSensPRLM2[iop];
      GetGradEvaluationFunctionAndSens(egs.eval,egs.deval,egs.simSens,
                                       penalty_coeff_joint, penalty_coeff_cont,
                                       od.itype,od.iobj,od.idim, itype_anchor_reloc,
                                       aRI,aRD,aFix,aJFixForce,aFixDir,aMyExForce2,
                                       gravity,
                                       //                                       contProp.cont_base,contProp.cont_norm,stiff_n,stiff_f,
                                       contProp,
                                       aFP);      
    }      
  }
  /////
  //    std::cout << "%%% " << itr << std::endl;
  
  std::vector<double> upd_in;
  upd_in.resize(aOptDof.size(),0);
  for(unsigned int iop=0;iop<aOptDof.size();iop++){
    if( aOptDof[iop].itype == JNT_POS ) continue;    
    const CEvalGradSensPRLM& egs1 = aEvalDevalSensPRLM1[iop];
    const CEvalGradSensPRLM& egs2 = aEvalDevalSensPRLM2[iop];    
    upd_in[iop] = -(egs1.deval+egs2.deval)*0.5*alpha;
  }  
  
  {
    std::vector<Com::CVector3D> aGradJnt;
    aGradJnt.resize(aFix.size());
    for(unsigned int iop=0;iop<aOptDof.size();iop++){
      const COptimizationDof& od = aOptDof[iop];   
      if( od.itype == JNT_POS ){          
        const CEvalGradSensPRLM& egs1 = aEvalDevalSensPRLM1[iop];
        const CEvalGradSensPRLM& egs2 = aEvalDevalSensPRLM2[iop];        
        aGradJnt[od.iobj][od.idim] = -(egs1.deval+egs2.deval)*0.5*alpha;
      }
    }      
    JointDegeneration(aRI,aFix,aFixDir, aOptDof,aGradJnt,upd_in);
  }
  
  std::vector<double> upd_out;
  Projection_JointConnectivity(aRI,aRD,aFix,aFP,aFixDir,aOptDof, 
                               upd_in,
                               upd_out,
                               contProp.cont_base,contProp.cont_norm, 1000);  
  
  std::vector<CEvalGradSensPRLM> aEvalDevalSensPRLM0;
  {
    aEvalDevalSensPRLM0.resize(nOptDof);      
    std::vector<CExternalForceOnRigidBody> aMyExForce0 = aExForce;
    //    aMyExForce0.clear();
    SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                           gravity,
                           contProp,
                           aFP);
    /*
    for(unsigned int i=0;i<1000;i++){
      double resn, updn;      
      StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(dt,
                                                                        aRI,aRD,aFix,aJFixForce,aMyExForce0,
                                                                        gravity,
                                                                        //                                                                        cont_base,cont_norm,stiff_n,stiff_f,
                                                                        contProp,
                                                                        aFP,
                                                                        resn,updn);
      if( updn < 1.0e-10 ){ break; }
    }      
     */
    for(unsigned int iop=0;iop<nOptDof;iop++){      
      const COptimizationDof& od = aOptDof[iop];
      CEvalGradSensPRLM& egs = aEvalDevalSensPRLM0[iop];
      GetGradEvaluationFunctionAndSens(egs.eval,egs.deval,egs.simSens,
                                       penalty_coeff_joint, penalty_coeff_cont,
                                       od.itype,od.iobj,od.idim, itype_anchor_reloc,
                                       aRI,aRD,aFix,aJFixForce,aFixDir,aMyExForce0,
                                       gravity,
                                       //                                       cont_base,cont_norm,stiff_n,stiff_f,
                                       contProp,
                                       aFP);      
    }      
  }  
  EditInitialConfig_JointedRigidBodyContactSystem_SlaveJoint(aRI,aRD,aFix,aJFixForce,aFP,aFixDir,
                                                             aOptDof,aEvalDevalSensPRLM0,
                                                             upd_out,
                                                             contProp.cont_norm);
  /*
   for(unsigned int iop=0;iop<nOptDof;iop++){
   CEvalGradSensPRLM& egs1 = aEvalDevalSensPRLM1[iop];      
   CEvalGradSensPRLM& egs2 = aEvalDevalSensPRLM2[iop];            
   const COptimizationDof& od = aOptDof[iop];      
   //      std::cout << od.itype << " " << od.iobj << " " << od.idim << "  " << egs1.deval << " " << egs2.deval << std::endl;            
   double eps = -(egs1.deval+egs2.deval)*0.5*alpha;
   MoveInitialConfig_JointedRigidBodyContactSystem(od.itype,od.iobj,od.idim, eps, aRB,aFix);
   
   MoveCurrentConfig_JointedRigidBodyContactSystem(egs1.aSensPR,egs1.aSensLM,eps, aRB,aFix);  
   MoveCurrentConfig_JointedRigidBodyContactSystem(egs2.aSensPR,egs2.aSensLM,eps, aRB,aFix);        
   
   RelocateAnchors_JointedRigidBodyContactSystem(aFP,
   od.itype,od.iobj,od.idim, 
   itype_anchor_reloc,
   egs1.aSensPR[od.iobj], eps, 
   aRB,cont_norm);       
   RelocateAnchors_JointedRigidBodyContactSystem(aFP,
   od.itype,od.iobj,od.idim, 
   itype_anchor_reloc,
   egs2.aSensPR[od.iobj], eps, 
   aRB,cont_norm);             
   }
   for(int i=0;i<100;i++){
   bool res = this->StepTime_BackwardEular_UntactAnchor(0.1);	// solve rigid motion    
   if( res ){ break; }
   }        
   */
}

void ItrOptimizeEvalFunc_GradMultDir
(CJointedRigidBody& jrb, double alpha,
 const std::vector<COptimizationDof>& aOptDof,
 double penalty_coeff_joint, double penalty_coeff_cont)
{  
  
  std::vector<CRigid3D_Ini>& aRI = jrb.aRI;
  std::vector<CJoint_Fix>& aFix = jrb.aFix;
  std::vector< std::pair<unsigned int, unsigned int> >& aFixDir = jrb.aFixDir;
  
  std::vector<CRigid3D_Def>& aRD = jrb.aRD;
  std::vector<CJoint_Fix_Force>& aJFixForce = jrb.aJFixForce;
  std::vector<CExternalForceOnRigidBody>& aExForce = jrb.aExForce;
  std::vector<CContactPoints_RigidBody3D>& aFP = jrb.aFP;
  
  Com::CVector3D& gravity = jrb.gravity;
  CContactProp& contProp = jrb.contProp;
  
//  double penalty_coeff_joint = jrb.penalty_coeff_joint;
//  double penalty_coeff_cont = jrb.penalty_coeff_cont;
  
  std::cout << "######################" << std::endl;      
  std::cout << "ItrOptimizeEvalFunc_Grad" << std::endl;  
  FindInitialConfiguration_SameFrictionAnchor(jrb);  
  const unsigned int nOptDof = aOptDof.size();
  const ANCHOR_RELOCATION_TYPE itype_anchor_reloc = MOVE_INITIAL;
  for(unsigned int itr=0;itr<10;itr++){ // this is needed optimization converge at some level until reevaluation of anchor
    std::vector<CEvalGradSensPRLM> aEvalDevalSensPRLM1;
    {
      aEvalDevalSensPRLM1.resize(nOptDof);      
      std::vector<CExternalForceOnRigidBody> aMyExForce1;     
      aMyExForce1.push_back( CExternalForceOnRigidBody(0,Com::CVector3D(0,0,0), Com::CVector3D(-3.0,-6,0) ) );
      SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                             gravity,
                             contProp,
                             aFP);
      /*
      for(unsigned int i=0;i<1000;i++){
        double resn, updn;
        StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(0.01,
                                                                          aRI,aRD,aFix,aJFixForce,aMyExForce1,
                                                                          gravity,
                                                                          //                                                                          cont_base,cont_norm,stiff_n,stiff_f,
                                                                          contProp,
                                                                          aFP,
                                                                          resn,updn);
        if( updn < 1.0-10 ){ break; }
      } 
       */
      for(unsigned int iop=0;iop<nOptDof;iop++){
        const COptimizationDof& od = aOptDof[iop];
        CEvalGradSensPRLM& egs = aEvalDevalSensPRLM1[iop];
        GetGradEvaluationFunctionAndSens(egs.eval,egs.deval,egs.simSens,
                                         penalty_coeff_joint, penalty_coeff_cont,
                                         od.itype,od.iobj,od.idim, itype_anchor_reloc,
                                         aRI,aRD,aFix,aJFixForce,aFixDir,aMyExForce1,
                                         gravity,
                                         //                                         cont_base,cont_norm,stiff_n,stiff_f,
                                         contProp,
                                         aFP);      
      }      
    }
    ////
    std::vector<CEvalGradSensPRLM> aEvalDevalSensPRLM2;
    {
      aEvalDevalSensPRLM2.resize(nOptDof);      
      std::vector<CExternalForceOnRigidBody> aMyExForce2;     
      aMyExForce2.push_back( CExternalForceOnRigidBody(0,Com::CVector3D(0,0,0), Com::CVector3D(+3.0,-6,0) ) );
      SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                             gravity,
                             contProp,
                             aFP);
      /*
      for(unsigned int i=0;i<1000;i++){
        double resn, updn;        
        StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(0.01,
                                                                          aRI,aRD,aFix,aJFixForce,aMyExForce2,
                                                                          gravity,
                                                                          //                                                                          cont_base,cont_norm,stiff_n,stiff_f,
                                                                          contProp,
                                                                          aFP,
                                                                          resn,updn); 
        if( updn < 1.0e-10 ){ break; }
      }      
       */
      for(unsigned int iop=0;iop<nOptDof;iop++){
        const COptimizationDof& od = aOptDof[iop];
        CEvalGradSensPRLM& egs = aEvalDevalSensPRLM2[iop];
        GetGradEvaluationFunctionAndSens(egs.eval,egs.deval,egs.simSens,
                                         penalty_coeff_joint, penalty_coeff_cont,
                                         od.itype,od.iobj,od.idim, itype_anchor_reloc,
                                         aRI,aRD,aFix,aJFixForce,aFixDir,aMyExForce2,
                                         gravity,
                                         //                                         cont_base,cont_norm,stiff_n,stiff_f,
                                         contProp,
                                         aFP);      
      }      
    }
    /////
    //    std::cout << "%%% " << itr << std::endl;
    for(unsigned int iop=0;iop<nOptDof;iop++){
      CEvalGradSensPRLM& egs1 = aEvalDevalSensPRLM1[iop];      
      CEvalGradSensPRLM& egs2 = aEvalDevalSensPRLM2[iop];            
      const COptimizationDof& od = aOptDof[iop];      
      //      std::cout << od.itype << " " << od.iobj << " " << od.idim << "  " << egs1.deval << " " << egs2.deval << std::endl;            
      double eps = -(egs1.deval+egs2.deval)*0.5*alpha;
      MoveInitialConfig_JointedRigidBodyContactSystem(od.itype,od.iobj,od.idim, eps, aRI,aFix);
      
      MoveCurrentDeform(egs1.simSens,eps, aRD,aJFixForce);  
      MoveCurrentDeform(egs2.simSens,eps, aRD,aJFixForce);        
      
      RelocateAnchors_JointedRigidBodyContactSystem(aFP,
                                                    od.itype,od.iobj,od.idim, 
                                                    itype_anchor_reloc,
                                                    egs1.simSens.aPlankPR[od.iobj], eps, 
                                                    aRI,aRD,contProp.cont_norm);       
      RelocateAnchors_JointedRigidBodyContactSystem(aFP,
                                                    od.itype,od.iobj,od.idim, 
                                                    itype_anchor_reloc,
                                                    egs2.simSens.aPlankPR[od.iobj], eps, 
                                                    aRI,aRD,contProp.cont_norm);             
    }
    for(int i=0;i<1000;i++){
      double resn, updn;
      StepTime_BackwardEular_UntactAnchor(jrb,0.01,resn,updn);	// solve rigid motion    
      if( updn < 1.0e-10 ){ break; }
    }
    // projection is nesesally because optimization uses penalty method and doesn't satisfy constraint    
    FindInitialConfiguration_SameFrictionAnchor(jrb);       
  }        
}

void PrintJointBendForceSensitivty(CJointedRigidBody& jrb,
                                   const std::vector<COptimizationDof>& aOptDof)
{
  std::vector<CRigid3D_Ini>& aRI = jrb.aRI;
  std::vector<CJoint_Fix>& aFix = jrb.aFix;
  std::vector< std::pair<unsigned int, unsigned int> >& aFixDir = jrb.aFixDir;
  
  std::vector<CRigid3D_Def>& aRD = jrb.aRD;
  std::vector<CJoint_Fix_Force>& aJFixForce = jrb.aJFixForce;
  std::vector<CExternalForceOnRigidBody>& aExForce = jrb.aExForce;
  std::vector<CContactPoints_RigidBody3D>& aFP = jrb.aFP;
  
  Com::CVector3D& gravity = jrb.gravity;
  CContactProp& contProp = jrb.contProp;
  
  std::cout << "%%%%%%%%%%%%%%%%%%%%%%%" << std::endl;
  std::cout << "JointBendingForceSensitivity" << std::endl;
  const unsigned int nOptDof = aOptDof.size();
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);
  /*
  for(unsigned int i=0;i<1000;i++){
    double resn, updn;    
    StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(0.01,
                                                                      aRI,aRD,aFix,aJFixForce,aExForce,
                                                                      gravity,
                                                                      //                                                                      cont_base,cont_norm,stiff_n,stiff_f,
                                                                      contProp,
                                                                      aFP,
                                                                      updn,resn);
    if( updn < 1.0e-10 ){ break; }
  } 
   */
  std::vector<Com::CVector3D> aQh;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;
    const CRigid3D_Ini& ri0 = aRI[aFix[ifix].irb0];
    const CRigid3D_Ini& ri1 = aRI[aFix[ifix].irb1];    
    if(      (idir0==4||idir0==5) && (idir1!=4&&idir1!=5) ){    
      const Com::CMatrix3& Q1 = ri1.GetRotMatrix_Ini();
      const Com::CVector3D& Qh1 = Q1.MatVec( Cross( Com::CVector3D(0,0,1), aFaceNormal[idir1] ) );
      aQh.push_back(Qh1);
    }
    else if( (idir0!=4&&idir0!=5) && (idir1==4||idir1==5) ){
      const Com::CMatrix3& Q0 = ri0.GetRotMatrix_Ini();
      const Com::CVector3D& Qh0 = Q0.MatVec( Cross( Com::CVector3D(0,0,1), aFaceNormal[idir0] ) );          
      aQh.push_back(Qh0);
    }
  }
  std::vector<double> A;
  A.resize( nOptDof*aFix.size() );
  assert( aQh.size() == aFix.size() );
  for(unsigned int iop=0;iop<nOptDof;iop++){
    const COptimizationDof& od = aOptDof[iop];
    CSimulationSensitivity simSens;
    GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem(simSens, 
                                                                      od.itype,od.iobj,od.idim, MOVE_INITIAL,
                                                                      aRI,aRD,aFix,aJFixForce,aExForce,
                                                                      gravity,
                                                                      contProp,
                                                                      aFP);  
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      A[ifix*(nOptDof)+iop] = Com::Dot(simSens.aJointLM[ifix].second,aQh[ifix]);      
    }
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    for(unsigned int iop=0;iop<nOptDof;iop++){
      std::cout << ifix << " " << iop << "  " << A[ifix*(nOptDof)+iop] << std::endl;
    }
  }
  FILE *fp = fopen("sens_bend.m","w");
  fprintf(fp,"A=[ ");
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    for(unsigned int iop=0;iop<nOptDof;iop++){
      fprintf(fp,"%lf ",A[ifix*(nOptDof)+iop]);
    }
    fprintf(fp,"; ");
  }
  fprintf(fp,"];");
  fclose(fp);
}

void OptimizeEvalFunc_GradPrj(CJointedRigidBody& jrb, double alpha,
                              const std::vector<COptimizationDof>& aOptDof)
{
  std::vector<CRigid3D_Ini>& aRI = jrb.aRI;
  std::vector<CJoint_Fix>& aFix = jrb.aFix;
  const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir = jrb.aFixDir;
  
  std::vector<CRigid3D_Def>& aRD = jrb.aRD;
  std::vector<CJoint_Fix_Force>& aJFixForce = jrb.aJFixForce;
  const std::vector<CExternalForceOnRigidBody>& aExForce = jrb.aExForce;
  std::vector<CContactPoints_RigidBody3D>& aFP = jrb.aFP;
  
  const Com::CVector3D& gravity = jrb.gravity;
  const CContactProp& contProp = jrb.contProp;
  
//  std::cout << "%%%%%%%%%%%%%%%%%%%%%%%" << std::endl;
//  std::cout << "OptimzeEvalFunc_GradPrj" << std::endl;
  const unsigned int nOptDof = aOptDof.size();
  SolveStaticEquilibrium(aRI,aRD,aFix,aJFixForce,aExForce,
                         gravity,
                         contProp,
                         aFP);
  /*
  for(unsigned int i=0;i<1000;i++){
    double resn, updn;    
    StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(0.01,
                                                                      aRI,aRD,aFix,aJFixForce,aExForce,
                                                                      gravity,
                                                                      //                                                                      cont_base,cont_norm,stiff_n,stiff_f,
                                                                      contProp,
                                                                      aFP,
                                                                      updn,resn);
    if( updn < 1.0e-10 ){ break; }
  } 
   */
  std::vector<CEvalGradSensPRLM> aEvalDevalSensPRLM;
  aEvalDevalSensPRLM.resize(nOptDof);
  const ANCHOR_RELOCATION_TYPE itype_anchor_reloc = MOVE_INITIAL;
  for(unsigned int iop=0;iop<nOptDof;iop++){
    const COptimizationDof& od = aOptDof[iop];
    CEvalGradSensPRLM& egs = aEvalDevalSensPRLM[iop];
    GetGradEvaluationFunctionAndSens(egs.eval,egs.deval,egs.simSens,
                                     0, 0,
                                     od.itype,od.iobj,od.idim, itype_anchor_reloc,
                                     aRI,aRD,aFix,aJFixForce,aFixDir,aExForce,
                                     gravity,
                                     contProp,
                                     aFP);      
  }
  std::vector<double> upd_in;
  upd_in.resize(aOptDof.size(),0);
  for(unsigned int iop=0;iop<aOptDof.size();iop++){
    if( aOptDof[iop].itype == JNT_POS ) continue;    
    const CEvalGradSensPRLM& egs = aEvalDevalSensPRLM[iop];
    upd_in[iop] = -egs.deval*alpha;
  }  
  
  {
    std::vector<Com::CVector3D> aGradJnt;
    aGradJnt.resize(aFix.size());
    for(unsigned int iop=0;iop<aOptDof.size();iop++){
      const COptimizationDof& od = aOptDof[iop];   
      if( od.itype == JNT_POS ){
        const CEvalGradSensPRLM& egs = aEvalDevalSensPRLM[iop];
        aGradJnt[od.iobj][od.idim] = -egs.deval*alpha;
      }
    }      
    JointDegeneration(aRI,aFix,aFixDir, aOptDof,aGradJnt,upd_in);
  }
  
  std::vector<double> upd_out;
  Projection_JointConnectivity(aRI,aRD,aFix,aFP,aFixDir,aOptDof, 
                               upd_in,
                               upd_out,
                               contProp.cont_base,contProp.cont_norm, 1000);  
  
  EditInitialConfig_JointedRigidBodyContactSystem_SlaveJoint(aRI,aRD,aFix,aJFixForce,aFP,aFixDir,
                                                             aOptDof,aEvalDevalSensPRLM,
                                                             upd_out,
                                                             contProp.cont_norm);
}




void FindInitialConfiguration_SameFrictionAnchor(CJointedRigidBody& jrb)
{  
  std::vector<CRigid3D_Ini>& aRI = jrb.aRI;  
  std::vector<CRigid3D_Def>& aRD = jrb.aRD;
  std::vector<CContactPoints_RigidBody3D>& aFP = jrb.aFP;
  CContactProp& contProp = jrb.contProp;
  
  unsigned int nRB = aRI.size();
  assert( aRD.size() == nRB );
  { // set initial value of ICP
    double ave_crv[3] = {0,0,0};
    for(unsigned int irb=0;irb<nRB;irb++){        
      //      const CRigidBody3D& rb = aRB[irb];
      const Com::CMatrix3& P = aRD[irb].GetRotMatrix_Inc();
      double crv[3]; P.GetCRV_RotMatrix(crv);
      ave_crv[0] += crv[0];
      ave_crv[1] += crv[1];      
      ave_crv[2] += crv[2];      
    }    
    ave_crv[0] /= nRB;
    ave_crv[1] /= nRB;
    ave_crv[2] /= nRB;
    Com::CMatrix3 avP; avP.SetRotMatrix_CRV(ave_crv);
    Com::CVector3D ave_cg_ini = GetGravityCenter(false,jrb.aRI,jrb.aRD);
    Com::CVector3D ave_cg_inc = GetGravityCenter(true,jrb.aRI,jrb.aRD)-ave_cg_ini;
    jrb.MoveInitialConfig_UnchangeCurrent(ave_cg_ini,ave_cg_inc,avP);
  }
  
  for(unsigned int irb=0;irb<nRB;irb++){        
    CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      if( aCP.aFP[i].itype_contact != 1 ){
        aCP.aFP[i].is_force_cont = false;
      }
      else{
        aCP.aFP[i].is_force_cont = true;    
      }      
    }
  }
  
  for(unsigned int itr=0;itr<3;itr++){
    double eng_n = 0, eng_f = 0;
    Com::CVector3D dq_du, dq_dr;
    Com::CMatrix3 ddq_dudu, ddq_dudr, ddq_drdu, ddq_drdr;
    Com::CVector3D ave_cg_ini = GetGravityCenter(false,jrb.aRI,jrb.aRD);
    for(unsigned int irb=0;irb<nRB;irb++){        
      const CRigid3D_Ini& ri = aRI[irb];          
      const CContactPoints_RigidBody3D& aCP = aFP[irb];
      for(unsigned int i=0;i<aCP.Size();i++){
        Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);                     
        if( !aCP.aFP[i].is_force_cont ) continue;
        const Com::CVector3D& anc = aCP.aFP[i].aloc; 
        const Com::CVector3D& v1 = ri.GetInitialPosition(v0);
        const Com::CVector3D c = ave_cg_ini;
        const Com::CVector3D v = v1-c;
        const Com::CVector3D a = anc;
        const Com::CVector3D n = contProp.cont_norm;
        const double kf = contProp.stiff_f;
        const double fe = GetStaticFrictionEnergy(c,v,a,n,kf);
        Com::CVector3D dfe_du, dfe_dr;
        Get1stDer_StaticFrictionEnergy(dfe_du,dfe_dr, c,v,a,n,kf);
        Com::CMatrix3 dfe_dudu, dfe_dudr, dfe_drdu, dfe_drdr;
        Get2ndDer_StaticFrictionEnergy(dfe_dudu,dfe_dudr,dfe_drdu,dfe_drdr, c,v,a,n,kf);    
        /////
        const Com::CVector3D b = contProp.cont_base;
        const double kn = contProp.stiff_n;
        const double ce = GetContactEnergy(c,v,b,n,kn);
        Com::CVector3D dce_du, dce_dr;
        Get1stDer_ContactEnergy(dce_du,dce_dr, c,v,b,n,kn);
        Com::CMatrix3 dce_dudu, dce_dudr, dce_drdr;
        Get2ndDer_ContactEnergy(dce_dudu, dce_dudr, dce_drdr, c,v,b,n,kn);      
        eng_n += ce;
        eng_f += fe;
        dq_du += dce_du + dfe_du;
        dq_dr += dce_dr + dfe_dr;
        ddq_dudu += dce_dudu + dfe_dudu;
        ddq_dudr += dce_dudr + dfe_dudr;
        ddq_drdu += dce_dudr.Trans() + dfe_drdu;        
        ddq_drdr += dce_drdr + dfe_drdr;
      }
    }      
    //    std::cout << eng_n << " " << eng_f << std::endl;  
    {
      double A[36];
      double b[6];
      for(unsigned int i=0;i<3;i++){      
        b[0+i] = -dq_du[i];
        b[3+i] = -dq_dr[i];      
        for(unsigned int j=0;j<3;j++){     
          A[6*(i+0)+(j+0)] = ddq_dudu.mat[3*i+j];
          A[6*(i+0)+(j+3)] = ddq_dudr.mat[3*i+j];        
          A[6*(i+3)+(j+0)] = ddq_drdu.mat[3*i+j];
          A[6*(i+3)+(j+3)] = ddq_drdr.mat[3*i+j];
        }
      }
      double nres = 0;
      for(unsigned int i=0;i<6;i++){
        nres += b[i]*b[i];        
      }
      int info;
      CalcInvMat(A, 6, info );
      double x[6];
      double nupd = 0;
      for(unsigned int i=0;i<6;i++){
        x[i] = 0;
        for(unsigned int j=0;j<6;j++){
          x[i] += A[6*i+j]*b[j];
        }
        nupd += x[i]*x[i];
      }
      Com::CVector3D U(x[0],x[1],x[2]);
      Com::CMatrix3 P; P.SetRotMatrix_Cartesian( Com::CVector3D(x[3],x[4],x[5]) );
      jrb.MoveInitialConfig_UnchangeCurrent(ave_cg_ini,U,P);      
    }    
  }
  
  { // anchor relocation
    double eng = 0; 
    for(unsigned int irb=0;irb<nRB;irb++){        
      const CRigid3D_Ini& ri = aRI[irb];            
      CContactPoints_RigidBody3D& aCP = aFP[irb];      
      for(unsigned int i=0;i<aCP.Size();i++){
        Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);                      
        if( !aCP.aFP[i].is_force_cont ) continue;      
        const Com::CVector3D p0 = ri.GetInitialPosition(v0);        
        eng += SquareDistance(p0,aCP.aFP[i].aloc);
        aCP.aFP[i].aloc = p0;
      }
    }    
  }
  
  {
    for(int i=0;i<100;i++){
      double resn, updn;
      StepTime_BackwardEular_UntactAnchor(jrb,0.1,resn,updn);	// solve rigid motion    
      if( updn < 1.0e-10 ) break;
    }
  }
  
  {
    const Com::CVector3D& cont_norm = contProp.cont_norm;    
    const Com::CVector3D& cont_base = contProp.cont_base;        
    const double stiff_f = contProp.stiff_f;
    const double stiff_n = contProp.stiff_n;    
    for(unsigned int irb=0;irb<nRB;irb++){        
      const CRigid3D_Ini& ri = aRI[irb];          
      const CRigid3D_Def& rd = aRD[irb];                
      const CContactPoints_RigidBody3D& aCP = aFP[irb];
      for(unsigned int i=0;i<aCP.Size();i++){
        Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);                      
        if( !aCP.aFP[i].is_force_cont ) continue;
        Com::CVector3D v1 = GetCurrentPosition(v0,ri,rd);        
        Com::CVector3D v2 = aCP.aFP[i].aloc;
        Com::CVector3D dir_f = v2-v1;
        dir_f -= cont_norm*Com::Dot(cont_norm,dir_f);
        Com::CVector3D dir_n = cont_norm*Com::Dot(cont_norm,cont_base-v1);
        double f = Length(stiff_f*dir_f);
        double n = Length(stiff_n*dir_n);
        if( n*jrb.myu_s-f < 0 ){
          std::cout << "detect slippery : " << irb << " " << i << " " << n*jrb.myu_s-f << std::endl;
        }        
      }
    }
  }
}

void GetGradEvaluationFunctionAndSens
(double& eval, double& deval,
 CSimulationSensitivity& simSens,
 ////
 double penalty_coeff_joint, double penalty_coeff_cont,
 ////
 PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s, 
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc,
 ////
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,                                        
 std::vector<CJoint_Fix>& aFix, 
 std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir,
 const std::vector<CExternalForceOnRigidBody>& aMyExForce,     
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP)  
{
  /*
   for(unsigned int i=0;i<100;i++){
   bool res = StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem(0.1,
   aRB,aFix,aMyExForce,
   gravity,
   cont_base,cont_norm,stiff_n,stiff_f,
   aFP);                                                                                        
   if( res ){ break; }
   }
   */
  GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem(simSens, 
                                                                    itype_s,iobj_s,idim_s, itype_anchor_reloc,
                                                                    aRI,aRD,aFix,aJFixForce,aMyExForce,
                                                                    gravity,
                                                                    //                                                                    cont_base,cont_norm,stiff_n,stiff_f,
                                                                    contProp,
                                                                    aFP);  
  /*
   eval = GetEvaluationFunction(penalty_coeff_joint,penalty_coeff_cont,
   aRB,aFix,aFixDir,
   cont_base,cont_norm,
   aFP);       
   */
  deval = GetDerivativeEvaluationFunction(itype_s,iobj_s,idim_s, 
                                          simSens,
                                          penalty_coeff_joint, penalty_coeff_cont,
                                          aRI,aRD,aFix,aJFixForce,aFixDir,
                                          contProp.cont_base,contProp.cont_norm,
                                          //                                          contProp,
                                          aFP);  
}


double GetEvaluationFunction
(double coeff_joint, double coeff_cont, 
 const std::vector<CRigidBody3D>& aRB, 
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CJoint_Fix_Force>& aJFixForce,
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir,
 ////
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 const std::vector<CContactPoints_RigidBody3D>& aFP)
{
  
  /*
   double eng_cont = 0;   
   {
   const double corner[8][3] = { 
   {-1,-1,-1}, {-1,-1,+1}, {-1,+1,-1}, {-1,+1,+1},
   {+1,-1,-1}, {+1,-1,+1}, {+1,+1,-1}, {+1,+1,+1} };      
   for(unsigned int irb=0;irb<aRB.size();irb++){
   const CRigidBody3D& rb = aRB[irb];
   for(unsigned int i=0;i<8;i++){
   Com::CVector3D v0 = Com::CVector3D(corner[i][0]*rb.xlen, corner[i][1]*rb.ylen, corner[i][2]*rb.zlen )*0.5;
   Com::CVector3D v1 = rb.GetCurrentPosition(v0);
   double hn = -Com::Dot(v1-cont_base,cont_norm);      
   if( hn < 0 ) continue;
   ////
   eng_cont += 0.5*stiff_n*hn*hn;      
   if( aFP[irb*8+i].itype_contact == 1 ){
   Com::CVector3D ap_t = v1-aFP[irb*8+i].aloc;      
   ap_t -= cont_norm*Dot(cont_norm,ap_t);    
   eng_cont += 0.5*stiff_f*Com::Dot(ap_t,ap_t);
   }
   }
   }    
   }
   */
  double eng_cont = 0;   
  { // evaluation function due to contact
    for(unsigned int irb=0;irb<aRB.size();irb++){
      const CRigidBody3D& rb = aRB[irb];
      const CRigid3D_Ini& ri = rb.ri;
      const CContactPoints_RigidBody3D& aCP = aFP[irb];
      for(unsigned int i=0;i<aCP.Size();i++){
        if( !aCP.aFP[i].is_force_cont ) continue;
        //        Com::CVector3D v0 = Com::CVector3D(corner[i][0]*rb.xlen, corner[i][1]*rb.ylen, corner[i][2]*rb.zlen )*0.5;
        Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);                      
        Com::CVector3D v1 = rb.ri.GetInitialPosition(v0);
        double hn = -Com::Dot(v1-cont_base,cont_norm);      
        eng_cont += 0.5*hn*hn;
      }
    }    
  }  
  double eng_joint;
  { 
    std::vector<CRigid3D_Ini> aRI;
    for(unsigned int irb=0;irb<aRB.size();irb++){ aRI.push_back(aRB[irb].ri); }      
    eng_joint = JointConnectivityEnergy(aRI,aFix,aFixDir);
  }
  double eng_bend = 0;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cjs = aFix[ifix];        
    const CJoint_Fix_Force& jff = aJFixForce[ifix];    
    Com::CVector3D lr(jff.lambda[3],jff.lambda[4],jff.lambda[5]);
    eng_bend += 0.5*(1.0/cjs.invstiff_rot)*Dot(lr,lr);
  }  
  return eng_bend + eng_joint*coeff_joint + eng_cont*coeff_cont;
}



double GetDerivativeEvaluationFunction
(PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s,
 const CSimulationSensitivity& simSens,
 double coeff_joint, double coeff_cont,
 /////
 std::vector<CRigid3D_Ini>& aRI,
 std::vector<CRigid3D_Def>& aRD,                        
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir, 
 ////
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, 
 const std::vector<CContactPoints_RigidBody3D>& aFP)
{
  //  assert( irb_s < aRB.size() );
  const std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aSensPR = simSens.aPlankPR;
  const std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aSensLM = simSens.aJointLM;
  assert( idim_s < 3 );
  assert( aSensPR.size() == aRI.size() );
  assert( aSensLM.size() == aFix.size() );
  assert( aFP.size() == aRI.size() );
  /*
   double deng_cont = 0;     
   {
   const double corner[8][3] = { 
   {-1,-1,-1}, {-1,-1,+1}, {-1,+1,-1}, {-1,+1,+1},
   {+1,-1,-1}, {+1,-1,+1}, {+1,+1,-1}, {+1,+1,+1} };      
   for(unsigned int irb=0;irb<aRB.size();irb++){
   const CRigidBody3D& rb = aRB[irb];
   for(unsigned int i=0;i<8;i++){
   const Com::CVector3D v0 = Com::CVector3D(corner[i][0]*rb.xlen, corner[i][1]*rb.ylen, corner[i][2]*rb.zlen )*0.5;
   const Com::CVector3D v1 = rb.GetCurrentPosition(v0);
   double hn = -Com::Dot(v1-cont_base,cont_norm);
   if( hn < 0 ) continue;
   const Com::CVector3D dv1 = GetSensitivityPointOnRigidBody(itype_s,iobj_s,idim_s, irb,rb, i, aSensPR[irb] );
   deng_cont += -stiff_n*hn*Com::Dot(dv1,cont_norm);
   if( aFP[irb*8+i].itype_contact == 1 ){
   Com::CVector3D ap_t = v1-aFP[irb*8+i].aloc;
   ap_t -= cont_norm*Dot(cont_norm,ap_t);
   Com::CVector3D dapt = dv1 - cont_norm*Dot(cont_norm,dv1);
   deng_cont += stiff_f*Com::Dot(ap_t,dapt);
   }
   }
   }    
   }
   */
  // evaluatino function due to contact
  double deng_cont = 0;
  if( itype_s == BOX_POS || itype_s == BOX_ROT || itype_s == BOX_LEN ){
    //    const CRigidBody3D& rb = aRB[iobj_s];
    const CRigid3D_Ini& ri = aRI[iobj_s];
    const CContactPoints_RigidBody3D& aCP = aFP[iobj_s];
    for(unsigned int i=0;i<aCP.Size();i++){        
      if( !aCP.aFP[i].is_force_cont ) continue;
      Com::CMatrix3 sU,sQ,sL;
      GetSensitivityInitialPointOnRigidBody(ri,i,aCP,sU,sQ,sL);
      Com::CVector3D e(0,0,0); e[idim_s]=1;
      Com::CVector3D scp(0,0,0);
      if(      itype_s == BOX_POS ){ scp = sU.MatVec(e); }
      else if( itype_s == BOX_ROT ){ scp = sQ.MatVec(e); }
      else if( itype_s == BOX_LEN ){ scp = sL.MatVec(e); }
      const Com::CVector3D& v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);
      Com::CVector3D v1 = ri.GetInitialPosition(v0);
      double hn = Com::Dot(v1-cont_base,cont_norm);      
      deng_cont += hn*Dot(cont_norm,scp);
    }    
  }
  
  double deng_joint = 0;
  { // energy due to joint connectivity
    std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensJE_BoxPosRot;
    std::vector<Com::CVector3D> aSensJE_BoxLen;  
    std::vector<Com::CVector3D> aSensJE_JntPos;  
    GetSensitivityJCE_InitialConfigBox(aSensJE_BoxPosRot,aSensJE_BoxLen,aSensJE_JntPos,
                                       aRI,aFix,aFixDir);          
    if(      itype_s == BOX_POS ){ deng_joint = aSensJE_BoxPosRot[iobj_s].first[idim_s]; }
    else if( itype_s == BOX_ROT ){ deng_joint = aSensJE_BoxPosRot[iobj_s].second[idim_s]; }
    else if( itype_s == BOX_LEN ){ deng_joint = aSensJE_BoxLen[iobj_s][idim_s]; }
    else if( itype_s == JNT_POS ){ deng_joint = aSensJE_JntPos[iobj_s][idim_s]; }    
  }
  double deng_bend = 0;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix_Force& jff = aJFixForce[ifix];   
    const CJoint_Fix& cjs = aFix[ifix];       
    Com::CVector3D lr(jff.lambda[3],jff.lambda[4],jff.lambda[5]);
    deng_bend += (1.0/cjs.invstiff_rot)*Dot(lr,aSensLM[ifix].second);
  }  
  return deng_bend + deng_joint*coeff_joint + deng_cont*coeff_cont;
}

void EditInitialConfig_JointedRigidBodyContactSystem_SlaveJoint
(std::vector<CRigid3D_Ini>& aRI,
 std::vector<CRigid3D_Def>& aRD, 
 std::vector<CJoint_Fix>& aFix,
 std::vector<CJoint_Fix_Force>& aJFixForce, 
 std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir, 
 ////
 const std::vector<COptimizationDof>& aOptDof,
 const std::vector<CEvalGradSensPRLM>& aEvalDevalSensPRLM,
 const std::vector<double>& upd,  // dim of nOptDof
 const Com::CVector3D& cont_norm)
{
  const ANCHOR_RELOCATION_TYPE itype_anchor_reloc = MOVE_INITIAL;
  for(unsigned int iop=0;iop<aOptDof.size();iop++){
    const CEvalGradSensPRLM& egs = aEvalDevalSensPRLM[iop];      
    const COptimizationDof& od = aOptDof[iop];      
    if( od.itype == JNT_POS ) continue;    
    double eps = upd[iop];
    MoveInitialConfig_JointedRigidBodyContactSystem(od.itype,od.iobj,od.idim, eps, aRI,aFix);
    MoveCurrentDeform(egs.simSens,eps, aRD,aJFixForce);  
    RelocateAnchors_JointedRigidBodyContactSystem(aFP,
                                                  od.itype,od.iobj,od.idim, 
                                                  itype_anchor_reloc,
                                                  egs.simSens.aPlankPR[od.iobj], eps, 
                                                  aRI,aRD,cont_norm);       
  }
  /*
   //  assert( upd.size() == aRB.size()*9 );
   for(unsigned int irb=0;irb<aRB.size();irb++){
   CRigidBody3D& rb = aRB[irb];    
   ////
   rb.SetIniPosCG( rb.ri.GetIniPosCG() + Com::CVector3D(upd[irb*9+0],upd[irb*9+1],upd[irb*9+2]));
   ////
   Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian( Com::CVector3D(upd[irb*9+3],upd[irb*9+4],upd[irb*9+5]) );
   rb.AddRotation_IniLeft(rot);  
   ////
   double xlen=ri.xlen; 
   double ylen=ri.ylen;
   double zlen=ri.zlen;    
   const double rho = rb.GetMass()/(xlen*ylen*zlen);    
   xlen += upd[irb*9+6];
   ylen += upd[irb*9+7];
   zlen += upd[irb*9+8];
   rb.SetModeBox(xlen, ylen, zlen, rho);
   }
   */  
  // check if all joint is free
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    for(unsigned int idim=0;idim<3;idim++){
      bool flg = false;
      for(unsigned int iop=0;iop<aOptDof.size();iop++){
        const COptimizationDof& od = aOptDof[iop];          
        if( od.itype == JNT_POS && od.idim == idim ){
          flg = true;
          break;
        }
      }
      if( !flg ){
        std::cout << "joint should be free" << std::endl;
        assert(0);
      }
    }
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    CJoint_Fix& cst = aFix[ifix];
    unsigned int idir0 = aFixDir[ifix].first;
    unsigned int idir1 = aFixDir[ifix].second;    
    if(      (idir0==4||idir0==5) && (idir1!=4&&idir1!=5) ){  // joint is at 1
      unsigned int irb1 = cst.irb1;      
      const CRigid3D_Ini& ri1 = aRI[irb1];           
      const Com::CVector3D& d1 = ri1.GetIniPosCG();            
      const Com::CMatrix3& Q1 = ri1.GetRotMatrix_Ini();      
      const Com::CVector3D a1(ri1.xlen,ri1.ylen,ri1.zlen);        
      const Com::CVector3D& b1 = aFaceNormal[idir1]*0.5;      
      const Com::CMatrix3 B1(b1.x,b1.y,b1.z);        
      const Com::CMatrix3& QB1 = Q1.MatMat(B1);        
      const Com::CVector3D& QBa1 = QB1.MatVec(a1);            
      cst.ini_pos_joint = d1 + QBa1;
    }
    else if( (idir0!=4&&idir0!=5) && (idir1==4||idir1==5) ){  // joint is at 0
      unsigned int irb0 = cst.irb0;
      const CRigid3D_Ini& ri0 = aRI[irb0];
      const Com::CVector3D& d0 = ri0.GetIniPosCG();      
      const Com::CMatrix3& Q0 = ri0.GetRotMatrix_Ini();      
      const Com::CVector3D a0(ri0.xlen,ri0.ylen,ri0.zlen);      
      const Com::CVector3D& b0 = aFaceNormal[idir0]*0.5;      
      const Com::CMatrix3 B0(b0.x,b0.y,b0.z);
      const Com::CMatrix3& QB0 = Q0.MatMat(B0);
      const Com::CVector3D& QBa0 = QB0.MatVec(a0);      
      cst.ini_pos_joint = d0 + QBa0;      
    }    
  }  
}

void Projection_JointConnectivity
(const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CRigid3D_Def>& aRD,  
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
 const std::vector<COptimizationDof>& aOptDof,
 const std::vector<double>& upd_in, // same size as aOptDof
 std::vector<double>& upd_out,      // same size as aEvalDevalSensPRLM
 const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm, double stiff_cont)
{
  unsigned int nRB = aRI.size();
  unsigned int nConst = aFix.size();
  const unsigned int nblk = nRB+nConst;  
  std::vector<unsigned int> m_aBlkSize;
  {
    m_aBlkSize.resize(nblk);
    for(unsigned int irb=0;irb<nRB;irb++){
      m_aBlkSize[irb] = 9;
    }
    for(unsigned int icst=0;icst<nConst;icst++){
      m_aBlkSize[nRB+icst] = 1;
    }
  }
  
  
  MatVec::CMatDia_BlkCrs m_mat(nblk,m_aBlkSize);
  MatVec::CVector_Blk m_r(nblk,m_aBlkSize);
  SetPattern_JointConnectivity(m_mat,m_r,aRI.size(),aFix);
  MatVec::CBCFlag m_flg(nblk,m_aBlkSize);
  {
    m_flg.SetAllFlag(1); 
    for(unsigned int iop=0;iop<aOptDof.size();iop++){
      const COptimizationDof& od = aOptDof[iop];
      if( od.itype == BOX_POS ){ m_flg.SetBC(od.iobj,0+od.idim,0); }
      if( od.itype == BOX_ROT ){ m_flg.SetBC(od.iobj,3+od.idim,0); }
      if( od.itype == BOX_LEN ){ m_flg.SetBC(od.iobj,6+od.idim,0); }      
    }
    for(unsigned int ifix=0;ifix<aFix.size();ifix++){
      m_flg.SetBC(ifix+nRB,0,0);
    }
  }
  MatVec::CMatDiaFrac_BlkCrs m_frac;
  m_frac.MakePattern_Initialize( m_mat );
  m_frac.AddFracPtn(-1);
  m_frac.MakePatternFinalize();
  
  ////////////////////////////////////////////////
  // Make Linear System
  m_mat.SetZero();
  m_r.SetVectorZero();
  
  for(unsigned int iop=0;iop<aOptDof.size();iop++){
    const COptimizationDof& od = aOptDof[iop];    
    if( od.itype == BOX_POS ){ m_r.SetValue(od.iobj,0+od.idim, upd_in[iop]); }
    if( od.itype == BOX_ROT ){ m_r.SetValue(od.iobj,3+od.idim, upd_in[iop]); }
    if( od.itype == BOX_LEN ){ m_r.SetValue(od.iobj,6+od.idim, upd_in[iop]); }
  }
  
  for(unsigned int irb=0;irb<nRB;irb++){
    Com::CMatrix3 m0; m0.SetIdentity();
    AddMatrix(m_mat, irb,0,  irb,0,  m0,1, true);
    AddMatrix(m_mat, irb,3,  irb,3,  m0,1, true);            
    AddMatrix(m_mat, irb,6,  irb,6,  m0,1, true);                
  }
  for(unsigned int irb=0;irb<nRB;irb++){
    const CRigid3D_Ini& ri = aRI[irb];        
    const CContactPoints_RigidBody3D& aCP = aFP[irb];    
    for(unsigned int i=0;i<aCP.Size();i++){      
      if( !aCP.aFP[i].is_force_cont ) continue;
      Com::CMatrix3 B0 = aCP.GetSensMat_ContPointParam(i); 
      Com::CVector3D L0(ri.xlen,ri.ylen,ri.zlen);
      Com::CVector3D a0 = B0.MatVec(L0);
      Com::CVector3D dce_dd, dce_ds, dce_da;
      GetDerIni_IniContactEnergy(dce_dd,dce_ds,dce_da,                                   
                                 ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                 a0,cont_base,cont_norm,stiff_cont);
      const Com::CVector3D& dce_dl = B0.MatVec(dce_da);
      AddVector(m_r,irb,0,dce_dd,-1);
      AddVector(m_r,irb,3,dce_ds,-1);
      AddVector(m_r,irb,6,dce_dl,-1);      
      
      ////
      Com::CMatrix3 dce_dddd, dce_ddds, dce_ddda;
      Com::CMatrix3 dce_dsdd, dce_dsds, dce_dsda;
      Com::CMatrix3 dce_dadd, dce_dads, dce_dada;      
      GetDerIniIni_IniContactEnergy(dce_dddd, dce_ddds, dce_ddda,
                                    dce_dsdd, dce_dsds, dce_dsda,
                                    dce_dadd, dce_dads, dce_dada, 
                                    ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                    a0,cont_base,cont_norm,stiff_cont);
      const Com::CMatrix3& dce_dddl = dce_ddda.MatMat(B0);
      const Com::CMatrix3& dce_dsdl = dce_dsda.MatMat(B0);
      const Com::CMatrix3& dce_dadl = dce_dada.MatMat(B0);      
      const Com::CMatrix3& dce_dldd = B0.MatMat(dce_dadd);
      const Com::CMatrix3& dce_dlds = B0.MatMat(dce_dads);
      const Com::CMatrix3& dce_dldl = B0.MatMat(dce_dadl);      
      AddMatrix(m_mat, irb,0,  irb,0,  dce_dddd,1, true);
      AddMatrix(m_mat, irb,0,  irb,3,  dce_ddds,1, true);            
      AddMatrix(m_mat, irb,0,  irb,6,  dce_dddl,1, true);
      AddMatrix(m_mat, irb,3,  irb,0,  dce_dsdd,1, true);
      AddMatrix(m_mat, irb,3,  irb,3,  dce_dsds,1, true);            
      AddMatrix(m_mat, irb,3,  irb,6,  dce_dsdl,1, true);                      
      AddMatrix(m_mat, irb,6,  irb,0,  dce_dldd,1, true);
      AddMatrix(m_mat, irb,6,  irb,3,  dce_dlds,1, true);            
      AddMatrix(m_mat, irb,6,  irb,6,  dce_dldl,1, true);                            
    }
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    AddLinSys_JointConnectivity(m_mat,m_r, ifix,aFix[ifix], aRI, 
                                aFixDir[ifix].first,aFixDir[ifix].second);
    AddLinSys_JointInside(m_mat,m_r, ifix,aFix[ifix], aRI,
                          aFixDir[ifix].first,aFixDir[ifix].second,
                          1);
  }
  
  ////////////////////////////////
  m_mat.SetBoundaryCondition(m_flg);
  m_flg.SetZeroToBCDof(m_r);
  m_frac.SetValue( m_mat );
  m_frac.Solve( m_r );       
  
  upd_out.resize(aOptDof.size());
  for(unsigned int iop=0;iop<aOptDof.size();iop++){
    const COptimizationDof& od = aOptDof[iop];    
    if( od.itype == BOX_POS ){ upd_out[iop] = m_r.GetValue(od.iobj,0+od.idim); }
    if( od.itype == BOX_ROT ){ upd_out[iop] = m_r.GetValue(od.iobj,3+od.idim); }
    if( od.itype == BOX_LEN ){ upd_out[iop] = m_r.GetValue(od.iobj,6+od.idim); }
  }  
}

void JointSensitivity
(const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,                       
 const std::vector<COptimizationDof>& aOptDof,
 const  std::vector<double>& upd, // aOptDof.size()
 std::vector<Com::CVector3D>& aJointSens)
{
  std::vector< std::vector< std::pair<unsigned int,unsigned int> > > aJntDir;
  aJntDir.resize(aRI.size());
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int irb0 = aFix[ifix].irb0;
    unsigned int irb1 = aFix[ifix].irb1;
    unsigned int idir0 = aFixDir[ifix].first;      
    unsigned int idir1 = aFixDir[ifix].second;            
    aJntDir[irb0].push_back( std::make_pair(ifix,idir0) );
    aJntDir[irb1].push_back( std::make_pair(ifix,idir1) );      
  }
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };      
  aJointSens.resize( aFix.size() );
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){ aJointSens[ifix] = Com::CVector3D(0,0,0); }
  for(unsigned int iop=0;iop<aOptDof.size();iop++){
    if( aOptDof[iop].itype == JNT_POS ) continue;  
    const COptimizationDof& od = aOptDof[iop];
    unsigned int irb = aOptDof[iop].iobj;
    //    const CRigidBody3D& rb = aRB[irb];
    const CRigid3D_Ini& ri = aRI[irb];
    Com::CVector3D e(0,0,0); e[od.idim] = 1;
    for(unsigned int ij=0;ij<aJntDir[irb].size();ij++){
      unsigned int ifix = aJntDir[irb][ij].first;
      unsigned int idir = aJntDir[irb][ij].second;
      //      std::cout << irb << " " << ifix << " " << idir << std::endl;
      if( idir == 4 || idir == 5 ) continue;
      const Com::CVector3D& b = aFaceNormal[idir]*0.5;
      const Com::CMatrix3 B(b.x,b.y,b.z);
      const Com::CVector3D a(ri.xlen,ri.ylen,ri.zlen);
      const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
      const Com::CMatrix3& QB = Q.MatMat(B);
      const Com::CVector3D& QBa = QB.MatVec(a);
      const Com::CVector3D v = ri.GetIniPosCG() + QBa;
      if( od.itype == BOX_POS ){ aJointSens[ifix] += upd[iop]*e; }
      if( od.itype == BOX_ROT ){ aJointSens[ifix] += upd[iop]*-Cross(QBa,e); }
      if( od.itype == BOX_LEN ){ aJointSens[ifix] += upd[iop]*QB.MatVec(e); }
    }
  }  
}


void JointDegeneration
(const std::vector<CRigid3D_Ini>& aRI,
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,                       
 const std::vector<COptimizationDof>& aOptDof,
 const std::vector<Com::CVector3D>& aGradJnt,
 std::vector<double>& upd )
{ // joint position degeneration
  std::vector< std::vector< std::pair<unsigned int,unsigned int> > > aJntDir;
  aJntDir.resize(aRI.size());
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int irb0 = aFix[ifix].irb0;
    unsigned int irb1 = aFix[ifix].irb1;
    unsigned int idir0 = aFixDir[ifix].first;      
    unsigned int idir1 = aFixDir[ifix].second;            
    aJntDir[irb0].push_back( std::make_pair(ifix,idir0) );
    aJntDir[irb1].push_back( std::make_pair(ifix,idir1) );      
  }
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };    
  for(unsigned int iop=0;iop<aOptDof.size();iop++){
    if( aOptDof[iop].itype == JNT_POS ) continue;  
    const COptimizationDof& od = aOptDof[iop];
    unsigned int irb = aOptDof[iop].iobj;
    //    const CRigidBody3D& rb = aRB[irb];
    const CRigid3D_Ini& ri = aRI[irb];        
    Com::CVector3D e(0,0,0); e[od.idim] = 1;
    for(unsigned int ij=0;ij<aJntDir[irb].size();ij++){
      unsigned int ifix = aJntDir[irb][ij].first;
      unsigned int idir = aJntDir[irb][ij].second;
      //      std::cout << irb << " " << ifix << " " << idir << std::endl;
      if( idir == 4 || idir == 5 ) continue;
      const Com::CVector3D& b = aFaceNormal[idir]*0.5;
      const Com::CMatrix3 B(b.x,b.y,b.z);
      const Com::CVector3D a(ri.xlen,ri.ylen,ri.zlen);
      const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
      const Com::CMatrix3& QB = Q.MatMat(B);
      const Com::CVector3D& QBa = QB.MatVec(a);
      const Com::CVector3D v = ri.GetIniPosCG() + QBa;
      if( od.itype == BOX_POS ){ upd[iop] += Com::Dot(aGradJnt[ifix],e); }
      if( od.itype == BOX_ROT ){ upd[iop] += Com::Dot(aGradJnt[ifix],-Cross(QBa,e)); }
      if( od.itype == BOX_LEN ){ upd[iop] += Com::Dot(aGradJnt[ifix],QB.MatVec(e)); }
    }
  }
}

void GetDesignSensitivity(CDesignSensitivity& desSens, 
                          const std::vector<CRigid3D_Ini>& aRI,
                          const std::vector<CRigid3D_Def>& aRD,                          
                          const std::vector<CJoint_Fix>& aFix,
                          const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir,
                          const std::vector<COptimizationDof>& aOptDof,                          
                          const Com::CVector3D& cont_base, const Com::CVector3D& cont_norm,
                          const std::vector<CContactPoints_RigidBody3D>& aFP,                          
                          unsigned int irb_pic, const Com::CVector3D& v)
{
  unsigned int nRB = aRI.size();
  std::vector<double> upd_in;
  {
    std::vector<double> upd_in_tmp;                      
    upd_in_tmp.resize( nRB*9,0);
    upd_in_tmp[irb_pic*9+0] = v.x;
    upd_in_tmp[irb_pic*9+1] = v.y;
    upd_in_tmp[irb_pic*9+2] = v.z;
    upd_in.resize(aOptDof.size());
    for(unsigned int iop=0;iop<aOptDof.size();iop++){
      const COptimizationDof& od = aOptDof[iop];
      double val = 0;
      if( od.itype == BOX_POS ){ val = upd_in_tmp[od.iobj*9+od.idim+0]; }
      if( od.itype == BOX_ROT ){ val = upd_in_tmp[od.iobj*9+od.idim+3]; }
      if( od.itype == BOX_LEN ){ val = upd_in_tmp[od.iobj*9+od.idim+6]; }                    
      upd_in[iop] = val;
    }
  }
  std::vector<double> upd_out_tmp;          
  Projection_JointConnectivity(aRI,aRD, aFix, aFP, aFixDir, aOptDof, 
                               upd_in, upd_out_tmp, 
                               cont_base, cont_norm, 1000);
  std::vector<double> upd_out;
  {
    upd_out.resize( nRB*9,0 );
    for(unsigned int iop=0;iop<aOptDof.size();iop++){
      const COptimizationDof& od = aOptDof[iop];
      double val = upd_out_tmp[iop];
      if( od.itype == BOX_POS ){ upd_out[od.iobj*9+od.idim+0] = val; }
      if( od.itype == BOX_ROT ){ upd_out[od.iobj*9+od.idim+3] = val; }
      if( od.itype == BOX_LEN ){ upd_out[od.iobj*9+od.idim+6] = val; }                    
    }            
  }
  JointSensitivity(aRI,aFix,aFixDir,aOptDof, upd_out_tmp,desSens.aJoint);                    
  desSens.aPlank.resize( aRI.size() );
  for(unsigned int irb=0;irb<aRI.size();irb++){
    desSens.aPlank[irb].p.x = upd_out[irb*9+0];
    desSens.aPlank[irb].p.y = upd_out[irb*9+1];
    desSens.aPlank[irb].p.z = upd_out[irb*9+2];
    desSens.aPlank[irb].r.x = upd_out[irb*9+3];
    desSens.aPlank[irb].r.y = upd_out[irb*9+4];
    desSens.aPlank[irb].r.z = upd_out[irb*9+5];
    desSens.aPlank[irb].l.x = upd_out[irb*9+6];
    desSens.aPlank[irb].l.y = upd_out[irb*9+7];
    desSens.aPlank[irb].l.z = upd_out[irb*9+8];                                   
  }  
}



/*
 void CJointedRigidBody::MoveJointTwardBendless(unsigned int ifix, double ratio)
 {  
 CJoint_Fix& cjs = aFix[ifix];
 const unsigned int irb0 = cjs.aIndRB[0];
 const CRigidBody3D& rb0 = aRB[irb0];
 
 const unsigned int irb1 = cjs.aIndRB[1];
 const CRigidBody3D& rb1 = aRB[irb1];
 
 const Com::CMatrix3& mrot0 = rb0.GetRotMatrix_Inc();
 const Com::CMatrix3& mrot1 = rb1.GetRotMatrix_Inc();
 const Com::CVector3D vlambda(cjs.lambda[0],cjs.lambda[1],cjs.lambda[2]);
 const Com::CVector3D Xdistfix0 = rb0.GetIniPosCG() - cjs.ini_pos_joint;
 const Com::CVector3D Xdistfix1 = rb1.ri.GetIniPosCG() - cjs.ini_pos_joint;  
 
 const Com::CVector3D& R0tvl = mrot0.MatVecTrans(vlambda);
 const Com::CVector3D vRes0 = Com::CMatrix3(Xdistfix0).MatVec(R0tvl);
 
 const Com::CVector3D& R0x = mrot0.MatVec(cjs.loc_coord[0]);
 const Com::CVector3D& R0y = mrot0.MatVec(cjs.loc_coord[1]);
 const Com::CVector3D& R0z = mrot0.MatVec(cjs.loc_coord[2]);
 ////
 const Com::CVector3D& R1x = mrot1.MatVec(cjs.loc_coord[0]);
 const Com::CVector3D& R1y = mrot1.MatVec(cjs.loc_coord[1]);
 const Com::CVector3D& R1z = mrot1.MatVec(cjs.loc_coord[2]);    
 ////
 const Com::CVector3D& wR0yR1z = Com::CMatrix3(R0y).MatVec(R1z);
 const Com::CVector3D& wR0zR1x = Com::CMatrix3(R0z).MatVec(R1x);
 const Com::CVector3D& wR0xR1y = Com::CMatrix3(R0x).MatVec(R1y);
 Com::CVector3D Rtwma[3];
 Rtwma[0] = mrot0.MatVecTrans(wR0yR1z);
 Rtwma[1] = mrot0.MatVecTrans(wR0zR1x);
 Rtwma[2] = mrot0.MatVecTrans(wR0xR1y);
 Com::CVector3D tRes0 = cjs.lambda[3]*Rtwma[0] + cjs.lambda[4]*Rtwma[1] + cjs.lambda[5]*Rtwma[2];
 
 Com::CVector3D Res0 = vRes0+tRes0;
 //  std::cout << "vRes : " << vRes0.x << " " << vRes0.y << " " << vRes0.z << std::endl;
 //  std::cout << "tRes : " << tRes0.x << " " << tRes0.y << " " << tRes0.z << std::endl;      
 std::cout << Res0.x << " " << Res0.y << " " << Res0.z << std::endl;
 
 
 //   {  
 //   Com::CVector3D tlambda(jF.lambda[3],jF.lambda[4],jF.lambda[5]);
 //   if( tlambda.Length() < 0.05 ) return;
 //   }
 
 
 Com::CVector3D a = R0tvl;
 Com::CVector3D b = -Res0+Com::CMatrix3(rb0.GetIniPosCG()).MatVec(R0tvl);
 Com::CVector3D wba = Cross(b,a);
 
 //   if( wba.Length() < 0.001 || a.Length() < 0.001 ){
 //   std::cout << wba.Length() << " " << a.Length() << std::endl;
 //   }
 
 wba.SetNormalizedVector();
 wba *= -(b.Length()/a.Length());
 
 double t = Dot(wba-cjs.ini_pos_joint,a) / Dot(a,a);
 wba -= t*a;
 
 cjs.ini_pos_joint = (1-ratio)*cjs.ini_pos_joint + ratio*wba;  
 }
 */





/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////

/*
 double CJointedRigidBody::GetHessianEvaluationFunction
 (PARAM_TYPE itype_s, unsigned int irb_s, unsigned int idim_s,
 const std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aiSensPR,
 const std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aiSensLM,
 PARAM_TYPE jtype_s, unsigned int jrb_s, unsigned int jdim_s,
 const std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& ajSensPR,
 const std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& ajSensLM,
 double coeff_joint) const
 {
 if( irb_s != jrb_s ) return 0;
 assert( irb_s < aRB.size() );
 assert( jrb_s < aRB.size() );  
 assert( idim_s < 3 );
 assert( aiSensPR.size() == aRB.size() );
 assert( aiSensLM.size() == aFix.size() );
 assert( ajSensPR.size() == aRB.size() );
 assert( ajSensLM.size() == aFix.size() );  
 assert( aFP.size() == aRB.size()*8 );
 
 //  double ddeng_cont = 0;
 //  {
 //    const double corner[8][3] = { 
 //      {-1,-1,-1}, {-1,-1,+1}, {-1,+1,-1}, {-1,+1,+1},
 //      {+1,-1,-1}, {+1,-1,+1}, {+1,+1,-1}, {+1,+1,+1} };        
 //    for(unsigned int irb=0;irb<aRB.size();irb++){
 //      const CRigidBody3D& rb = aRB[irb];
 //      for(unsigned int i=0;i<8;i++){
 //        const Com::CVector3D v0 = Com::CVector3D(corner[i][0]*rb.xlen, corner[i][1]*rb.ylen, corner[i][2]*rb.zlen )*0.5;
 //        const Com::CVector3D v1 = rb.GetCurrentPosition(v0);
 //        double hn = -Com::Dot(v1-cont_base,cont_norm);
 //        if( hn < 0 ) continue;
 //        ////
 //        Com::CVector3D div1 = GetSensitivityPointOnRigidBody(itype_s,irb_s,idim_s, irb,rb, i, aiSensPR[irb] );
 //        Com::CVector3D djv1 = GetSensitivityPointOnRigidBody(jtype_s,jrb_s,jdim_s, irb,rb, i, aiSensPR[irb] );      
 //        //      ddeng += stiff_n*Com::Dot(div1,cont_norm)*Com::Dot(djv1,cont_norm);
 //        if( aFP[irb*8+i].itype_contact == 1 ){
 //          Com::CVector3D diapt = div1 - cont_norm*Dot(cont_norm,div1);
 //          Com::CVector3D djapt = djv1 - cont_norm*Dot(cont_norm,djv1);        
 //          //        ddeng += stiff_f*Com::Dot(djapt,diapt);
 //        }
 //      }
 //    }    
 //  }
 double ddeng_bend = 0;
 for(unsigned int ifix=0;ifix<aFix.size();ifix++){
 const CJoint_Fix& cjs = aFix[ifix];    
 ddeng_bend += (1.0/cjs.invstiff_rot)*Dot(ajSensLM[ifix].second,aiSensLM[ifix].second);
 }  
 return ddeng_bend;    
 }*/

/*
 // using hessian plank wise
 void CJointedRigidBody::OptimizeEvalFunc_HessPlank
 (unsigned int irb_s, double alpha, 
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc)
 {
 double a[9];
 
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensPR_pos0;
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensLM_pos0;
 this->GetSensitivityPRLM_InitialConfigBox2(aSensPR_pos0,aSensLM_pos0, BOX_POS,irb_s,0, itype_anchor_reloc); 
 
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensPR_pos1;
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensLM_pos1;
 this->GetSensitivityPRLM_InitialConfigBox2(aSensPR_pos1,aSensLM_pos1, BOX_POS,irb_s,1, itype_anchor_reloc);   
 
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensPR_rot2;
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensLM_rot2;
 this->GetSensitivityPRLM_InitialConfigBox2(aSensPR_rot2,aSensLM_rot2, BOX_ROT,irb_s,2, itype_anchor_reloc);     
 
 a[0*3+0] = this->GetHessianEvaluationFunction(BOX_POS,irb_s,0,aSensPR_pos0,aSensLM_pos0, BOX_POS,irb_s,0,aSensPR_pos0,aSensLM_pos0, 0);
 a[0*3+1] = this->GetHessianEvaluationFunction(BOX_POS,irb_s,0,aSensPR_pos0,aSensLM_pos0, BOX_POS,irb_s,1,aSensPR_pos1,aSensLM_pos1, 0);  
 a[0*3+2] = this->GetHessianEvaluationFunction(BOX_POS,irb_s,0,aSensPR_pos0,aSensLM_pos0, BOX_ROT,irb_s,2,aSensPR_rot2,aSensLM_rot2, 0);
 a[1*3+0] = a[0*3+1];
 a[1*3+1] = this->GetHessianEvaluationFunction(BOX_POS,irb_s,1,aSensPR_pos1,aSensLM_pos1, BOX_POS,irb_s,1,aSensPR_pos1,aSensLM_pos1, 0);  
 a[1*3+2] = this->GetHessianEvaluationFunction(BOX_POS,irb_s,1,aSensPR_pos1,aSensLM_pos1, BOX_ROT,irb_s,2,aSensPR_rot2,aSensLM_rot2, 0);  
 a[2*3+0] = a[0*3+2];
 a[2*3+1] = a[1*3+2];
 a[2*3+2] = this->GetHessianEvaluationFunction(BOX_ROT,irb_s,2,aSensPR_rot2,aSensLM_rot2, BOX_ROT,irb_s,2,aSensPR_rot2,aSensLM_rot2, 0);    
 
 {
 int info;
 CalcInvMat(a,3,info);
 }
 
 const double b[3] = {
 -GetDerivativeEvaluationFunction(BOX_POS,irb_s,0, aSensPR_pos0,aSensLM_pos0, 0,0),
 -GetDerivativeEvaluationFunction(BOX_POS,irb_s,1, aSensPR_pos1,aSensLM_pos1, 0,0),    
 -GetDerivativeEvaluationFunction(BOX_ROT,irb_s,2, aSensPR_rot2,aSensLM_rot2, 0,0) };
 
 const double x[3] = {
 a[0*3+0]*b[0] + a[0*3+1]*b[1] + a[0*3+2]*b[2], 
 a[1*3+0]*b[0] + a[1*3+1]*b[1] + a[1*3+2]*b[2], 
 a[2*3+0]*b[0] + a[2*3+1]*b[1] + a[2*3+2]*b[2] };
 
 
 this->MoveInitialConfig(BOX_POS,irb_s,0,x[0]*alpha);
 this->MoveCurrentConfig(aSensPR_pos0,aSensLM_pos0,x[0]*alpha);
 this->RelocateAnchors(BOX_POS, irb_s, 0, aSensPR_pos0[irb_s], x[0]*alpha, itype_anchor_reloc);  
 
 
 this->MoveInitialConfig(BOX_POS,irb_s,1,x[1]*alpha);
 this->MoveCurrentConfig(aSensPR_pos1,aSensLM_pos1,x[1]*alpha);
 this->RelocateAnchors(BOX_POS, irb_s, 1, aSensPR_pos1[irb_s], x[1]*alpha, itype_anchor_reloc); 
 
 this->MoveInitialConfig(BOX_ROT,irb_s,2,x[2]*alpha);
 this->MoveCurrentConfig(aSensPR_rot2,aSensLM_rot2,x[2]*alpha);
 this->RelocateAnchors(BOX_ROT, irb_s, 2, aSensPR_rot2[irb_s], x[2]*alpha, itype_anchor_reloc);
 
 //  double eng0 = GetEvaluationFunction();  
 //  std::cout << irb_s << "  eng:" << eng0 << std::endl;
 }
 */

/*
 void CJointedRigidBody::OptimizeEvalFunc_Sym_Grad
 (PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s, 
 unsigned int iobj_m, unsigned int iaxis_m,
 double alpha, ANCHOR_RELOCATION_TYPE itype_anchor_reloc)
 {
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensPR_s;
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensLM_s;
 GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem(aSensPR_s,aSensLM_s, 
 itype_s,iobj_s,idim_s, itype_anchor_reloc,
 aRB,aFix,aExForce,
 gravity,
 cont_base,cont_norm,stiff_n,stiff_f,
 aFP);
 
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensPR_m;
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensLM_m;
 GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem(aSensPR_m,aSensLM_m, 
 itype_s,iobj_m,idim_s, itype_anchor_reloc,
 aRB,aFix,aExForce,
 gravity,
 cont_base,cont_norm,stiff_n,stiff_f,
 aFP);  
 
 //  double eng0 = GetEvaluationFunction(penalty_coeff_joint,penalty_coeff_cont,
 //                                      aRB,aFix,aFixDir,
 //                                      cont_base,cont_norm,
 //                                      aFP);  
 
 double gradient_s = GetDerivativeEvaluationFunction(itype_s,iobj_s,idim_s, 
 aSensPR_s,aSensLM_s,
 penalty_coeff_joint, penalty_coeff_cont,
 aRB,aFix,aFixDir,
 cont_base,cont_norm,
 aFP);
 
 double gradient_m = GetDerivativeEvaluationFunction(itype_s,iobj_m,idim_s, 
 aSensPR_m,aSensLM_m,
 penalty_coeff_joint, penalty_coeff_cont,
 aRB,aFix,aFixDir,
 cont_base,cont_norm,
 aFP);
 
 double flg = 1;
 if( iaxis_m == 0 ){
 if( itype_s == BOX_ROT ){
 if( idim_s == 1 || idim_s == 2 ){ flg = -1; }
 }
 else{
 if( idim_s == 0 ){ flg = -1; }
 }
 }
 
 double eps = -( gradient_s + flg*gradient_m )*alpha;
 MoveInitialConfig_JointedRigidBodyContactSystem(itype_s, iobj_s, idim_s, eps, aRB,aFix);
 MoveCurrentConfig_JointedRigidBodyContactSystem(aSensPR_s,aSensLM_s,eps, aRB,aFix);  
 
 MoveInitialConfig_JointedRigidBodyContactSystem(itype_s, iobj_m, idim_s, flg*eps, aRB,aFix);
 MoveCurrentConfig_JointedRigidBodyContactSystem(aSensPR_m,aSensLM_m,flg*eps, aRB,aFix);  
 
 if( itype_s == BOX_POS || itype_s == BOX_ROT || itype_s == BOX_LEN ){
 RelocateAnchors_JointedRigidBodyContactSystem(aFP,
 itype_s,iobj_s,idim_s, 
 itype_anchor_reloc,
 aSensPR_s[iobj_s], eps, 
 aRB,cont_norm);     
 RelocateAnchors_JointedRigidBodyContactSystem(aFP,
 itype_s,iobj_m,idim_s, 
 itype_anchor_reloc,
 aSensPR_m[iobj_s], flg*eps, 
 aRB,cont_norm);     
 }
 }*/




/*
 void AddLinSys_JointConnectivity
 (MatVec::CMatDia_BlkCrs& m_mat,     
 MatVec::CVector_Blk& m_residual,
 unsigned int icst, const CJoint_Fix& cst,
 const std::vector<CRigidBody3D>& aRB,
 unsigned int idir0, unsigned int idir1) 
 */



/*
 void CJointedRigidBody::OptimizeEvalFunc_MultDir()
 { 
 FindInitialConfiguration_SameFrictionAnchor();    
 for(int i=0;i<100;i++){
 bool res = this->StepTime_BackwardEular_UntactAnchor(0.1);	// solve rigid motion    
 if( res ) break;
 }      
 
 //  std::cout << "Get Sensitivity w.r.t initial configuration (improved version) " << std::endl;
 MatVec::CMatDia_BlkCrs m_mat;
 MatVec::CVector_Blk m_r;
 SetPatternDynamicSystem(m_mat,m_r,aRB,aFix);
 
 MatVec::CMatDiaFrac_BlkCrs m_frac;
 m_frac.MakePattern_Initialize( m_mat );
 m_frac.AddFracPtn(-1); // compute exact inverse
 m_frac.MakePatternFinalize();
 
 ////////////////////////////////////////////////////////////////
 // Make Linear System
 m_mat.SetZero();
 for(unsigned int irb=0;irb<aRB.size();irb++){
 AddMatrix_StaticContact(m_mat, irb,aRB[irb],
 cont_base,cont_norm,stiff_n,stiff_f,aFP);
 {
 Com::CMatrix3 m0; m0.SetIdentity();
 AddMatrix(m_mat, irb,0,  irb,0,  m0,1.0e-8, true);
 AddMatrix(m_mat, irb,3,  irb,3,  m0,1.0e-8, true);            
 } 
 }
 for(unsigned int ifix=0;ifix<aFix.size();ifix++){
 const CJoint_Fix& cjs = aFix[ifix];
 AddMatrix_StaticJoint(m_mat,ifix,cjs,aRB);    
 }
 for(unsigned int ief=0;ief<aExForce.size();ief++){
 const CExternalForceOnRigidBody& ef = aExForce[ief];
 unsigned int irb = ef.irb;
 AddMatrix_ExternalForce_Static(m_mat, irb, aRB[irb], ef.pos, ef.force );
 }
 // Solve Linear System
 m_frac.SetValue( m_mat );
 
 
 std::vector<Com::CVector3D> aForce;
 aForce.push_back( Com::CVector3D(1,0,0) );
 aForce.push_back( Com::CVector3D(0,0,1) );
 
 std::vector< std::vector<Com::CVector3D> > aBend;
 aBend.resize(2);
 
 
 for(unsigned int i=0;i<2;i++){
 m_r.SetVectorZero();
 
 for(unsigned int irb=0;irb<aRB.size();irb++){
 AddVector(m_r, irb,0, aForce[i],  1);
 }
 
 m_frac.Solve( m_r );
 
 double tmp[6];    
 for(unsigned int icst=0;icst<aFix.size();icst++){
 const unsigned int nlen = m_r.Len(icst+aRB.size()); assert( nlen == 6 );
 for(unsigned int ilen=0;ilen<nlen;ilen++){ tmp[ilen] = m_r.GetValue(icst+aRB.size(),ilen); }
 aBend[i].push_back( Com::CVector3D(tmp[3],tmp[4],tmp[5]) );
 }      
 }
 
 ////////////////////////////////////////////////////////////////  
 }
 */


/*
 void CJointedRigidBody::OptimizeEvalFunc_GradMultDir(PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s, 
 double alpha, ANCHOR_RELOCATION_TYPE itype_anchor_reloc)
 {
 double eval1, deval1;
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensPR1;
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensLM1;
 {
 std::vector<CExternalForceOnRigidBody> aMyExForce;     
 aMyExForce.push_back( CExternalForceOnRigidBody(0,Com::CVector3D(0,0,0), Com::CVector3D(+3.0,-6,0) ) );
 GetGradEvaluationFunctionAndSens(eval1,deval1,
 aSensPR1,aSensLM1,
 penalty_coeff_joint,penalty_coeff_cont,
 itype_s,iobj_s,idim_s,
 itype_anchor_reloc,
 aRB,aFix,aFixDir,aMyExForce,
 gravity,
 cont_base,cont_norm,stiff_n,stiff_f,
 aFP);
 }  
 double eval2, deval2;
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensPR2;
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensLM2;  
 {
 std::vector<CExternalForceOnRigidBody> aMyExForce;     
 aMyExForce.push_back( CExternalForceOnRigidBody(0,Com::CVector3D(0,0,0), Com::CVector3D(-3.0,-6,0) ) );
 GetEvaluationFunctionAndGrad(eval2,deval2,
 aSensPR2,aSensLM2, 
 penalty_coeff_joint,penalty_coeff_cont,
 itype_s,iobj_s,idim_s,
 itype_anchor_reloc,
 aRB,aFix,aFixDir,aMyExForce,
 gravity,
 cont_base,cont_norm,stiff_n,stiff_f,
 aFP);
 }      
 double eval0, deval0;  
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensPR0;
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensLM0;  
 {
 std::vector<CExternalForceOnRigidBody> aMyExForce;     
 aMyExForce.push_back( CExternalForceOnRigidBody(0,Com::CVector3D(0,0,0), Com::CVector3D(+0.0,-6,0) ) );    
 GetEvaluationFunctionAndGrad(eval0,deval0,
 aSensPR0,aSensLM0, 
 penalty_coeff_joint,penalty_coeff_cont,
 itype_s,iobj_s,idim_s,
 itype_anchor_reloc,
 aRB,aFix,aFixDir,aMyExForce,
 gravity,
 cont_base,cont_norm,stiff_n,stiff_f,
 aFP);
 }  
 //  std::cout << eval0 << std::endl;
 double eps = -(0.1*deval1+0.1*deval2+0.8*deval0)*alpha;
 //  double eps = -0.5*deval0*alpha;  
 //  double eps = -deval2*alpha;  
 //  std::cout << eps << std::endl;
 MoveInitialConfig_JointedRigidBodyContactSystem(itype_s,iobj_s,idim_s, eps, aRB,aFix);
 MoveCurrentConfig_JointedRigidBodyContactSystem(aSensPR0,aSensLM0,eps, aRB,aFix);  
 RelocateAnchors_JointedRigidBodyContactSystem(aFP,
 itype_s,iobj_s,idim_s, 
 itype_anchor_reloc,
 aSensPR0[iobj_s], eps, 
 aRB,cont_norm); 
 
 }
 */


/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////



void MakeSuggestionSupportMaterial(CJointedRigidBody& jrb)
{
  std::vector<CRigid3D_Ini> aRI = jrb.aRI;
  std::vector<CRigid3D_Def> aRD = jrb.aRD;
  const unsigned int nRB = aRI.size();
  double aJntCand[9][2] = {
    {-0.9,-0.7}, {+0.0,-0.7}, {+0.9,-0.7},
    {-0.9,-0.0}, {+0.0,-0.0}, {+0.9,-0.0},
    {-0.9,+0.7}, {+0.0,+0.7}, {+0.9,+0.7},            
  };
  
  Com::CVector3D U;
  Com::CMatrix3 P;
  { // set initial value of ICP
    double ave_crv[3] = {0,0,0};
    for(unsigned int irb=0;irb<nRB;irb++){        
      const Com::CMatrix3& P = aRD[irb].GetRotMatrix_Inc();
      double crv[3]; P.GetCRV_RotMatrix(crv);
      ave_crv[0] += crv[0];
      ave_crv[1] += crv[1];      
      ave_crv[2] += crv[2];      
    }    
    ave_crv[0] /= nRB;
    ave_crv[1] /= nRB;
    ave_crv[2] /= nRB;
    Com::CMatrix3 avP; avP.SetRotMatrix_CRV(ave_crv);
    const Com::CVector3D C = GetGravityCenter(false,aRI,aRD);
    P.SetRotMatrix_CRV(ave_crv);
    U = GetGravityCenter(true ,aRI,aRD) - C - P.MatVec(C);
  }
  
  Com::CMatrix3 aCnr[4] = {
    Com::CMatrix3(-0.5,-0.5,+0.0),
    Com::CMatrix3(+0.5,-0.5,+0.0),
    Com::CMatrix3(-0.5,+0.5,+0.0),
    Com::CMatrix3(+0.5,+0.5,+0.0) };  
  
  std::cout.setf( std::ios::scientific );
  for(unsigned int itr=0;itr<5;itr++){
    double eng = 0;
    Com::CVector3D dq_du, dq_dr;
    Com::CMatrix3 ddq_dudu, ddq_dudr, ddq_drdu, ddq_drdr;
    Com::CVector3D ave_cg_ini = GetGravityCenter(false,aRI,aRD);
    for(unsigned int irb=0;irb<nRB;irb++){
      const CRigid3D_Ini& ri = aRI[irb];
      const CRigid3D_Def& rd = aRD[irb];
      const Com::CVector3D l(ri.xlen,ri.ylen,ri.zlen);         
      const Com::CMatrix3& q = ri.GetRotMatrix_Ini();
      const Com::CMatrix3& p = rd.GetRotMatrix_Inc();
      for(unsigned int i=0;i<4;i++){
        Com::CVector3D v0 = aCnr[i].MatVec(l);
        const Com::CVector3D& v1 = q.MatVec(v0) + ri.GetIniPosCG();
        const Com::CVector3D& v2 = p.MatVec( q.MatVec(v0) ) + rd.GetDispCG() + ri.GetIniPosCG();
        const Com::CVector3D& v3 = P.MatVec(v1) + U;
        const double fe = GetSpringEnergy(v3,v2);
        Com::CVector3D dfe_du, dfe_dr;
        Get1stDer_SpringEnergy(dfe_du,dfe_dr, v3,v2);
        Com::CMatrix3 dfe_dudu, dfe_dudr, dfe_drdu, dfe_drdr;
        Get2ndDer_SpringEnergy(dfe_dudu,dfe_dudr,dfe_drdu,dfe_drdr, v3,v2); 
        eng += fe;
        dq_du += dfe_du;
        dq_dr += dfe_dr;
        ddq_dudu += dfe_dudu;
        ddq_dudr += dfe_dudr;
        ddq_drdu += dfe_drdu;  
        ddq_drdr += dfe_drdr;
      }
    }      
    std::cout << "energy " << sqrt(eng*2) << std::endl;
    {
      double A[36];
      double b[6];
      for(unsigned int i=0;i<3;i++){      
        b[0+i] = -dq_du[i];
        b[3+i] = -dq_dr[i];      
        for(unsigned int j=0;j<3;j++){     
          A[6*(i+0)+(j+0)] = ddq_dudu.mat[3*i+j];
          A[6*(i+0)+(j+3)] = ddq_dudr.mat[3*i+j];        
          A[6*(i+3)+(j+0)] = ddq_drdu.mat[3*i+j];
          A[6*(i+3)+(j+3)] = ddq_drdr.mat[3*i+j];
        }
      }
      double nres = 0;
      for(unsigned int i=0;i<6;i++){
        nres += b[i]*b[i];        
        //        std::cout << "b " << i << " " << b[i] << std::endl;      
      }
      int info;
      CalcInvMat(A, 6, info );
      double x[6];
      double nupd = 0;
      for(unsigned int i=0;i<6;i++){
        x[i] = 0;
        for(unsigned int j=0;j<6;j++){
          x[i] += A[6*i+j]*b[j];
        }
        nupd += x[i]*x[i];
        //        std::cout << "x " << i << " " << x[i] << std::endl;
      }
      //      std::cout << "nres : " << nres << "   nupd : " << nupd << std::endl;
      Com::CVector3D dU(x[0],x[1],x[2]);
      Com::CMatrix3 dP; dP.SetRotMatrix_Cartesian( Com::CVector3D(x[3],x[4],x[5]) );
      U = dP.MatVec(U) + dU;
      P = dP.MatMat(P);
    }    
  }
  
  //  std::cout << "cent " << C << " disp  " << U << std::endl;
  std::vector<Com::CVector3D> aDist;
  aDist.reserve(aRI.size()*9);
  for(unsigned int iri=0;iri<aRI.size();iri++){
    const CRigid3D_Ini& ri = aRI[iri];
    const Com::CMatrix3& q = ri.GetRotMatrix_Ini();
    const Com::CVector3D l(ri.xlen,ri.ylen,ri.zlen);
    const Com::CVector3D& d = ri.GetIniPosCG();
    const CRigid3D_Def& rd = aRD[iri];    
    const Com::CMatrix3& p = rd.GetRotMatrix_Inc();
    const Com::CVector3D& u = rd.GetDispCG();  
    for(unsigned int i=0;i<9;i++){
      const Com::CMatrix3 B(aJntCand[i][0]*0.5,aJntCand[i][1]*0.5,0);
      const Com::CVector3D& p0 = B.MatVec(l);      
      const Com::CVector3D& p1 = q.MatVec(p0) + d;
      const Com::CVector3D& p2 = p.MatVec(q.MatVec(p0)) + u + d;
      const Com::CVector3D& p3 = P.MatVec(p1) + U;
      const Com::CVector3D& del = p2-p3;
      aDist.push_back(del);
      //      std::cout << iri << " " << i << " --> " << del << "   " << Length(del) << std::endl;
    }
  }
  Com::CVector3D distMax0;
  unsigned int indMax0;
  {
    distMax0 = aDist[0];
    double maxlen = Length(aDist[0]); 
    indMax0 = 0;
    for(unsigned int i=1;i<aRI.size()*9;i++){
      double len = Length(aDist[i]);
      if( len > maxlen ){
        indMax0 = i;
        maxlen = len;
        distMax0 = aDist[i];
      }
    }      
  }
  unsigned int irbMax0 = indMax0/9;
  unsigned int ilocMax0 = indMax0-irbMax0*9;
  Com::CVector3D posMax0;
  {
    const CRigid3D_Ini& ri = aRI[irbMax0];
    const Com::CMatrix3& q = ri.GetRotMatrix_Ini();
    const Com::CVector3D l(ri.xlen,ri.ylen,ri.zlen);
    const Com::CVector3D& d = ri.GetIniPosCG();
    const Com::CMatrix3 B(aJntCand[ilocMax0][0]*0.5,aJntCand[ilocMax0][1]*0.5,0);
    const Com::CVector3D& p0 = B.MatVec(l);      
    posMax0 = q.MatVec(p0) + d;
  }
  std::cout << indMax0 << " " << irbMax0 << " " << ilocMax0 << "   " << posMax0 << std::endl;  
  for(unsigned int jri=0;jri<aRI.size();jri++){
    if( jri == irbMax0 ) continue;
    unsigned int ilocMax1 = 0;
    double dotMax1 = Com::Dot(aDist[jri*9+0],distMax0);
    for(unsigned int j=0;j<9;j++){
      double dot = Com::Dot(aDist[jri*9+j],distMax0);
      if( dot < dotMax1 ){
        dotMax1 = dot;
        ilocMax1 = j;
      }
    }
    Com::CVector3D posMax1;
    {
      const CRigid3D_Ini& ri = aRI[jri];
      const Com::CMatrix3& q = ri.GetRotMatrix_Ini();
      const Com::CVector3D l(ri.xlen,ri.ylen,ri.zlen);
      const Com::CVector3D& d = ri.GetIniPosCG();
      const Com::CMatrix3 B(aJntCand[ilocMax1][0]*0.5,aJntCand[ilocMax1][1]*0.5,0);
      const Com::CVector3D& p0 = B.MatVec(l);      
      posMax1 = q.MatVec(p0) + d;
    }
    std::cout << ilocMax1 << " " << dotMax1 << " " << posMax1 << std::endl;
    double lenx = Distance(posMax0, posMax1);
    double leny = 0.1;    
    double lenz = aRI[jri].zlen;
    double rho = aRI[jri].GetMass()/(aRI[jri].xlen*aRI[jri].ylen+aRI[jri].zlen);
    Com::CMatrix3 rot(1.0);
    {
      Com::CVector3D a(1,0,0);
      Com::CVector3D b = (posMax0-posMax1);
      b.SetNormalizedVector();
      Com::CQuaternion q(a,b);
      double m[9];
      q.RotMatrix33(m);
      for(unsigned int i=0;i<9;i++){ rot.mat[i] = m[i]; }
    }
    Com::CVector3D cg = (posMax0+posMax1)*0.5;
    jrb.AddRigidBody_Box(lenx,leny,lenz,rho, cg, rot);
  }
}

class CJointConnection{
public:
  CJointConnection(unsigned int idir0,
                   unsigned int irb1,unsigned int idir1, 
                   unsigned int ifix){
    this->idir0 = idir0;
    this->irb1 = irb1;
    this->idir1 = idir1;
    this->ifix = ifix;
  }
  CJointConnection( const CJointConnection& jc ){
    idir0 = jc.idir0;
    irb1 = jc.irb1;    
    idir1 = jc.idir1;
    ifix = jc.ifix;
  }
  CJointConnection(){
    idir0 = 0;
    irb1 = 0;    
    idir1 = 0;
    ifix = 0;    
  }
public:
  unsigned int idir0;  
  unsigned int irb1; 
  unsigned int idir1;
  unsigned int ifix;
};

bool ConnectPlank(unsigned int irb1, unsigned int irb2,
                  CJointedRigidBody& jrb)
{  
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };      
  const double aJntCand[9][2] = {
    {+0.0,-0.0}, {-0.9,-0.0}, {+0.9,-0.0},    
    {+0.0,-0.7}, {+0.0,+0.7}, {-0.9,-0.7}, 
    {+0.9,+0.7}, {-0.9,+0.7}, {+0.9,-0.7},           
  };    
  const CRigid3D_Ini& ri1 = jrb.aRI[irb1];
  Com::CMatrix3 Q1 = ri1.GetRotMatrix_Ini();
  Com::CVector3D n1 = ri1.GetRotMatrix_Ini().MatVec( Com::CVector3D(0,0,1) );  
//  unsigned int irb2 = jc2.irb1;
  const CRigid3D_Ini& ri2 = jrb.aRI[irb2];                
  Com::CVector3D n2 = ri2.GetRotMatrix_Ini().MatVec( Com::CVector3D(0,0,1) );        
  if( fabs(Dot(n1,n2)) < 0.5 ) return false;        
  //        std::cout << irb << " " << ijd << " " << jjd << "  " << irb1 << " " << irb2 << " " << Dot(n1,n2) << std::endl;
  Com::CVector3D y = Cross(n1,n2);
  if( SquareLength(y) < 1.0e-5 ){
    y = Cross(n1,Com::CVector3D(0,0,1));          
  }
  y.SetNormalizedVector();        
  Com::CVector3D x = ri2.GetIniPosCG()-ri1.GetIniPosCG();
  if( fabs(x.x) > fabs(x.y) && fabs(x.x) > fabs(x.z) ){ x = (x.x>0) ? Com::CVector3D(+1,0,0) : Com::CVector3D(-1,0,0); }
  if( fabs(x.y) > fabs(x.z) && fabs(x.y) > fabs(x.x) ){ x = (x.y>0) ? Com::CVector3D(0,+1,0) : Com::CVector3D(0,-1,0); }
  if( fabs(x.z) > fabs(x.x) && fabs(x.z) > fabs(x.y) ){ x = (x.z>0) ? Com::CVector3D(0,0,+1) : Com::CVector3D(0,0,-1); }
  x -= Dot(x,y)*y;
  x.SetNormalizedVector();
  Com::CVector3D z = Cross(x,y);
  //        std::cout << "vector " << "  " << x << "  " << y << "  " << z << std::endl;
  Com::CVector3D hit_pos;
  unsigned int iface;
  for(unsigned int icnd=0;icnd<9;icnd++){
    double rx1 = aJntCand[icnd][0];
    double ry1 = aJntCand[icnd][1];
    Com::CVector3D src = ri1.GetIniPosCG() + Q1.MatVec(Com::CVector3D(rx1*ri1.xlen*0.5,ry1*ri1.ylen*0.5,0));
    bool res = IsRayCollide_Ini(src,x,hit_pos,iface,ri2);
    if( !res ){ continue; }
    bool is_intsec = false;
    {            
      double min_depth = 0;
      int ir_min = -1;
      for(unsigned int ir=0;ir<jrb.aRI.size();ir++){
        if( ir == irb1 ) continue;
        bool res = IsRayCollide_Ini(src,x,hit_pos,iface,jrb.aRI[ir]);
        if( !res ) continue;
        double depth = Com::Dot(hit_pos-src,x);
//        std::cout << "depth " << depth << std::endl;
        if( depth < 0 ) continue;
        if( ir_min == -1 ){ 
          min_depth = depth; 
          ir_min = ir; 
          continue;
        }
        if( depth < min_depth ){
          min_depth = depth;
          ir_min = ir;
        }
        //              std::cout << ir_min << " " << irb1 << " " << irb2 << " " << min_depth << std::endl;
      }
      if( ir_min != irb2 || ir_min == -1 ){ is_intsec = true; }
    }
    if( is_intsec ){
      //            std::cout << "intersection " << std::endl;
      continue;
    }
    //          std::cout << "addplank " << std::endl;
    unsigned int irb3 = jrb.ConnectPlanks(irb1,irb2,src,hit_pos);
    for(unsigned int i=2;i<20;i++){
      CRigid3D_Ini& ri3 = jrb.aRI[irb3];
      ri3.ylen = i*ri3.zlen*2;
//      std::cout << "planks size" << ri3.xlen << " " << ri3.ylen << " " << ri3.zlen << std::endl;
      const Com::CMatrix3& Q3 = ri3.GetRotMatrix_Ini();
      const Com::CVector3D& srcr = ri3.GetIniPosCG() + Q3.MatVec(Com::CVector3D(0,+ri3.ylen*0.5,0));
      const Com::CVector3D& srcl = ri3.GetIniPosCG() + Q3.MatVec(Com::CVector3D(0,-ri3.ylen*0.5,0));            
      bool res0 = IsRayCollide_Ini(srcr,x,hit_pos,iface,jrb.aRI[irb1]);
      bool res1 = IsRayCollide_Ini(srcl,x,hit_pos,iface,jrb.aRI[irb2]);
      bool res2 = IsRayCollide_Ini(srcr,x,hit_pos,iface,jrb.aRI[irb1]);
      bool res3 = IsRayCollide_Ini(srcl,x,hit_pos,iface,jrb.aRI[irb2]);
      if( !res0 || !res1 || !res2 || !res3 ){ break; }            
      //            std::cout << res0 << " " << res1 << " " << res2 << " " << res3 << std::endl;
      jrb.PlaceNails();
      SolveStaticEquilibrium(jrb.aRI,jrb.aRD,jrb.aFix,jrb.aJFixForce,jrb.aExForce,
                             jrb.gravity,
                             jrb.contProp,
                             jrb.aFP);   
      bool is_stable = true;
      for(unsigned int ifix=0;ifix<jrb.aFix.size();ifix++){
        double shear, pull;
        DisplayJointCapacity(ifix, shear, pull,
                             jrb.aRI, jrb.aRD, 
                             jrb.aFix, 
                             jrb.aJFixForce, 
                             jrb.aFixDir, 
                             jrb.aNailJoint);        
        if( shear > 1 ){ is_stable = false; break; }
        if( pull  > 1 ){ is_stable = false; break; }
        //      std::cout << ifix << "   shear : " << shear*100 << "    pull : " << pull*100 << std::endl;              
      }  // ifix;
      if( is_stable ){ return true; }
    } // ilen
    jrb.DeleteRigidBody(irb3);
  } // icnd  
  return false;
}

void MakeSuggestionSupportMaterial2(const CJointedRigidBody& jrb, 
                                    std::vector<CJointedRigidBody>& aSgg)
{
  std::cout << "adding support material" << std::endl;
  std::vector<bool> aIsStableJoint;
  const unsigned int nFix = jrb.aFix.size();
  aIsStableJoint.resize(nFix,true);
  for(unsigned int ifix=0;ifix<nFix;ifix++){    
    double shear,pull;
    DisplayJointCapacity(ifix,shear,pull,
                         jrb.aRI,jrb.aRD,
                         jrb.aFix,jrb.aJFixForce,jrb.aFixDir,jrb.aNailJoint);
    if( shear > 1 || pull > 1 ){
      aIsStableJoint[ifix] = false;
    }    
  }
  const unsigned int nRI = jrb.aRI.size();      
  std::vector< std::vector<CJointConnection> > aJntDir;
  aJntDir.resize(nRI);
  for(unsigned int ifix=0;ifix<nFix;ifix++){
    unsigned int irb0 = jrb.aFix[ifix].irb0;
    unsigned int irb1 = jrb.aFix[ifix].irb1;
    unsigned int idir0 = jrb.aFixDir[ifix].first;      
    unsigned int idir1 = jrb.aFixDir[ifix].second;  
    aJntDir[irb0].push_back( CJointConnection(idir0,irb1,idir1,ifix) );
    aJntDir[irb1].push_back( CJointConnection(idir1,irb0,idir0,ifix) );
  }
  std::vector< std::pair<unsigned int,unsigned int> > aPairConnect;
  for(unsigned int irb=0;irb<nRI;irb++){
    unsigned int njd = aJntDir[irb].size();
    if( njd < 2 ) continue;
    for(unsigned int ijd=0;ijd<njd;ijd++){
      const CJointConnection& jc1 = aJntDir[irb][ijd];
      unsigned int ifix1 = jc1.ifix;      
      unsigned int irb1 = jc1.irb1;      
      for(unsigned int jjd=ijd+1;jjd<njd;jjd++){
        const CJointConnection& jc2 = aJntDir[irb][jjd]; 
        unsigned int ifix2 = jc2.ifix;
        unsigned int irb2 = jc2.irb1;
        if( aIsStableJoint[ifix1] && aIsStableJoint[ifix2] ){ continue; }        
        {
          bool is_searched = false;
          for(unsigned int ip=0;ip<aPairConnect.size();ip++){
            if( aPairConnect[ip].first == irb1 && aPairConnect[ip].second == irb2 ){ is_searched=true; break; }
            if( aPairConnect[ip].first == irb2 && aPairConnect[ip].second == irb1 ){ is_searched=true; break; }
          }
          if( is_searched ) continue;
        }
        aPairConnect.push_back( std::make_pair(irb1,irb2) );
        std::cout << "try conect " << irb1 << " " << irb2 << std::endl;            
        CJointedRigidBody jrb1 = jrb;
        bool res = ConnectPlank(irb1, irb2, jrb1);
        if( res ){
          aSgg.push_back(jrb1);
        }
      } // jjd
    } // ijd
  }
}



