/*
 *  joint_connectivity.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 9/25/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#include <vector>
#include <stdlib.h>

#include "delfem/vector3d.h"

#include "delfem/indexed_array.h"
#include "delfem/matvec/matdia_blkcrs.h"
#include "delfem/matvec/vector_blk.h"
#include "delfem/matvec/matdiafrac_blkcrs.h"
#include "delfem/matvec/bcflag_blk.h"

#include "rigidbody.h"
#include "furniture_utility.h"
#include "joint_fix.h"
#include "joint_connectivity.h"
#include "rigid_contact.h"

void EnforceJointConnectivity_ChangeBoxPosRotLen
(std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CJoint_Fix>& aFix, 
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir )
{  
  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensCE_BoxPosRot;
  std::vector<Com::CVector3D> aSensCE_BoxLen;  
  std::vector<Com::CVector3D> aSensCE_JntPos;  
  GetSensitivityJCE_InitialConfigBox(aSensCE_BoxPosRot,aSensCE_BoxLen,aSensCE_JntPos,
                                     aRI,aFix,aFixDir);
  
  double eps = -0.2;
  for(unsigned int irb=0;irb<aRI.size();irb++){ // BOX_POS
    CRigid3D_Ini& ri1 = aRI[irb];    
    Com::CVector3D e_s = aSensCE_BoxPosRot[irb].first * eps;
    ri1.SetIniPosCG( ri1.GetIniPosCG()+e_s );
  } 
  for(unsigned int irb=0;irb<aRI.size();irb++){ // BOX_ROT
    CRigid3D_Ini& ri1 = aRI[irb];    
    Com::CVector3D e_s = aSensCE_BoxPosRot[irb].second * eps;
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e_s);
    ri1.AddRotation_IniLeft(rot);  
  }
  for(unsigned int irb=0;irb<aRI.size();irb++){ // BOX_LEN
    CRigid3D_Ini& ri1 = aRI[irb];    
    Com::CVector3D e_s = aSensCE_BoxLen[irb] * eps;
    double xlen=ri1.xlen; 
    double ylen=ri1.ylen;
    double zlen=ri1.zlen;    
    const double rho = ri1.GetMass()/(xlen*ylen*zlen);    
    xlen += e_s.x;
    ylen += e_s.y;
    //    zlen += e_s.z;        
    ri1.SetModeBox(xlen,ylen,zlen,rho);
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){ // JNT_POS
    Com::CVector3D e_s = aSensCE_JntPos[ifix] * eps;
    CJoint_Fix& cjs = aFix[ifix];
    cjs.ini_pos_joint += e_s;
  }
  
//  double eng1 = JointConnectivityEnergy(aRB,aFix,aFixDir);
  //  std::cout << "Connectivitiy Energy : " << eng1 << std::endl;
}

const Com::CVector3D aFaceNormal[6] = {
  Com::CVector3D(+1, 0, 0),
  Com::CVector3D(-1, 0, 0),
  Com::CVector3D(+0,+1, 0),
  Com::CVector3D(-0,-1, 0),
  Com::CVector3D(+0, 0,+1),
  Com::CVector3D(-0, 0,-1) };

const Com::CVector3D aFaceNormalAbs[6] = {
  Com::CVector3D(+1, 0, 0),
  Com::CVector3D(+1, 0, 0),
  Com::CVector3D(+0,+1, 0),
  Com::CVector3D(-0,+1, 0),
  Com::CVector3D(+0, 0,+1),
  Com::CVector3D(-0, 0,+1) };

void AddSensJointOnFace
(std::pair<Com::CVector3D,Com::CVector3D>& Sens_BoxPosRot,
 Com::CVector3D& Sens_BoxLen,
 Com::CVector3D& Sens_JntPos,
 ////
 unsigned int idir0,
 const Com::CVector3D& ini_pos_joint,
 const Com::CVector3D& ini_pos_cg,             
 const Com::CMatrix3& rotmat_ini,                        
 double lx, double ly, double lz)
{  
  Com::CVector3D n0 = rotmat_ini.MatVec(aFaceNormal[idir0]);
  Com::CVector3D d0 = aFaceNormal[idir0];
  d0.x *= lx*0.5;
  d0.y *= ly*0.5;
  d0.z *= lz*0.5;  
  Com::CVector3D b0 = rotmat_ini.MatVec(d0)+ini_pos_cg;
  
  Com::CVector3D e0 = ini_pos_joint - b0;      
  Com::CVector3D f0 = ini_pos_joint - ini_pos_cg;
  double A0 = Com::Dot(e0,n0);
  Sens_JntPos += A0*n0;       
  Sens_BoxPosRot.first -= +A0*n0;      
  Sens_BoxPosRot.second -= A0*Com::CMatrix3( f0 ).MatVec(n0);
  Com::CVector3D t0 = aFaceNormalAbs[idir0];
  Sens_BoxLen -= (0.5*A0)*Com::CVector3D(t0.x,t0.y,t0.z);        
}

void AddSensJointOnEdge
(std::pair<Com::CVector3D,Com::CVector3D>& Sens_BoxPosRot,
 Com::CVector3D& Sens_BoxLen,
 Com::CVector3D& Sens_JntPos,
 ////
 unsigned int idir0,
 const Com::CVector3D& ini_pos_joint,
 const Com::CVector3D& ini_pos_cg,
 const Com::CMatrix3& rotmat_ini,
 double lx, double ly, double lz)
{  
  Com::CVector3D n0 = rotmat_ini.MatVec(aFaceNormal[idir0]);   
  Com::CVector3D d0 = aFaceNormal[idir0];
  d0.x *= lx*0.5;
  d0.y *= ly*0.5;
  d0.z *= lz*0.5;  
  Com::CVector3D b0 = rotmat_ini.MatVec(d0)+ini_pos_cg;
  
  Com::CVector3D e0 = ini_pos_joint - b0;      
  Com::CVector3D f0 = ini_pos_joint - ini_pos_cg;
  Sens_JntPos += e0;
  Sens_BoxPosRot.first += -e0;
  Com::CVector3D c0 = rotmat_ini.MatVec(d0);  
  Sens_BoxPosRot.second -= Com::CMatrix3( -f0 ).MatVec(c0);
  Com::CVector3D t0 = aFaceNormalAbs[idir0];
  Sens_BoxLen += 0.25*Com::CVector3D(t0.x*lx,t0.y*ly,t0.z*lz);
  Com::CVector3D s0 = rotmat_ini.MatVecTrans(f0);        
  s0.x *= aFaceNormal[idir0].x*0.5;
  s0.y *= aFaceNormal[idir0].y*0.5;
  s0.z *= aFaceNormal[idir0].z*0.5;
  Sens_BoxLen -= s0;
}



double GetSensitivityJCE_InitialConfigBox
(std::vector<std::pair<Com::CVector3D,Com::CVector3D> >& aSens_BoxPosRot,
 std::vector<Com::CVector3D>& aSens_BoxLen,
 std::vector<Com::CVector3D>& aSens_JntPos,
 ////
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CJoint_Fix>& aFix, 
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir)
{
  for(unsigned int irb=0;irb<aRI.size();irb++){
    aSens_BoxPosRot.push_back( std::make_pair(Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) );
    aSens_BoxLen.push_back( Com::CVector3D(0,0,0) );    
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    aSens_JntPos.push_back( Com::CVector3D(0,0,0) );    
  }
  double jeng = 0;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cjs = aFix[ifix];
    const unsigned int irb0 = cjs.irb0;
    const CRigid3D_Ini& ri0 = aRI[irb0]; 
    
    const unsigned int irb1 = cjs.irb1;
//    const CRigidBody3D& rb1 = aRB[irb1];
    const CRigid3D_Ini& ri1 = aRI[irb1];
    
    const unsigned int idir0 = aFixDir[ifix].first;
    const unsigned int idir1 = aFixDir[ifix].second;
    
    if( idir0==4 || idir0==5 ){      
      AddSensJointOnFace(aSens_BoxPosRot[irb0],aSens_BoxLen[irb0],aSens_JntPos[ifix],
                         idir0,cjs.ini_pos_joint,
                         ri0.GetIniPosCG(),ri0.GetRotMatrix_Ini(),ri0.xlen,ri0.ylen,ri0.zlen);
    }
    else {
      AddSensJointOnEdge(aSens_BoxPosRot[irb0],aSens_BoxLen[irb0],aSens_JntPos[ifix],
                         idir0,cjs.ini_pos_joint,
                         ri0.GetIniPosCG(),ri0.GetRotMatrix_Ini(),ri0.xlen,ri0.ylen,ri0.zlen);
    }
    ////
    if( idir1==4||idir1==5 ){      
      AddSensJointOnFace(aSens_BoxPosRot[irb1],aSens_BoxLen[irb1],aSens_JntPos[ifix],
                         idir1,cjs.ini_pos_joint,
                         ri1.GetIniPosCG(),ri1.GetRotMatrix_Ini(),ri1.xlen,ri1.ylen,ri1.zlen);
    }
    else{      
      AddSensJointOnEdge(aSens_BoxPosRot[irb1],aSens_BoxLen[irb1],aSens_JntPos[ifix],
                         idir1,cjs.ini_pos_joint,
                         ri1.GetIniPosCG(),ri1.GetRotMatrix_Ini(),ri1.xlen,ri1.ylen,ri1.zlen);
    }    
  }
  return jeng;  
}


double JointConnectivityEnergy
(const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CJoint_Fix>& aFix, 
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir)
{
  double jeng = 0;
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cjs = aFix[ifix];    
    const unsigned int irb0 = cjs.irb0;
    const CRigid3D_Ini& ri0 = aRI[irb0];
    
    const unsigned int irb1 = cjs.irb1;
    const CRigid3D_Ini& ri1 = aRI[irb1];      
    
    const unsigned int idir0 = aFixDir[ifix].first;
    const unsigned int idir1 = aFixDir[ifix].second;
    
    Com::CVector3D n0,b0;
    {    
      n0 = ri0.GetRotMatrix_Ini().MatVec(aFaceNormal[idir0]);
      b0 = aFaceNormal[idir0];
      b0.x *= ri0.xlen*0.5;
      b0.y *= ri0.ylen*0.5;
      b0.z *= ri0.zlen*0.5;
      b0 = ri0.GetRotMatrix_Ini().MatVec(b0)+ri0.GetIniPosCG();
    }  
    
    Com::CVector3D n1,b1;
    {    
      n1 = ri1.GetRotMatrix_Ini().MatVec(aFaceNormal[idir1]);            
      b1 = aFaceNormal[idir1];
      b1.x *= ri1.xlen*0.5;
      b1.y *= ri1.ylen*0.5;
      b1.z *= ri1.zlen*0.5;
      b1 = ri1.GetRotMatrix_Ini().MatVec(b1)+ri1.GetIniPosCG();
    }
    
    if( idir0==4 || idir0==5 ){      
      double A = Com::Dot(cjs.ini_pos_joint - b0,n0);
      jeng += 0.5*A*A;
    }
    else{
      double B = SquareDistance(cjs.ini_pos_joint,b0);
      jeng += 0.5*B;
    }
    ////
    if( idir1==4 || idir1==5 ){          
      double A = Com::Dot(cjs.ini_pos_joint - b1,n1);      
      jeng += 0.5*A*A;      
    }
    else{
      double B = SquareDistance(cjs.ini_pos_joint,b1);
      jeng += 0.5*B;      
    }
  }
  return jeng;
}

static double rand2(double min, double max){
  return rand()/(RAND_MAX+1.0)*(max-min)+min;
}

void TestSensitive_JointConnectivityEnergy()
{      
  std::vector<CRigidBody3D> aRB;
  std::vector<CJoint_Fix> aFix;
  std::vector< std::pair<unsigned int, unsigned int> > aFixDir;
  
  ///////  
  unsigned int nRB = 10;
  aRB.resize(nRB);
  for(unsigned int irb=0;irb<nRB;irb++){
    aRB[irb].ri.SetModeBox( rand2(0.1,10.0), rand2(0.1,10.0), rand2(0.1,10.0), 1 );     
    aRB[irb].ri.SetIniPosCG( Com::CVector3D(rand2(-1,1), rand2(-1,1), rand2(-1,1)) );
    Com::CMatrix3 m; m.SetRotMatrix_BryantAngle(rand2(-2,2),rand2(-2,2),rand2(-2,2));
    aRB[irb].ri.SetIniRotation(m);   
  }        
  double invstiff_pos = 1.0e-4;
  double invstiff_rot = 1.0e-4;        
  for(unsigned int ifix=0;ifix<nRB-1;ifix++){
    aFixDir.push_back( std::make_pair(0,5) );      
    CJoint_Fix fix(ifix,ifix+1); 
    fix.SetIniPosJoint(rand2(-1,1), rand2(-1,1), rand2(-1,1));
    fix.SetIniJointDirection(1,0,0,  0,1,0,  0,0,1);
    fix.SetInvStiffJoint(invstiff_pos, invstiff_rot);            
    aFix.push_back( fix );                
  }
  
  
  
  ///////
  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensCE_BoxPosRot;
  std::vector<Com::CVector3D> aSensCE_BoxLen;  
  std::vector<Com::CVector3D> aSensCE_JntPos;  
  double eng0;
  {
    std::vector<CRigid3D_Ini> aRI;
    for(unsigned int irb=0;irb<aRB.size();irb++){ aRI.push_back( aRB[irb].ri ); }    
    GetSensitivityJCE_InitialConfigBox(aSensCE_BoxPosRot,aSensCE_BoxLen,aSensCE_JntPos,
                                       aRI,aFix,aFixDir);      
    eng0 = JointConnectivityEnergy(aRI,aFix,aFixDir);
  }
  
  double eps = 1.0e-8;
  double diff = 0;
  for(unsigned int irb=0;irb<aRB.size();irb++){ // BOX_POS
    CRigidBody3D& rb1 = aRB[irb];    
    Com::CVector3D e_s(rand2(-1,1),rand2(-1,1),rand2(-1,1)); e_s *= eps;
    rb1.ri.SetIniPosCG( rb1.ri.GetIniPosCG()+e_s );
    diff += Dot(aSensCE_BoxPosRot[irb].first,e_s);    
  } 
  for(unsigned int irb=0;irb<aRB.size();irb++){ // BOX_ROT
    CRigidBody3D& rb1 = aRB[irb];    
    Com::CVector3D e_s(rand2(-1,1),rand2(-1,1),rand2(-1,1)); e_s *= eps;    
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e_s);
    rb1.ri.AddRotation_IniLeft(rot);  
    diff += Dot(aSensCE_BoxPosRot[irb].second,e_s);
  }
  for(unsigned int irb=0;irb<aRB.size();irb++){ // BOX_LEN
    CRigidBody3D& rb1 = aRB[irb];
    const CRigid3D_Ini& ri1 = rb1.ri;    
    Com::CVector3D e_s(rand2(-1,1),rand2(-1,1),rand2(-1,1)); e_s *= eps;        
    double xlen=ri1.xlen; 
    double ylen=ri1.ylen;
    double zlen=ri1.zlen;    
    const double rho = rb1.ri.GetMass()/(xlen*ylen*zlen);    
    xlen += e_s.x;
    ylen += e_s.y;
    zlen += e_s.z;        
    rb1.ri.SetModeBox(xlen,ylen,zlen,rho);
    diff += Dot(aSensCE_BoxLen[irb],e_s);
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){ // JNT_POS
    Com::CVector3D e_s(rand2(-1,1),rand2(-1,1),rand2(-1,1)); e_s *= eps;
    CJoint_Fix& cjs = aFix[ifix];
    cjs.ini_pos_joint += e_s;
    diff += Dot(aSensCE_JntPos[ifix],e_s);
  }
  
  double eng1;
  {  
    std::vector<CRigid3D_Ini> aRI;
    for(unsigned int irb=0;irb<aRB.size();irb++){ aRI.push_back( aRB[irb].ri ); }    
    eng1 = JointConnectivityEnergy(aRI,aFix,aFixDir);
  }
  
  std::cout << "TestSensitivity_JointConnectivityEnergy" << std::endl;
  std::cout << (eng1-eng0)/eps << " " << diff/eps << std::endl;
}

bool GetIntersectionOfTwoPlanks(Com::CVector3D& a, Com::CVector3D& b,
                                const CRigid3D_Ini& ri0, const CRigid3D_Ini& ri1,
                                unsigned int idir0, unsigned int idir1)
{  
  const Com::CVector3D aFaceNormal[6][3] = {
    { Com::CVector3D(+1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0,+1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-0,-1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0, 0,+1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) },
    { Com::CVector3D(-0, 0,-1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) } };    
    
  Com::CMatrix3 R0 = ri0.GetRotMatrix_Ini();
  Com::CVector3D n0 = R0.MatVec(aFaceNormal[idir0][0]);
  Com::CMatrix3 B0(ri0.xlen*0.5, ri0.ylen*0.5, ri0.zlen*0.5);
  Com::CMatrix3 RL0 = R0.MatMat(B0);
  Com::CVector3D b0 = RL0.MatVec(aFaceNormal[idir0][0])+ri0.GetIniPosCG();
  /////
  Com::CMatrix3 R1 = ri1.GetRotMatrix_Ini();
  Com::CVector3D n1 = R1.MatVec(aFaceNormal[idir1][0]);
  Com::CMatrix3 L1(ri1.xlen*0.5, ri1.ylen*0.5, ri1.zlen*0.5);
  Com::CMatrix3 RL1 = R1.MatMat(L1);  
  Com::CVector3D b1 = RL1.MatVec(aFaceNormal[idir1][0])+ri1.GetIniPosCG();
  Com::CVector3D h1 = RL1.MatVec(aFaceNormal[idir1][1]);  
  /////
  const Com::CVector3D& h0 = RL0.MatVec( aFaceNormal[idir0][1]);
  const Com::CVector3D& v0 = RL0.MatVec( aFaceNormal[idir0][2]);
  const double dlen1 = Com::Dot(b0-b1,n0) / Com::Dot(n1,n0);      
  double rha0, rva0;
  bool resa0 = IsRayCollideQuad(b1+dlen1*n1+h1,n0, b0,n0,h0,v0, rha0,rva0); 
  double rhb0, rvb0;
  bool resb0 = IsRayCollideQuad(b1+dlen1*n1-h1,n0, b0,n0,h0,v0, rhb0,rvb0); 
  double rha1=rha0, rva1=rva0,  rhb1=rhb0,rvb1=rvb0;
  if( !resa0 ){
    double t0=2, t1=2, t2=2, t3=2;
    if( fabs(rvb1-rva1) > 1.0-5 ){ 
      t0 = (+1-rva0)/(rvb0-rva0);
      t1 = (-1-rva0)/(rvb0-rva0);
    }
    if( fabs(rhb0-rha0) > 1.0e-5 ){
      t2 = (+1-rha0)/(rhb0-rha0);
      t3 = (-1-rha0)/(rhb0-rha0);
    }
    double tmin = 1-1.0e-30;
    if( t0 < tmin && t0 > 0 ){ double s=(1-t0)*rha0+t0*rhb0; if(s>-1&&s<+1){ tmin = t0; } }
    if( t1 < tmin && t1 > 0 ){ double s=(1-t1)*rha0+t1*rhb0; if(s>-1&&s<+1){ tmin = t1; } }
    if( t2 < tmin && t2 > 0 ){ double s=(1-t2)*rva0+t2*rvb0; if(s>-1&&s<+1){ tmin = t2; } }
    if( t3 < tmin && t3 > 0 ){ double s=(1-t3)*rva0+t3*rvb0; if(s>-1&&s<+1){ tmin = t3; } }
    if( tmin > 0 && tmin < 1 ){
      resa0 = true;
      rva1 = (1-tmin)*rva0+tmin*rvb0;
      rha1 = (1-tmin)*rha0+tmin*rhb0;
    }
  }
  if( !resb0 ){
    double t0=2, t1=2, t2=2, t3=2;
    if( fabs(rvb0-rva0) > 1.0-5 ){ 
      t0 = (+1-rvb0)/(rva0-rvb0);
      t1 = (-1-rvb0)/(rva0-rvb0);
    }
    if( fabs(rhb0-rha0) > 1.0e-5 ){
      t2 = (+1-rhb0)/(rha0-rhb0);
      t3 = (-1-rhb0)/(rha0-rhb0);
    }
    double tmin = 1-1.0e-30;
    if( t0 < tmin && t0 > 0 ){ double s=(1-t0)*rhb0+t0*rha0; if(s>-1&&s<+1){ tmin = t0; } }
    if( t1 < tmin && t1 > 0 ){ double s=(1-t1)*rhb0+t1*rha0; if(s>-1&&s<+1){ tmin = t1; } }        
    if( t2 < tmin && t2 > 0 ){ double s=(1-t2)*rvb0+t2*rva0; if(s>-1&&s<+1){ tmin = t2; } }
    if( t3 < tmin && t3 > 0 ){ double s=(1-t3)*rvb0+t3*rva0; if(s>-1&&s<+1){ tmin = t3; } }
    if( tmin > 0 && tmin < 1 ){
      resb0 = true;
      rvb1 = (1-tmin)*rvb0+tmin*rva0;
      rhb1 = (1-tmin)*rhb0+tmin*rha0;
    }
  }            
  if( !resa0 || !resb0 ) return false;
  a = rva1*v0+rha1*h0+b0;
  b = rvb1*v0+rhb1*h0+b0;  
  return true;
}

void EnforceJointConnectivity_ChangeBoxLen
(std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CJoint_Fix>& aFix, 
 const std::vector< std::pair<unsigned int, unsigned int> >& aFixDir,
 bool is_chg_len0)
{
  const Com::CVector3D aFaceNormal[6][3] = {
    { Com::CVector3D(+1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0,+1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-0,-1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0, 0,+1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) },
    { Com::CVector3D(-0, 0,-1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) } };    
    
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    CJoint_Fix& cjs = aFix[ifix];    
    const unsigned int irb0 = cjs.irb0;
    CRigid3D_Ini& ri0 = aRI[irb0];    
    const double rho0 = ri0.GetMass()/(ri0.xlen * ri0.ylen * ri0.zlen);      
    unsigned int idir0 = aFixDir[ifix].first;    
    Com::CMatrix3 R0 = ri0.GetRotMatrix_Ini();
    Com::CVector3D n0 = R0.MatVec(aFaceNormal[idir0][0]);
    Com::CMatrix3 B0(ri0.xlen*0.5, ri0.ylen*0.5, ri0.zlen*0.5);
    Com::CMatrix3 RL0 = R0.MatMat(B0);
    Com::CVector3D b0 = RL0.MatVec(aFaceNormal[idir0][0])+ri0.GetIniPosCG();
    
    const unsigned int irb1 = cjs.irb1;
    CRigid3D_Ini& ri1 = aRI[irb1];  
    const double rho1 = ri1.GetMass()/(ri1.xlen * ri1.ylen * ri1.zlen);    
    unsigned int idir1 = aFixDir[ifix].second;    
    Com::CMatrix3 R1 = ri1.GetRotMatrix_Ini();
    Com::CVector3D n1 = R1.MatVec(aFaceNormal[idir1][0]);
    Com::CMatrix3 L1(ri1.xlen*0.5, ri1.ylen*0.5, ri1.zlen*0.5);
    Com::CMatrix3 RL1 = R1.MatMat(L1);
    Com::CVector3D b1 = RL1.MatVec(aFaceNormal[idir1][0])+ri1.GetIniPosCG();
    Com::CVector3D h1 = RL1.MatVec(aFaceNormal[idir1][1]);
    
    assert( idir0==4 || idir0==5 );
    if( idir1!=4 && idir1!=5 ){
      const double dlen1 = Com::Dot(b0-b1,n0) / Com::Dot(n1,n0);      
      { // set joint position
        Com::CVector3D a, b;
        bool res = GetIntersectionOfTwoPlanks(a,b,ri0,ri1,idir0,idir1);
        if( res ){
          cjs.ini_pos_joint = (a+b)*0.5;
        }
        else{
          cjs.ini_pos_joint = b1+dlen1*n1;
        }
      }
      // set Plank 1 
      if( idir1==0 || idir1==1 ){ 
        ri1.SetModeBox(ri1.xlen+dlen1, ri1.ylen,ri1.zlen,rho1); 
        ri1.SetIniPosCG( ri1.GetIniPosCG() + dlen1*0.5*n1 );
      }
      if( idir1==2 || idir1==3 ){
        ri1.SetModeBox(ri1.xlen,ri1.ylen+dlen1, ri1.zlen,rho1);
        ri1.SetIniPosCG( ri1.GetIniPosCG() + dlen1*0.5*n1 );        
      }              
      if( is_chg_len0 ){ // set plank 0
        const Com::CVector3D& vec_h = RL0.MatVec( aFaceNormal[idir0][1]*0.5);
        const Com::CVector3D& vec_v = RL0.MatVec( aFaceNormal[idir0][2]*0.5);
        double rh, rv;
        bool res = IsRayCollideQuad(cjs.ini_pos_joint,n0, b0,n0,vec_h,vec_v, rv,rh); 
        if(      rh > rv && rh > -rv ){ // +y
        }
        else if( rh < rv && rh < -rv ){ // -y
        }
        else if( rv > rh && rv > -rh ){ // +x
          Com::CVector3D e = R0.MatVec( Com::CVector3D(1,0,0) );
          double len0 = Dot(cjs.ini_pos_joint-b0,e);
          if( len0 <= 0 ) continue;
          assert( len0 > 0 );
          double dlen0 = len0+ri1.zlen-ri0.xlen*0.5;          
          if( dlen0 > 0 ){
            ri0.SetModeBox(ri0.xlen+dlen0,ri0.ylen, ri0.zlen,rho0);
            ri0.SetIniPosCG( ri0.GetIniPosCG() + dlen0*0.5*e );                    
          }          
        }   
        else if( rv < rh && rv < -rh ){ // -x
          Com::CVector3D e = R0.MatVec( Com::CVector3D(-1,0,0) );
          double len0 = Dot(cjs.ini_pos_joint-b0,e);
          if( len0 <= 0 ) continue;          
          assert( len0 > 0 );
          double dlen0 = len0+ri1.zlen-ri0.xlen*0.5;          
          if( dlen0 > 0 ){
            ri0.SetModeBox(ri0.xlen+dlen0,ri0.ylen, ri0.zlen,rho0);
            ri0.SetIniPosCG( ri0.GetIniPosCG() + dlen0*0.5*e );                    
          }                    
        }        
      }
    }
  }
}

/*
 void MoveRotateRigidBodyTowardJoint
 (CJoint_Fix& jF,
 std::vector<CRigidBody3D>& aRB,
 unsigned int idir0, unsigned int idir1,
 double ratio_r, double ratio_t)
 {  
 const unsigned int irb0 = jF.aIndRB[0];
 const CRigidBody3D& rb0 = aRB[irb0];
 
 const unsigned int irb1 = jF.aIndRB[1];
 CRigidBody3D& rb1 = aRB[irb1];
 
 Com::CVector3D root1;
 {    
 if( idir1 == 0 ){ root1 = Com::CVector3D(+rb1.xlen,0,0); }
 if( idir1 == 1 ){ root1 = Com::CVector3D(-rb1.xlen,0,0); }
 if( idir1 == 2 ){ root1 = Com::CVector3D(0,+rb1.ylen,0); }
 if( idir1 == 3 ){ root1 = Com::CVector3D(0,-rb1.ylen,0); }
 if( idir1 == 4 ){ root1 = Com::CVector3D(0,0,+rb1.zlen); }
 if( idir1 == 5 ){ root1 = Com::CVector3D(0,0,-rb1.zlen); }    
 root1 *= 0.5;
 }
 Com::CVector3D cur_joint_pos = rb0.GetIniPosCG() + rb0.GetDispCG() + rb0.GetRotMatrix_Inc().MatVec(jF.ini_pos_joint-rb0.GetIniPosCG());
 Com::CVector3D Rot_root1 = rb1.GetRotMatrix().MatVec(root1);
 Com::CVector3D goal_pos_cg = (1-ratio_t)*rb1.GetIniPosCG() + ratio_t*(cur_joint_pos-Rot_root1-rb1.GetDispCG() );
 rb1.SetIniPosCG( goal_pos_cg );
 
 Com::CVector3D a0 = Rot_root1; a0.Normalize();
 Com::CVector3D a1 = cur_joint_pos-rb1.GetIniPosCG()-rb1.GetDispCG(); a1.Normalize();
 Com::CVector3D axis = Com::Cross(a0,a1); axis*=ratio_r;
 double axs[3] = {axis.x,axis.y,axis.z};
 Com::CMatrix3 m0; m0.SetRotMatrix_Cartesian(axs);
 Com::CMatrix3 m2 = rb1.GetRotMatrix_Inc().MatMatTrans( m0.MatMat( rb1.GetRotMatrix() ) );
 rb1.SetIniRotation(m2);
 }*/


