/*
 *  jointed_rigidbody_contact.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 10/2/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */


#include "delfem/indexed_array.h"
#include "delfem/matvec/matdia_blkcrs.h"
#include "delfem/matvec/matdiafrac_blkcrs.h"
#include "delfem/matvec/vector_blk.h"

#include "rigid_pointset.h"
#include "furniture_utility.h"
#include "joint_connectivity.h"
#include "rigid_contact.h"
#include "rigid_contact_ls.h"

#include "jointed_rigidbody_contact.h"

void GetResidual_JointedRigidBodyContactSystem
(
 CSimulationSensitivity& resSim,
 ////
 const std::vector<CRigid3D_Ini>& aRI,   
 const std::vector<CRigid3D_Def>& aRD,    
 const std::vector<CJoint_Fix>& aFix, 
 const std::vector<CJoint_Fix_Force>& aJFixForce,  
 const std::vector<CExternalForceOnRigidBody>& aExForce, 
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP)
{                
  resSim.aPlankPR.clear();
  for(unsigned int irb=0;irb<aRI.size();irb++){
    resSim.aPlankPR.push_back( std::make_pair(Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) );
  }
  resSim.aJointLM.clear();  
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    resSim.aJointLM.push_back( std::make_pair(Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) );    
  }  
  
  ////////////////////////////////////////////////////////////
  for(unsigned int irb=0;irb<aRI.size();irb++){
    resSim.aPlankPR[irb].first += aRI[irb].GetMass()*gravity;
//    const CRigidBody3D& rb = aRB[irb];
    const CRigid3D_Ini& ri = aRI[irb];    
    const CRigid3D_Def& rd = aRD[irb];
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);
      AddResidual_StaticContact(ri.GetIniPosCG(), ri.GetRotMatrix_Ini(),
                                rd.GetDispCG(),   rd.GetRotMatrix_Inc(),
                                v0,irb,resSim.aPlankPR,
                                contProp.cont_base,contProp.cont_norm,contProp.stiff_n,contProp.stiff_f,
                                aCP.aFP[i]);
    }
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cjs = aFix[ifix];    
    const CJoint_Fix_Force& jff = aJFixForce[ifix];        
    unsigned int irb0 = cjs.irb0; 
    unsigned int irb1 = cjs.irb1; 
    AddResidual_JointFix(resSim.aPlankPR,resSim.aJointLM,
                         aRI.size(),
                         ifix,cjs,jff,
                         aRI[irb0], aRD[irb0],
                         aRI[irb1], aRD[irb1]);
  }
  for(unsigned int ief=0;ief<aExForce.size();ief++){
    const CExternalForceOnRigidBody& ef = aExForce[ief];
    //    if( !ef.is_active ) continue;
    unsigned int irb = ef.irb;
    AddResidual_ExternalForce_Static(resSim.aPlankPR[irb].first, resSim.aPlankPR[irb].second, 
                                     aRI[irb], aRD[irb], 
                                     ef.pos, ef.force);
  }
}

void AddSensitivityResidual_JointedRigidBodyContactSystem
(std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aSensResPR,
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aSensResLM,
 PARAM_TYPE itype_s, int iobj_s, unsigned int idim_s, double alpha,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc,
 /////
 const std::vector<CRigid3D_Ini>& aRI,    
 const std::vector<CRigid3D_Def>& aRD,     
 const std::vector<CJoint_Fix>& aFix, 
 const std::vector<CJoint_Fix_Force>& aJFixForce,  
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 /////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP) 
{                
  Com::CVector3D e_s(0,0,0); e_s[idim_s] = alpha;
  if( itype_s == BOX_POS || itype_s == BOX_ROT || itype_s == BOX_LEN ){
    assert( iobj_s < aRI.size() );
    Com::CMatrix3 ResP(0.0), ResR(0.0);
    AddSensitivityResidual_StaticContact
    (iobj_s, aRI[iobj_s], aRD[iobj_s], 
     ResP, ResR,
     itype_s,iobj_s,
     contProp.cont_base,contProp.cont_norm,contProp.stiff_n,contProp.stiff_f,     
     aFP[iobj_s],
     itype_anchor_reloc);    
    aSensResPR[iobj_s].first  += ResP.MatVec(e_s);
    aSensResPR[iobj_s].second += ResR.MatVec(e_s);    
  }
  if( itype_s == BOX_LEN ){  // sensitivity of box mass gravitational force
    assert( iobj_s < aRI.size() );    
    const CRigid3D_Ini& ri = aRI[iobj_s];    
    double rho = ri.GetMass()/(ri.xlen*ri.ylen*ri.zlen);
    Com::CVector3D tmp = gravity*rho;
    if( idim_s == 0 ){ tmp *= ri.ylen*ri.zlen; }
    if( idim_s == 1 ){ tmp *= ri.xlen*ri.zlen; }
    if( idim_s == 2 ){ tmp *= ri.xlen*ri.ylen; }
    aSensResPR[iobj_s].first += tmp*alpha; 
  }  
  // sensitivity of residual from fix force 
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cjs = aFix[ifix];        
    const CJoint_Fix_Force& jff = aJFixForce[ifix];            
    unsigned int irb0 = cjs.irb0;
    unsigned int irb1 = cjs.irb1;
    if( itype_s == BOX_POS ){
      assert( iobj_s < aRI.size() );
      AddSensitivityResidual_StaticJoint(aSensResPR,aSensResLM,
                                         aRI.size(),
                                         ifix,cjs,jff,
                                         aRI[irb0], aRD[irb0],
                                         aRI[irb1], aRD[irb1],
                                         true,iobj_s,idim_s,alpha);
    }
    else if( itype_s == JNT_POS ){
      assert( iobj_s < aFix.size() );      
      AddSensitivityResidual_StaticJoint(aSensResPR,aSensResLM,
                                         aRI.size(),
                                         ifix,cjs,jff,
                                         aRI[irb0], aRD[irb0],
                                         aRI[irb1], aRD[irb1],
                                         false,iobj_s,idim_s,alpha);
    }
  }
  // sensitivity of residual from external force
  if( itype_s == BOX_ROT ){
    assert( iobj_s < aRI.size() );    
    for(unsigned int ief=0;ief<aExForce.size();ief++){
      const CExternalForceOnRigidBody& ef = aExForce[ief];
      if( ef.irb != iobj_s ) continue;
      unsigned int irb = ef.irb;
      const CRigid3D_Ini& ri = aRI[irb];          
      const CRigid3D_Def& rd = aRD[irb];          
      const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
      const Com::CMatrix3& P = rd.GetRotMatrix_Inc();    
      const Com::CVector3D& Ptf = P.MatVecTrans(ef.force);
      const Com::CVector3D& Qa = Q.MatVec(ef.pos);
      const Com::CMatrix3& wPtfwQa = Com::CMatrix3(Ptf).MatMat( Com::CMatrix3(Qa) );      
      aSensResPR[iobj_s].second += wPtfwQa.MatVec(e_s);
    }    
  }
}


void AddSensitivityResidual_JointedRigidBodyContactSystem
(std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aSensResPR,
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aSensResLM,
 PARAM_TYPE itype_s, int iobj_s, const Com::CVector3D& e_s,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc,
 /////
 const std::vector<CRigid3D_Ini>& aRI,   
 const std::vector<CRigid3D_Def>& aRD,     
 const std::vector<CJoint_Fix>& aFix, 
 const std::vector<CJoint_Fix_Force>& aJFixForce,  
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 /////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP) 
{                
  if( itype_s == BOX_POS || itype_s == BOX_ROT || itype_s == BOX_LEN ){
    assert( iobj_s < aRI.size() );
    Com::CMatrix3 ResP(0.0), ResR(0.0);
    AddSensitivityResidual_StaticContact
    (iobj_s, aRI[iobj_s], aRD[iobj_s], 
     ResP, ResR,
     itype_s,iobj_s,
     contProp.cont_base,contProp.cont_norm,contProp.stiff_n,contProp.stiff_f,     
     aFP[iobj_s],
     itype_anchor_reloc);    
    aSensResPR[iobj_s].first  += ResP.MatVec(e_s);
    aSensResPR[iobj_s].second += ResR.MatVec(e_s);    
  }
  if( itype_s == BOX_LEN ){  // sensitivity of box mass gravitational force
    assert( iobj_s < aRI.size() );    
    const CRigid3D_Ini& ri = aRI[iobj_s];    
    const double rho = ri.GetMass()/(ri.xlen*ri.ylen*ri.zlen);
    double tmp2 = ri.ylen*ri.zlen*e_s.x + ri.xlen*ri.zlen*e_s.y + ri.xlen*ri.ylen*e_s.z;
    aSensResPR[iobj_s].first += (rho*tmp2)*gravity; 
  }  
  // sensitivity of residual from fix force 
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cjs = aFix[ifix];        
    const CJoint_Fix_Force& jff = aJFixForce[ifix];            
    unsigned int irb0 = cjs.irb0;
    unsigned int irb1 = cjs.irb1;
    if( itype_s == BOX_POS ){
      assert( iobj_s < aRI.size() );
      AddSensitivityResidual_StaticJoint(aSensResPR,aSensResLM,
                                         aRI.size(),
                                         ifix,cjs,jff,
                                         aRI[irb0], aRD[irb0],
                                         aRI[irb1], aRD[irb1],
                                         true,iobj_s,e_s);
    }
    else if( itype_s == JNT_POS ){
      assert( iobj_s < aFix.size() );      
      AddSensitivityResidual_StaticJoint(aSensResPR,aSensResLM,
                                         aRI.size(),
                                         ifix,cjs,jff,
                                         aRI[irb0], aRD[irb0],
                                         aRI[irb1], aRD[irb1],
                                         false,iobj_s,e_s);
    }
  }
  // sensitivity of residual from external force
  if( itype_s == BOX_ROT ){
    assert( iobj_s < aRI.size() );    
    for(unsigned int ief=0;ief<aExForce.size();ief++){
      const CExternalForceOnRigidBody& ef = aExForce[ief];
      if( ef.irb != iobj_s ) continue;
      unsigned int irb = ef.irb;
      const CRigid3D_Ini& ri = aRI[irb];          
      const CRigid3D_Def& rd = aRD[irb];          
      const Com::CMatrix3& Q = ri.GetRotMatrix_Ini();
      const Com::CMatrix3& P = rd.GetRotMatrix_Inc();    
      const Com::CVector3D& Ptf = P.MatVecTrans(ef.force);
      const Com::CVector3D& Qa = Q.MatVec(ef.pos);
      const Com::CMatrix3& wPtfwQa = Com::CMatrix3(Ptf).MatMat( Com::CMatrix3(Qa) );      
      aSensResPR[iobj_s].second += wPtfwQa.MatVec(e_s);
    }    
  }
}


void AddSensitivityResidual_JointedRigidBodyContactSystem
(std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aSensResPR,
 std::vector< std::pair<Com::CVector3D, Com::CVector3D> >& aSensResLM,
 const CDesignSensitivity& desSens,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc,
 /////
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CRigid3D_Def>& aRD,  
 const std::vector<CJoint_Fix>& aFix, 
 const std::vector<CJoint_Fix_Force>& aJFixForce,  
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 /////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP)
{
  for(unsigned int irb=0;irb<aRI.size();irb++){
    AddSensitivityResidual_JointedRigidBodyContactSystem(aSensResPR,aSensResLM,
                                                         BOX_POS,irb,desSens.aPlank[irb].p,itype_anchor_reloc,                                                           
                                                         aRI,aRD,aFix,aJFixForce,aExForce,
                                                         gravity,
                                                         contProp,
                                                         aFP);
    AddSensitivityResidual_JointedRigidBodyContactSystem(aSensResPR,aSensResLM,
                                                         BOX_ROT,irb,desSens.aPlank[irb].r,itype_anchor_reloc,
                                                         aRI,aRD,aFix,aJFixForce,aExForce,
                                                         gravity,
                                                         contProp,
                                                         aFP);
    AddSensitivityResidual_JointedRigidBodyContactSystem(aSensResPR,aSensResLM,
                                                         BOX_LEN,irb,desSens.aPlank[irb].l,itype_anchor_reloc,
                                                         aRI,aRD,aFix,aJFixForce,aExForce,
                                                         gravity,
                                                         contProp,
                                                         aFP);          
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    AddSensitivityResidual_JointedRigidBodyContactSystem(aSensResPR,aSensResLM,
                                                         JNT_POS,ifix,desSens.aJoint[ifix],itype_anchor_reloc,
                                                         aRI,aRD,aFix,aJFixForce,aExForce,
                                                         gravity,
                                                         contProp,
                                                         aFP);              
  }  
}

void RelocateAnchors_JointedRigidBodyContactSystem
(std::vector<CContactPoints_RigidBody3D>& aFP,
 PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc, 
 const std::pair< Com::CVector3D, Com::CVector3D>& sns_prlm, 
 double eps,
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CRigid3D_Def>& aRD,   
 const Com::CVector3D& cont_norm)
{
  if( itype_s != BOX_POS && itype_s != BOX_ROT && itype_s != BOX_LEN ) return;  // incase of jnt_movement return;
  if( itype_anchor_reloc == DONT_MOVE) return;  
  if( itype_anchor_reloc == MOVE_TOGETHER ){
    const CRigid3D_Ini& ri = aRI[iobj_s];    
    CContactPoints_RigidBody3D& aCP = aFP[iobj_s];
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);      
      Com::CVector3D dv1 = GetSensitivityCurrentPointOnRigidBody(itype_s,iobj_s,idim_s, iobj_s, 
                                                                 aRI[iobj_s],aRD[iobj_s],
                                                                 i,aCP, sns_prlm );
      dv1 *= eps;
      dv1 -= cont_norm*Com::Dot(cont_norm,dv1);
      aCP.aFP[i].aloc += dv1;
    }    
  }
  if( itype_anchor_reloc == MOVE_INITIAL ){    
    const CRigid3D_Ini& ri = aRI[iobj_s];   
    CContactPoints_RigidBody3D& aCP = aFP[iobj_s];    
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;    
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);            
      Com::CVector3D v1 = ri.GetInitialPosition(v0);
      aCP.aFP[i].aloc = v1;
    }
  }
}
/*
void RelocateAnchors_JointedRigidBodyContactSystem
(std::vector<CContactPoints_RigidBody3D>& aFP,
 const std::vector<CRigid3D_Ini>& aRI)
{
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const CRigid3D_Ini& ri = aRI[irb];   
    CContactPoints_RigidBody3D& aCP = aFP[irb];    
    for(unsigned int i=0;i<aCP.Size();i++){
      if( aCP.aFP[i].itype_contact != 1 ) continue;    
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);            
      Com::CVector3D v1 = ri.GetInitialPosition(v0);
      aCP.aFP[i].aloc = v1;
    }  
  }    
}*/




void MoveInitialConfig_JointedRigidBodyContactSystem
(PARAM_TYPE itype_s, unsigned int iobj_s, unsigned int idim_s, double eps,
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CJoint_Fix>& aFix )
{
  if(      itype_s == BOX_POS ){
    CRigid3D_Ini& ri1 = aRI[iobj_s];    
    Com::CVector3D e_s(0,0,0); e_s[idim_s] = eps;
    ri1.SetIniPosCG( ri1.GetIniPosCG()+e_s );
  }
  else if( itype_s == BOX_ROT ){
    CRigid3D_Ini& ri1 = aRI[iobj_s];        
    Com::CVector3D e_s(0,0,0); e_s[idim_s] = eps;
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e_s);
    ri1.AddRotation_IniLeft(rot);  
  }      
  else if( itype_s == BOX_LEN ){
    CRigid3D_Ini& ri1 = aRI[iobj_s];        
    double xlen=ri1.xlen; 
    double ylen=ri1.ylen;
    double zlen=ri1.zlen;    
    const double rho = ri1.GetMass()/(xlen*ylen*zlen);
    if( idim_s == 0 ){ xlen += eps; }
    if( idim_s == 1 ){ ylen += eps; }
    if( idim_s == 2 ){ zlen += eps; }            
    ri1.SetModeBox(xlen, ylen, zlen, rho);
  }
  else if( itype_s == JNT_POS ){
    Com::CVector3D e_s(0,0,0); e_s[idim_s] = eps;
    assert( iobj_s <= aFix.size() );
    CJoint_Fix& cjs = aFix[iobj_s];
    cjs.ini_pos_joint += e_s;
  }  
}

void SetPatternDynamicSystem
(MatVec::CMatDia_BlkCrs& m_mat,     
 MatVec::CVector_Blk& m_residual,
 unsigned int nRB,
 const std::vector<CJoint_Fix>& aFix )
{
  unsigned int nConst = aFix.size();
  std::vector<unsigned int> m_aBlkSize;
  const unsigned int nblk = nRB+nConst;
  {
    m_aBlkSize.resize(nblk);
    for(unsigned int irb=0;irb<nRB;irb++){
      m_aBlkSize[irb] = 6;
    }
    for(unsigned int icst=0;icst<nConst;icst++){
      m_aBlkSize[nRB+icst] = 6;
    }
    m_mat.Initialize(nblk,m_aBlkSize, nblk,m_aBlkSize);
    m_residual.Initialize(nblk,m_aBlkSize);
  }
  /*
  {
    Com::CIndexedArray crs;
    crs.InitializeSize(nblk);
    crs.index.resize(nblk+1);
    crs.index[0] = 0;
    for(unsigned int i=0;i<nRB;i++){ crs.index[i+1]=0; }
    for(unsigned int i=nRB;i<nRB+nConst;i++){ crs.index[i+1]=0; }
    for(unsigned int icst=0;icst<nConst;icst++){
      const std::vector<unsigned int>& aIndRB = aFix[icst].GetAry_IndexRB();
      assert( aIndRB.size() == 2 );
      for(unsigned int i=0;i<aIndRB.size();i++){
        const unsigned int irb0 = aIndRB[i];
        crs.index[irb0+1] += 2;
      }
      crs.index[icst+nRB+1] += 2;
    }
    for(unsigned int i=0;i<nblk;i++){ 
      crs.index[i+1] = crs.index[i+1] + crs.index[i];
    }
    const unsigned int ncrs = crs.index[nblk];
    crs.array.resize(ncrs);
    for(unsigned int icst=0;icst<nConst;icst++){
      const std::vector<unsigned int>& aIndRB = aFix[icst].GetAry_IndexRB();
      assert( aIndRB.size() == 2 );
      unsigned int irb0 = aIndRB[0];
      unsigned int irb1 = aIndRB[1];
      {
        crs.array[ crs.index[irb0] ] = irb1;
        crs.index[irb0]++;
        crs.array[ crs.index[irb0] ] = icst+nRB;
        crs.index[irb0]++;
      }
      {
        crs.array[ crs.index[irb1] ] = irb0;
        crs.index[irb1]++;
        crs.array[ crs.index[irb1] ] = icst+nRB;
        crs.index[irb1]++;
      }
      {
        crs.array[ crs.index[icst+nRB] ] = irb0;
        crs.index[icst+nRB]++;
        crs.array[ crs.index[icst+nRB] ] = irb1;
        crs.index[icst+nRB]++;
      }
    }
    for(int i=nblk;i>0;i--){ 
      crs.index[i] = crs.index[i-1];
    }
    crs.index[0] = 0;
    crs.Sort();
    assert( crs.index[nRB+nConst] == ncrs );
    m_mat.AddPattern(crs);      
  }  */
  {
    Com::CIndexedArray crs;
    crs.InitializeSize(nblk);
    crs.index.resize(nblk+1);
    crs.index[0] = 0;
    for(unsigned int i=0;i<nRB;i++){ crs.index[i+1]=0; }
    for(unsigned int i=nRB;i<nRB+nConst;i++){ crs.index[i+1]=0; }
    for(unsigned int icst=0;icst<nConst;icst++){
      {
        const unsigned int irb0 = aFix[icst].irb0;
        crs.index[irb0+1] += 1;
      }
      {
        const unsigned int irb1 = aFix[icst].irb1;
        crs.index[irb1+1] += 1;
      }      
      crs.index[icst+nRB+1] += 2;
    }
    for(unsigned int i=0;i<nblk;i++){ 
      crs.index[i+1] = crs.index[i+1] + crs.index[i];
    }
    const unsigned int ncrs = crs.index[nblk];
    crs.array.resize(ncrs);
    for(unsigned int icst=0;icst<nConst;icst++){
      unsigned int irb0 = aFix[icst].irb0;
      unsigned int irb1 = aFix[icst].irb1;
      {
        crs.array[ crs.index[irb0] ] = icst+nRB;
        crs.index[irb0]++;
      }
      {
        crs.array[ crs.index[irb1] ] = icst+nRB;
        crs.index[irb1]++;
      }
      {
        crs.array[ crs.index[icst+nRB] ] = irb0;
        crs.index[icst+nRB]++;
        crs.array[ crs.index[icst+nRB] ] = irb1;
        crs.index[icst+nRB]++;
      }
    }
    for(int i=nblk;i>0;i--){ 
      crs.index[i] = crs.index[i-1];
    }
    crs.index[0] = 0;
    crs.Sort();
    assert( crs.index[nRB+nConst] == ncrs );
    m_mat.AddPattern(crs);      
  }    
}


void GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem
(CSimulationSensitivity& simSens,
 ////
 PARAM_TYPE itype_s, unsigned int irb_s, unsigned int idim_s,
 ANCHOR_RELOCATION_TYPE itype_anchor_reloc,
 ////
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CRigid3D_Def>& aRD,  
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp,
 const std::vector<CContactPoints_RigidBody3D>& aFP)
{
  //  std::cout << "Get Sensitivity w.r.t initial configuration (improved version) " << std::endl;
  MatVec::CMatDia_BlkCrs m_mat;
  MatVec::CVector_Blk m_r;
  SetPatternDynamicSystem(m_mat,m_r,aRI.size(),aFix);
  
  MatVec::CMatDiaFrac_BlkCrs m_frac;
  m_frac.MakePattern_Initialize( m_mat );
  m_frac.AddFracPtn(-1); // compute exact inverse
  m_frac.MakePatternFinalize();
  
  ////////////////////////////////////////////////////////////////
  // Make Linear System
  m_mat.SetZero();
  m_r.SetVectorZero();
  unsigned int nRB = aRI.size();
  for(unsigned int irb=0;irb<nRB;irb++){
    AddMatrix_StaticContact_Sensitivity(m_mat, irb,aRI[irb],aRD[irb],
                                        itype_s,irb_s,idim_s,
                                        contProp.cont_base,contProp.cont_norm,contProp.stiff_n,contProp.stiff_f,
                                        aFP[irb],
                                        itype_anchor_reloc);
    {
      Com::CMatrix3 m0; m0.SetIdentity();
      AddMatrix(m_mat, irb,0,  irb,0,  m0,1.0e-8, true);
      AddMatrix(m_mat, irb,3,  irb,3,  m0,1.0e-8, true);            
    } 
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cjs = aFix[ifix];
    const CJoint_Fix_Force& jff = aJFixForce[ifix];    
    unsigned int irb0 = cjs.irb0; 
    unsigned int irb1 = cjs.irb1; 
    AddMatrix_StaticJoint(m_mat,aRI.size(),ifix,cjs,jff,
                          aRI[irb0], aRD[irb0],
                          aRI[irb1], aRD[irb1]);    
  }
  for(unsigned int ief=0;ief<aExForce.size();ief++){
    const CExternalForceOnRigidBody& ef = aExForce[ief];
    unsigned int irb = ef.irb;
    AddMatrix_ExternalForce_Static(m_mat, irb, aRI[irb],aRD[irb], ef.pos, ef.force );
  }
  
  
  // residual 
  std::vector< std::pair<Com::CVector3D, Com::CVector3D> > aSensResPR, aSensResLM;
  for(unsigned int irb=0;irb<aRI.size();irb++){     aSensResPR.push_back( std::make_pair(Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) ); }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){ aSensResLM.push_back( std::make_pair(Com::CVector3D(0,0,0), Com::CVector3D(0,0,0) ) ); }  
  AddSensitivityResidual_JointedRigidBodyContactSystem(aSensResPR,aSensResLM,
                                                       itype_s,irb_s,idim_s,1,itype_anchor_reloc,
                                                       aRI,aRD,aFix,aJFixForce,aExForce,
                                                       gravity,
                                                       contProp,
                                                       aFP);
  for(unsigned int irb=0;irb<aRI.size();irb++){
    AddVector(m_r, irb,0, aSensResPR[irb].first,  1);
    AddVector(m_r, irb,3, aSensResPR[irb].second, 1);    
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    AddVector(m_r, ifix+nRB,0, aSensResLM[ifix].first,  1);
    AddVector(m_r, ifix+nRB,3, aSensResLM[ifix].second, 1);        
  }
  
  ////////////////////////////////////////////////////////////////
  // Solve Linear System
  m_frac.SetValue( m_mat );
  m_frac.Solve( m_r );
  
  double tmp[6];
  simSens.aPlankPR.clear();
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const unsigned int nlen = m_r.Len(irb); assert( nlen == 6 );
    for(unsigned int ilen=0;ilen<nlen;ilen++){ tmp[ilen] = m_r.GetValue(irb,ilen); }
    simSens.aPlankPR.push_back( std::make_pair(Com::CVector3D(tmp[0],tmp[1],tmp[2]), 
                                               Com::CVector3D(tmp[3],tmp[4],tmp[5]) ) );    
  }    
  simSens.aJointLM.clear();
  for(unsigned int icst=0;icst<aFix.size();icst++){
    const unsigned int nlen = m_r.Len(icst+nRB); assert( nlen == 6 );
    for(unsigned int ilen=0;ilen<nlen;ilen++){ tmp[ilen] = m_r.GetValue(icst+nRB,ilen); }
    simSens.aJointLM.push_back( std::make_pair(Com::CVector3D(tmp[0],tmp[1],tmp[2]), 
                                               Com::CVector3D(tmp[3],tmp[4],tmp[5]) ) );
  }
}
void MakeLinearSystemFrac_JointedRigidBodyContactSystem
(MatVec::CMatDia_BlkCrs& m_mat,
 MatVec::CVector_Blk& m_r,
 MatVec::CMatDiaFrac_BlkCrs& m_frac,
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CRigid3D_Def>& aRD,   
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp, 
 const std::vector<CContactPoints_RigidBody3D>& aFP)
{
  //  std::cout << "Get Sensitivity w.r.t initial configuration (improved version) " << std::endl;
  SetPatternDynamicSystem(m_mat,m_r,aRI.size(),aFix);
  
  m_frac.MakePattern_Initialize( m_mat );
  m_frac.AddFracPtn(-1); // compute exact inverse
  m_frac.MakePatternFinalize();
  
  ////////////////////////////////////////////////////////////////
  // Make Linear System
  m_mat.SetZero();
  m_r.SetVectorZero();
  unsigned int nRB = aRI.size();
  for(unsigned int irb=0;irb<aRI.size();irb++){
    AddMatrix_StaticContact(m_mat, irb,aRI[irb],aRD[irb],
                            contProp.cont_base,contProp.cont_norm,contProp.stiff_n,contProp.stiff_f,
                            aFP[irb]);
    {
      Com::CMatrix3 m0; m0.SetIdentity();
      AddMatrix(m_mat, irb,0,  irb,0,  m0,1.0e-8, true);
      AddMatrix(m_mat, irb,3,  irb,3,  m0,1.0e-8, true);            
    } 
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cjs = aFix[ifix];
    const CJoint_Fix_Force& jff = aJFixForce[ifix];    
    unsigned int irb0 = cjs.irb0;
    unsigned int irb1 = cjs.irb1;
    AddMatrix_StaticJoint(m_mat,aRI.size(),ifix,cjs,jff,
                          aRI[irb0], aRD[irb0],
                          aRI[irb1], aRD[irb1]);    
  }
  for(unsigned int ief=0;ief<aExForce.size();ief++){
    const CExternalForceOnRigidBody& ef = aExForce[ief];
    unsigned int irb = ef.irb;
    AddMatrix_ExternalForce_Static(m_mat, irb, aRI[irb], aRD[irb], ef.pos, ef.force );
  }
  // GetInverse
  m_frac.SetValue( m_mat );  
}
  

void GetSensitivityPRLM_InitialConfigBox_JointedRigidBodyContactSystem
(CSimulationSensitivity& simSens,
 ////
 const CDesignSensitivity& desSens,
 ////
 const std::vector<CRigid3D_Ini>& aRI, 
 const std::vector<CRigid3D_Def>& aRD,   
 const std::vector<CJoint_Fix>& aFix,
 const std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp, 
 const std::vector<CContactPoints_RigidBody3D>& aFP)
{
  assert( desSens.aPlank.size() == aRI.size() );
  assert( desSens.aJoint.size() == aFix.size() );  
  
  //  std::cout << "Get Sensitivity w.r.t initial configuration (improved version) " << std::endl;
  MatVec::CMatDia_BlkCrs m_mat;
  MatVec::CVector_Blk m_r;
  SetPatternDynamicSystem(m_mat,m_r,aRI.size(),aFix);
  
  MatVec::CMatDiaFrac_BlkCrs m_frac;
  m_frac.MakePattern_Initialize( m_mat );
  m_frac.AddFracPtn(-1); // compute exact inverse
  m_frac.MakePatternFinalize();
  
  ////////////////////////////////////////////////////////////////
  // Make Linear System
  m_mat.SetZero();
  m_r.SetVectorZero();
  unsigned int nRB = aRI.size();
  for(unsigned int irb=0;irb<aRI.size();irb++){
    AddMatrix_StaticContact(m_mat, irb,aRI[irb],aRD[irb],
                            contProp.cont_base,contProp.cont_norm,contProp.stiff_n,contProp.stiff_f,
                            aFP[irb]);
    {
      Com::CMatrix3 m0; m0.SetIdentity();
      AddMatrix(m_mat, irb,0,  irb,0,  m0,1.0e-8, true);
      AddMatrix(m_mat, irb,3,  irb,3,  m0,1.0e-8, true);            
    } 
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cjs = aFix[ifix];
    const CJoint_Fix_Force& jff = aJFixForce[ifix];    
    unsigned int irb0 = cjs.irb0;
    unsigned int irb1 = cjs.irb1;
    AddMatrix_StaticJoint(m_mat,aRI.size(),ifix,cjs,jff,
                          aRI[irb0], aRD[irb0],
                          aRI[irb1], aRD[irb1]);    
  }
  for(unsigned int ief=0;ief<aExForce.size();ief++){
    const CExternalForceOnRigidBody& ef = aExForce[ief];
    unsigned int irb = ef.irb;
    AddMatrix_ExternalForce_Static(m_mat, irb, aRI[irb], aRD[irb], ef.pos, ef.force );
  }
  // GetInverse
  m_frac.SetValue( m_mat );
  
  // residual 
  CSimulationSensitivity resSim; resSim.Initialize(aRI.size(), aFix.size());
  AddSensitivityResidual_JointedRigidBodyContactSystem(resSim.aPlankPR,resSim.aJointLM,
                                                       desSens,MOVE_INITIAL,
                                                       aRI,aRD,aFix,aJFixForce,aExForce,
                                                       gravity,
                                                       contProp,
                                                       aFP);
  
  for(unsigned int irb=0;irb<aRI.size();irb++){
    AddVector(m_r, irb,0, resSim.aPlankPR[irb].first,  1);
    AddVector(m_r, irb,3, resSim.aPlankPR[irb].second, 1);    
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    AddVector(m_r, ifix+nRB,0, resSim.aJointLM[ifix].first,  1);
    AddVector(m_r, ifix+nRB,3, resSim.aJointLM[ifix].second, 1);        
  }
  
  ////////////////////////////////////////////////////////////////
  // Solve Linear System
  m_frac.Solve( m_r );
  
  double tmp[6];
  simSens.aPlankPR.clear();
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const unsigned int nlen = m_r.Len(irb); assert( nlen == 6 );
    for(unsigned int ilen=0;ilen<nlen;ilen++){ tmp[ilen] = m_r.GetValue(irb,ilen); }
    simSens.aPlankPR.push_back( std::make_pair(Com::CVector3D(tmp[0],tmp[1],tmp[2]), 
                                               Com::CVector3D(tmp[3],tmp[4],tmp[5]) ) );    
  }    
  simSens.aJointLM.clear();
  for(unsigned int icst=0;icst<aFix.size();icst++){
    const unsigned int nlen = m_r.Len(icst+nRB); assert( nlen == 6 );
    for(unsigned int ilen=0;ilen<nlen;ilen++){ tmp[ilen] = m_r.GetValue(icst+nRB,ilen); }
    simSens.aJointLM.push_back( std::make_pair(Com::CVector3D(tmp[0],tmp[1],tmp[2]), 
                                               Com::CVector3D(tmp[3],tmp[4],tmp[5]) ) );
  }
}



bool UpdateValueOfRigidSystem_BackwardEular_JointedRigidBodySystem
(const std::vector<CRigid3D_Ini>& aRI,
 std::vector<CRigid3D_Def>& aRD,
 const std::vector<CJoint_Fix>& aFix, 
 std::vector<CJoint_Fix_Force>& aJFixForce,  
 MatVec::CVector_Blk& upd,
 double dt, bool is_first)
{
  double tmp[6];
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const unsigned int nlen = upd.Len(irb);
    assert( nlen <= 6 );
    for(unsigned int ilen=0;ilen<nlen;ilen++){
      tmp[ilen] = upd.GetValue(irb,ilen);
    }
    UpdateSolution_BackwardEular(aRD[irb], tmp, dt, is_first);
  }
  const unsigned int nRB = aRI.size();
  for(unsigned int icst=0;icst<aFix.size();icst++){
    const unsigned int nlen = upd.Len(icst+nRB);
    assert( nlen <= 6 );
    for(unsigned int ilen=0;ilen<nlen;ilen++){
      tmp[ilen] = upd.GetValue(icst+nRB,ilen);
    }
    UpdateLambda_BackwardEular(aJFixForce[icst], tmp, dt );  
  }
  return true;
}

bool UpdateValueOfRigidSystem_JointedRigidBodySystem_Static
(const std::vector<CRigid3D_Ini>& aRI,
 std::vector<CRigid3D_Def>& aRD,
 const std::vector<CJoint_Fix>& aFix, 
 std::vector<CJoint_Fix_Force>& aJFixForce,  
 MatVec::CVector_Blk& upd)
{
  double tmp[6];
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const unsigned int nlen = upd.Len(irb);
    assert( nlen <= 6 );
    for(unsigned int ilen=0;ilen<nlen;ilen++){
      tmp[ilen] = upd.GetValue(irb,ilen);
    }
//    std::cout << irb << "  " << tmp[0] << " " << tmp[1] << " " << tmp[2] << "  ";
//    std::cout << tmp[3] << " " <<  tmp[4] << " " << tmp[5] << std::endl;
    {
      CRigid3D_Def& rd = aRD[irb];
      rd.disp_cg.x += tmp[0];
      rd.disp_cg.y += tmp[1];
      rd.disp_cg.z += tmp[2];
      {	
        Com::CVector3D rot(tmp[3],tmp[4],tmp[5]);
        Com::CMatrix3 m; m.SetRotMatrix_Cartesian(rot);
        rd.AddRotation_IncRight( m );
      }      
    }
  }
  const unsigned int nRB = aRI.size();
  for(unsigned int icst=0;icst<aFix.size();icst++){
    const unsigned int nlen = upd.Len(icst+nRB);
    assert( nlen <= 6 );
    for(unsigned int ilen=0;ilen<nlen;ilen++){
      tmp[ilen] = upd.GetValue(icst+nRB,ilen);
    }
//    std::cout << icst << "  " << tmp[0] << " " << tmp[1] << " " << tmp[2] << "  ";
//    std::cout << tmp[3] << " " <<  tmp[4] << " " << tmp[5] << std::endl;
    for(unsigned int i=0;i<6;i++){ aJFixForce[icst].lambda[i] += tmp[i]; };    
  }
  return true;
}


void Static_UntactAnchor_JointedRigidBodyContactSystem
(const std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,  
 const std::vector<CJoint_Fix>& aFix,
 std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp, 
 const std::vector<CContactPoints_RigidBody3D>& aFP,
 double& res_norm, double& upd_norm)
{
  MatVec::CMatDia_BlkCrs m_mat;
  MatVec::CVector_Blk m_r;
  SetPatternDynamicSystem(m_mat,m_r,aRI.size(),aFix);
  MatVec::CMatDiaFrac_BlkCrs m_frac;
  m_frac.MakePattern_Initialize( m_mat );
  m_frac.AddFracPtn(-1);
  m_frac.MakePatternFinalize();
  ////////////////
  m_mat.SetZero();
  m_r.SetVectorZero();
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const CRigid3D_Ini& ri = aRI[irb];
    AddVector(m_r, irb,0,  gravity, ri.mass );
    {
      Com::CMatrix3 ind0(1.0e-5);
      AddMatrix(m_mat, irb,0,  irb,0,     ind0, 1, true  ); 
      AddMatrix(m_mat, irb,3,  irb,3,     ind0, 1, true  ); 
    }
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);            
      const CFrictionPoint& fp = aCP.aFP[i];
      if( !fp.is_force_cont ) continue;
      AddLinSys_FrictionalContact_Penalty_Static_Contact
      (m_mat,m_r,
       irb, aRI[irb], aRD[irb],
       v0, 
       contProp.cont_base,contProp.cont_norm,contProp.stiff_n,contProp.stiff_f,
       fp);
    }
  }
  for(unsigned int ief=0;ief<aExForce.size();ief++){
    const CExternalForceOnRigidBody& ef = aExForce[ief];    
    unsigned int irb = ef.irb;
    AddLinearSystem_ExternalForce_Static(m_mat,m_r,
                                         irb,
                                         aRI[irb], aRD[irb],
                                         ef.pos, ef.force);
  }  
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){    
    const CJoint_Fix& cjs = aFix[ifix];
    const CJoint_Fix_Force& jff = aJFixForce[ifix];    
    unsigned int irb0 = cjs.irb0; 
    unsigned int irb1 = cjs.irb1; 
    AddLinearSystem_JointFix_Static(m_mat,m_r,
                                    aRI.size(),
                                    ifix,cjs,jff,
                                    aRI[irb0],aRD[irb0],
                                    aRI[irb1],aRD[irb1]);
  }  
  ////////////////////////////////
  res_norm = sqrt( m_r.GetSquaredVectorNorm() );  
  /*
  double tmp[6];
  for(unsigned int irb=0;irb<aRI.size();irb++){
    const unsigned int nlen = m_r.Len(irb); assert( nlen <= 6 );
    for(unsigned int ilen=0;ilen<nlen;ilen++){
      tmp[ilen] = m_r.GetValue(irb,ilen);
    }
    std::cout << "rf " << irb << "  " << tmp[0] << " " << tmp[1] << " " << tmp[2] << "  ";
    std::cout << tmp[3] << " " <<  tmp[4] << " " << tmp[5] << std::endl;
  }
  const unsigned int nRB = aRI.size();
  for(unsigned int icst=0;icst<aFix.size();icst++){
    const unsigned int nlen = m_r.Len(icst+nRB); assert( nlen <= 6 );
    for(unsigned int ilen=0;ilen<nlen;ilen++){
      tmp[ilen] = m_r.GetValue(icst+nRB,ilen);
    }
    std::cout << "jf " << icst << "  " << tmp[0] << " " << tmp[1] << " " << tmp[2] << "  ";
    std::cout << tmp[3] << " " <<  tmp[4] << " " << tmp[5] << std::endl;
  }    
   */
  m_frac.SetValue( m_mat );
  m_frac.Solve( m_r ); 
  ////////////////////////////////
  UpdateValueOfRigidSystem_JointedRigidBodySystem_Static(aRI,aRD,aFix,aJFixForce,m_r);
  upd_norm = sqrt( m_r.GetSquaredVectorNorm() );    
//  std::cout << "upd : " << ls.DOT(-2,-2) << " res : " << ls.DOT(-1,-1) << std::endl;  
//  std::cout << upd_norm << " " << res_norm << std::endl;
}

void StepTime_BackwardEular_UntactAnchor_JointedRigidBodyContactSystem
(double dt,
 const std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,  
 const std::vector<CJoint_Fix>& aFix,
 std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp, 
 const std::vector<CContactPoints_RigidBody3D>& aFP,
 double& res_norm, double& upd_norm)
{
  MatVec::CMatDia_BlkCrs m_mat;
  MatVec::CVector_Blk m_r;
  SetPatternDynamicSystem(m_mat,m_r,aRI.size(),aFix);
  MatVec::CMatDiaFrac_BlkCrs m_frac;
  m_frac.MakePattern_Initialize( m_mat );
  m_frac.AddFracPtn(-1);
  m_frac.MakePatternFinalize();
  ////////////////
  m_mat.SetZero();
  m_r.SetVectorZero();
  for(unsigned int irb=0;irb<aRI.size();irb++){
    AddLinearSystem_RigidBody_BackwardEular(m_mat,m_r,irb,
                                            aRI[irb],aRD[irb],
                                            dt,gravity,true);    
    const CContactPoints_RigidBody3D& aCP = aFP[irb];
    const CRigid3D_Ini& ri = aRI[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);            
      const CFrictionPoint& fp = aCP.aFP[i];
      if( !fp.is_force_cont ) continue;
      AddLinSys_FrictionalContact_Penalty_NonStatic_BackwardEular_Contact
      (m_mat,m_r,
       irb, aRI[irb], aRD[irb],
       v0, dt,
       contProp.cont_base,contProp.cont_norm,contProp.stiff_n,contProp.stiff_f,
       fp);
    }
  }
  for(unsigned int ief=0;ief<aExForce.size();ief++){
    const CExternalForceOnRigidBody& ef = aExForce[ief];    
    unsigned int irb = ef.irb;
    AddLinearSystem_ExternalForce_BackwardEular(m_mat,m_r,
                                                irb,
                                                aRI[irb], aRD[irb],
                                                dt,true,
                                                ef.pos, ef.force);
  }  
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){    
    const CJoint_Fix& cjs = aFix[ifix];
    const CJoint_Fix_Force& jff = aJFixForce[ifix];    
    unsigned int irb0 = cjs.irb0; 
    unsigned int irb1 = cjs.irb1; 
    AddLinearSystem_JointFix_BackwardEular(m_mat,m_r,
                                           aRI.size(),
                                           ifix,cjs,jff,
                                           aRI[irb0],aRD[irb0],
                                           aRI[irb1],aRD[irb1],
                                           dt);
  }  
  ////////////////////////////////
  res_norm = sqrt( m_r.GetSquaredVectorNorm() );  
  m_frac.SetValue( m_mat );
  m_frac.Solve( m_r ); 
  ////////////////////////////////
  UpdateValueOfRigidSystem_BackwardEular_JointedRigidBodySystem(aRI,aRD,aFix,aJFixForce,m_r,dt,true);    
  //  std::cout << "upd : " << ls.DOT(-2,-2) << " res : " << ls.DOT(-1,-1) << std::endl;
  upd_norm = sqrt( m_r.GetSquaredVectorNorm() );    
}

void StepTime_BackwardEular_JointedRigidBodyContactSystem
(double dt,
 const std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD, 
 const std::vector<CJoint_Fix>& aFix,
 std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp, 
 double myu_s, double myu_k,
 std::vector<CContactPoints_RigidBody3D>& aFP, 
 double& res_norm, double& upd_norm)
{
  MatVec::CMatDia_BlkCrs m_mat;
  MatVec::CVector_Blk m_r;
  SetPatternDynamicSystem(m_mat,m_r,aRI.size(),aFix);
  MatVec::CMatDiaFrac_BlkCrs m_frac;
  m_frac.MakePattern_Initialize( m_mat );
  m_frac.AddFracPtn(-1);
  m_frac.MakePatternFinalize();
  
  ////////////////////////////////////////////////
  m_mat.SetZero();
  m_r.SetVectorZero();
  for(unsigned int irb=0;irb<aRI.size();irb++){
    AddLinearSystem_RigidBody_BackwardEular(m_mat,m_r,
                                            irb,
                                            aRI[irb],aRD[irb],  
                                            dt,gravity,true);
    const CRigid3D_Ini& ri = aRI[irb];    
    CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);            
      AddLinSys_FrictionalContact_Penalty_NonStatic_BackwardEular
      (m_mat,m_r,
       irb, aRI[irb], aRD[irb],
       v0, dt,
       contProp.cont_base,contProp.cont_norm,contProp.stiff_n,contProp.stiff_f,       
       myu_s,myu_k,       
       aCP.aFP[i]);
    }
  }
  for(unsigned int ief=0;ief<aExForce.size();ief++){
    const CExternalForceOnRigidBody& ef = aExForce[ief];    
    unsigned int irb = ef.irb;
    AddLinearSystem_ExternalForce_BackwardEular(m_mat,m_r,
                                                irb,
                                                aRI[irb],aRD[irb], 
                                                dt,true,
                                                ef.pos, ef.force);
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    const CJoint_Fix& cjs = aFix[ifix];
    const CJoint_Fix_Force& jff = aJFixForce[ifix];
    unsigned int irb0 = cjs.irb0; 
    unsigned int irb1 = cjs.irb1; 
    AddLinearSystem_JointFix_BackwardEular(m_mat,m_r,
                                           aRI.size(),
                                           ifix,cjs,jff,
                                           aRI[irb0], aRD[irb0],
                                           aRI[irb1], aRD[irb1],
                                           dt);
  }
  ////////////////////////////////
  res_norm = sqrt( m_r.GetSquaredVectorNorm() );
  m_frac.SetValue( m_mat );
  m_frac.Solve( m_r ); 
  
  ////////////////////////////////
  UpdateValueOfRigidSystem_BackwardEular_JointedRigidBodySystem(aRI,aRD,aFix,aJFixForce,m_r,dt,true);  
  for(unsigned int irb=0;irb<aRI.size();irb++){
    UpdateContact_RigidBody(aRI[irb], aRD[irb],
//                            cont_base,cont_norm,stiff_n,stiff_f,
                            contProp.cont_base,contProp.cont_norm,contProp.stiff_n,contProp.stiff_f,                            
                            myu_s,myu_k,
                            aFP[irb]);
  }
  upd_norm = sqrt( m_r.GetSquaredVectorNorm() );
}


// get sensitivity of joint
void SetJointSensitivity
(CDesignSensitivity& desSens,
 const std::vector<CRigid3D_Ini>& aRI,                                                                       
 const std::vector<CJoint_Fix>& aFix,
 const std::vector< std::pair<unsigned int,unsigned int> >& aFixDir)
{
  std::vector< std::vector< std::pair<unsigned int,unsigned int> > > aJntDir;
  aJntDir.resize(aRI.size());
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    unsigned int irb0 = aFix[ifix].irb0;
    unsigned int irb1 = aFix[ifix].irb1;
    unsigned int idir0 = aFixDir[ifix].first;      
    unsigned int idir1 = aFixDir[ifix].second;            
    aJntDir[irb0].push_back( std::make_pair(ifix,idir0) );
    aJntDir[irb1].push_back( std::make_pair(ifix,idir1) );      
  }
  const Com::CVector3D aFaceNormal[6][3] = {
    { Com::CVector3D(+1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-1, 0, 0), Com::CVector3D(0, 1, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0,+1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(-0,-1, 0), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 0, 1) },
    { Com::CVector3D(+0, 0,+1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) },
    { Com::CVector3D(-0, 0,-1), Com::CVector3D(1, 0, 0), Com::CVector3D(0, 1, 0) } };      
  desSens.aJoint.resize( aFix.size() );
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){ desSens.aJoint[ifix] = Com::CVector3D(0,0,0); }
  for(unsigned int irb0=0;irb0<aRI.size();irb0++){
    const CRigid3D_Ini& ri0 = aRI[irb0];     
    for(unsigned int ij=0;ij<aJntDir[irb0].size();ij++){
      unsigned int ifix = aJntDir[irb0][ij].first;
      const CJoint_Fix& fix = aFix[ifix];
      unsigned int idir0 = aJntDir[irb0][ij].second;
      if( idir0 == 4 || idir0 == 5 ){        
        Com::CMatrix3 R0 = ri0.GetRotMatrix_Ini();
        Com::CVector3D n0 = R0.MatVec(aFaceNormal[idir0][0]);
        Com::CMatrix3 L0(ri0.xlen*0.5, ri0.ylen*0.5, ri0.zlen*0.5);
        Com::CMatrix3 RL0 = R0.MatMat(L0);
        Com::CVector3D b0 = RL0.MatVec(aFaceNormal[idir0][0])+ri0.GetIniPosCG();
        /////
        unsigned int irb1, idir1;
        if( fix.irb0 == irb0 ){
          irb1 = fix.irb1;
          idir1 = aFixDir[ifix].second;
        }
        else{
          irb1 = fix.irb0;
          idir1 = aFixDir[ifix].first;
        }
        const CRigid3D_Ini& ri1 = aRI[irb1];  
        Com::CMatrix3 R1 = ri1.GetRotMatrix_Ini();
        Com::CVector3D n1 = R1.MatVec(aFaceNormal[idir1][0]);
        Com::CMatrix3 L1(ri1.xlen*0.5, ri1.ylen*0.5, ri1.zlen*0.5);
        Com::CMatrix3 RL1 = R1.MatMat(L1);
        Com::CVector3D b1 = RL1.MatVec(aFaceNormal[idir1][0])+ri1.GetIniPosCG();
        Com::CVector3D h1 = RL1.MatVec(aFaceNormal[idir1][1]);
        //////
        const Com::CVector3D& h0 = RL0.MatVec( aFaceNormal[idir0][1]);
        const Com::CVector3D& v0 = RL0.MatVec( aFaceNormal[idir0][2]);
        const double dlen1 = Com::Dot(b0-b1,n0) / Com::Dot(n1,n0);              
        double rha0, rva0;
        bool resa0 = IsRayCollideQuad(b1+dlen1*n1+h1,n0, b0,n0,h0,v0, rha0,rva0); 
        double rhb0, rvb0;
        bool resb0 = IsRayCollideQuad(b1+dlen1*n1-h1,n0, b0,n0,h0,v0, rhb0,rvb0);         
        if( !resa0 || !resb0 ){
          Com::CVector3D s = desSens.aPlank[irb0].p;
          s -= Dot(n0,s)*n0;
          desSens.aJoint[ifix] += s;
        }        
        continue;
      }
      else{
        const Com::CVector3D& b = aFaceNormal[idir0][0]*0.5;
        const Com::CMatrix3 B(b.x,b.y,b.z);
        const Com::CVector3D l(ri0.xlen,ri0.ylen,ri0.zlen);
        const Com::CMatrix3& Q = ri0.GetRotMatrix_Ini();
        const Com::CMatrix3& QB = Q.MatMat(B);
        const Com::CVector3D& QBa = QB.MatVec(l);
        const Com::CVector3D v = ri0.GetIniPosCG() + QBa;
        desSens.aJoint[ifix] +=  desSens.aPlank[irb0].p;
        desSens.aJoint[ifix] += -Cross(QBa,desSens.aPlank[irb0].r);
        desSens.aJoint[ifix] += QB.MatVec(desSens.aPlank[irb0].l);        
      }
    }
  }    
}

/*
void MoveCurrentConfig_JointedRigidBodyContactSystem
(const CSimulationSensitivity& simSens, double eps,
 std::vector<CRigid3D_Def>& aRD,   
 std::vector<CJoint_Fix_Force>& aJFixForce)
{
  const std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aSensPR = simSens.aPlankPR;
  const std::vector< std::pair<Com::CVector3D,Com::CVector3D> >& aSensLM = simSens.aJointLM;
  for(unsigned int irb=0;irb<aRD.size();irb++){ // BOX_POS
    CRigid3D_Def& rd1 = aRD[irb];    
    Com::CVector3D e_s = aSensPR[irb].first * eps;
    rd1.AddDispCG( e_s );
  } 
  for(unsigned int irb=0;irb<aRD.size();irb++){ // BOX_ROT
    CRigid3D_Def& rd1 = aRD[irb];    
    Com::CVector3D e_s = aSensPR[irb].second * eps;
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e_s);
    rd1.AddRotation_IncRight(rot);  
  }
  for(unsigned int ifix=0;ifix<aJFixForce.size();ifix++){ // JNT_FORCE
    CJoint_Fix_Force& jff = aJFixForce[ifix];
    jff.lambda[0] += aSensLM[ifix].first.x*eps;
    jff.lambda[1] += aSensLM[ifix].first.y*eps;
    jff.lambda[2] += aSensLM[ifix].first.z*eps;
    jff.lambda[3] += aSensLM[ifix].second.x*eps;
    jff.lambda[4] += aSensLM[ifix].second.y*eps;
    jff.lambda[5] += aSensLM[ifix].second.z*eps;  
  }
}*/

void MoveCurrentDeform(const CSimulationSensitivity& simSens, double eps,
                       std::vector<CRigid3D_Def>& aRD,                       
                       std::vector<CJoint_Fix_Force>& aJFixForce)
{
  assert( simSens.aPlankPR.size() == aRD .size() );
  assert( simSens.aJointLM.size() == aJFixForce.size() );
  for(unsigned int irb=0;irb<aRD.size();irb++){
    const Com::CVector3D& u = simSens.aPlankPR[irb].first *eps;
    const Com::CVector3D& p = simSens.aPlankPR[irb].second*eps;
    CRigid3D_Def& rd = aRD[irb];
    rd.AddDispCG( u );
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(p);
    rd.AddRotation_IncRight( rot ); 
  }
  for(unsigned int ifix=0;ifix<aJFixForce.size();ifix++){
    const Com::CVector3D& p = simSens.aJointLM[ifix].first *eps;
    const Com::CVector3D& q = simSens.aJointLM[ifix].second*eps;
//    std::cout << ifix << " " << p << "   " << q << std::endl;          
    for(unsigned int idim=0;idim<3;idim++){
      aJFixForce[ifix].lambda[idim+0] += +p[idim];
      aJFixForce[ifix].lambda[idim+3] += +q[idim]; 
    }
  }    
}


  /////
  /*
   const Com::CVector3D aFaceNormal[6] = {
   Com::CVector3D(+1, 0, 0),
   Com::CVector3D(-1, 0, 0),
   Com::CVector3D(+0,+1, 0),
   Com::CVector3D(-0,-1, 0),
   Com::CVector3D(+0, 0,+1),
   Com::CVector3D(-0, 0,-1) };
   for(unsigned int ifix=0;ifix<aFix.size();ifix++){
   CJoint_Fix& cst = aFix[ifix];
   unsigned int idir0 = aFixDir[ifix].first;
   unsigned int idir1 = aFixDir[ifix].second;    
   if(      (idir0==4||idir0==5) && (idir1!=4&&idir1!=5) ){  // joint is at 1
   unsigned int irb1 = cst.aIndRB[1];      
   const CRigidBody3D& rb1 = aRB[irb1]; 
   const Com::CVector3D& d1 = rb1.ri.GetIniPosCG();            
   const Com::CMatrix3& Q1 = rb1.GetRotMatrix_Ini();      
   const Com::CVector3D a1(ri1.xlen,ri1.ylen,ri1.zlen);        
   const Com::CVector3D& b1 = aFaceNormal[idir1]*0.5;      
   const Com::CMatrix3 B1(b1.x,b1.y,b1.z);        
   const Com::CMatrix3& QB1 = Q1.MatMat(B1);        
   const Com::CVector3D& QBa1 = QB1.MatVec(a1);            
   cst.ini_pos_joint = d1 + QBa1;
   }
   else if( (idir0!=4&&idir0!=5) && (idir1==4||idir1==5) ){  // joint is at 0
   unsigned int irb0 = cst.aIndRB[0];
   const CRigidBody3D& rb0 = aRB[irb0];            
   const Com::CVector3D& d0 = rb0.ri.GetIniPosCG();      
   const Com::CMatrix3& Q0 = rb0.ri.GetRotMatrix_Ini();      
   const Com::CVector3D a0(ri1.xlen,ri0.ylen,ri0.zlen);      
   const Com::CVector3D& b0 = aFaceNormal[idir0]*0.5;      
   const Com::CMatrix3 B0(b0.x,b0.y,b0.z);
   const Com::CMatrix3& QB0 = Q0.MatMat(B0);
   const Com::CVector3D& QBa0 = QB0.MatVec(a0);      
   cst.ini_pos_joint = d0 + QBa0;      
   }    
   }    
   */
//}


void MoveInitialDesignAnchor
(const CDesignSensitivity& desSens, double eps, 
 std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CJoint_Fix>& aFix,
 std::vector<CContactPoints_RigidBody3D>& aFP)
{
  for(unsigned int irb=0;irb<aRI.size();irb++){
    Com::CVector3D q = desSens.aPlank[irb].p*eps;
    Com::CVector3D s = desSens.aPlank[irb].r*eps;
    Com::CVector3D l = desSens.aPlank[irb].l*eps;
    CRigid3D_Ini& ri = aRI[irb]; 
    ri.SetIniPosCG( ri.GetIniPosCG() + q );    
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(s);
    ri.AddRotation_IniLeft( rot );
    const double rho = ri.GetMass() / (ri.xlen*ri.ylen*ri.zlen);    
    const double xlen = (ri.xlen + l.x > ri.zlen ) ? ri.xlen + l.x : ri.xlen; // prevent negative xlen
    const double ylen = (ri.ylen + l.y > ri.zlen ) ? ri.ylen + l.y : ri.ylen; // prevent negative ylen
    const double zlen = ri.zlen + l.z;
    ri.SetModeBox(xlen,ylen,zlen,rho);
    CContactPoints_RigidBody3D& aCP = aFP[irb];
    for(unsigned int i=0;i<aCP.Size();i++){
      if( !aCP.aFP[i].is_force_cont ) continue;    
      Com::CVector3D v0 = aCP.GetPosContactPoint(i,ri.xlen,ri.ylen,ri.zlen);            
      Com::CVector3D v1 = ri.GetInitialPosition(v0);
      aCP.aFP[i].aloc = v1;
    }
  }
  for(unsigned int ifix=0;ifix<aFix.size();ifix++){
    Com::CVector3D p = desSens.aJoint[ifix]*eps;
    aFix[ifix].ini_pos_joint += p;
  }  
}

void EnforceContact_ChangeBoxLen
(std::vector<CRigid3D_Ini>& aRI,  
 std::vector<CContactPoints_RigidBody3D>& aCP,
 const Com::CVector3D& n, const Com::CVector3D& b)
{
  const Com::CVector3D aFaceNormal[6] = {
    Com::CVector3D(+1, 0, 0),
    Com::CVector3D(-1, 0, 0),
    Com::CVector3D(+0,+1, 0),
    Com::CVector3D(-0,-1, 0),
    Com::CVector3D(+0, 0,+1),
    Com::CVector3D(-0, 0,-1) };
  
  for(unsigned int irb=0;irb<aRI.size();irb++){    
    int icont_flg = 0;
    if( aCP[irb].aFP[1].is_force_cont && aCP[irb].aFP[2].is_force_cont ){ icont_flg += 1; } // +x  
    if( aCP[irb].aFP[3].is_force_cont && aCP[irb].aFP[0].is_force_cont ){ icont_flg += 2; } // -x
    if( aCP[irb].aFP[2].is_force_cont && aCP[irb].aFP[3].is_force_cont ){ icont_flg += 4; } // +y  
    if( aCP[irb].aFP[0].is_force_cont && aCP[irb].aFP[1].is_force_cont ){ icont_flg += 8; } // -y    
    
    unsigned int idir0;
    if(      icont_flg == 1 ){ idir0 = 0; } // +x
    else if( icont_flg == 2 ){ idir0 = 1; } // -x  
    else if( icont_flg == 4 ){ idir0 = 2; }// +y
    else if( icont_flg == 8 ){ idir0 = 3; }// -y
    else{ continue; }    
    
    CRigid3D_Ini& ri0 = aRI[irb];
    const Com::CMatrix3 R0 = ri0.GetRotMatrix_Ini();
    const Com::CVector3D n0 = R0.MatVec(aFaceNormal[idir0]);
    const Com::CVector3D l0(ri0.xlen, ri0.ylen, ri0.zlen);    
    const Com::CMatrix3 B0(0.5*aFaceNormal[idir0].x,0.5*aFaceNormal[idir0].y,0.5*aFaceNormal[idir0].z);    
    const Com::CMatrix3 RB0 = R0.MatMat(B0);
    const Com::CVector3D t0 = RB0.MatVec(l0)+ri0.GetIniPosCG();  
    
    double dlen0 = Com::Dot(b-t0,n)/Com::Dot(n0,n);              
//    std::cout << dlen0 << std::endl;
    if( idir0 == 0 || idir0 == 1 ){
      ri0.xlen += dlen0;
      ri0.SetIniPosCG( ri0.GetIniPosCG() + dlen0*0.5*n0 );
    }
    else {
      ri0.ylen += dlen0;
      ri0.SetIniPosCG( ri0.GetIniPosCG() + dlen0*0.5*n0 );
    }        
//    CContactPoints_RigidBody3D& aCP = aCP_in[irb]; 
    for(unsigned int i=0;i<aCP[irb].Size();i++){
      CFrictionPoint& fp = aCP[irb].aFP[i];
      if( !fp.is_force_cont ) continue;    
      Com::CVector3D v0 = aCP[irb].GetPosContactPoint(i,ri0.xlen,ri0.ylen,ri0.zlen);            
      Com::CVector3D v1 = ri0.GetInitialPosition(v0);
      fp.aloc = v1;
    }    
  }
}


void SolveStaticEquilibrium
(const std::vector<CRigid3D_Ini>& aRI, 
 std::vector<CRigid3D_Def>& aRD,   
 const std::vector<CJoint_Fix>& aFix,
 std::vector<CJoint_Fix_Force>& aJFixForce, 
 const std::vector<CExternalForceOnRigidBody>& aExForce,
 ////
 const Com::CVector3D& gravity,
 const CContactProp& contProp, 
 const std::vector<CContactPoints_RigidBody3D>& aFP)
{
  for(unsigned int i=0;i<1000;i++){
    double resn, updn;
    Static_UntactAnchor_JointedRigidBodyContactSystem(aRI,aRD,aFix,aJFixForce,aExForce,
                                                      gravity,
                                                      //                                                                      cont_base,cont_norm,stiff_n,stiff_f,
                                                      contProp,
                                                      aFP,
                                                      resn,updn);    
//    std::cout << i << " "<< resn << " " << updn << std::endl;
    if( updn < 1.0e-10 ) break;
  }  
}

