/*
 *  untitled.cpp
 *  rigid3d
 *
 *  Created by Nobuyuki Umetani on 9/24/11.
 *  Copyright 2011 The University of Tokyo. All rights reserved.
 *
 */

#include <stdlib.h>
#include "rigid_pointset.h"

static double rand2(double min, double max){
  return rand()/(RAND_MAX+1.0)*(max-min)+min;
}

double GetContactEnergy
(const Com::CVector3D& c,
 const Com::CVector3D& v,
 const Com::CVector3D& b, 
 const Com::CVector3D& n, 
 double k)
{
  const Com::CVector3D& v1 = c+v;
  const double h = Dot(b-v1,n);
  return 0.5*h*h*k;
}

void Get1stDer_ContactEnergy
(Com::CVector3D& dce_du, Com::CVector3D& dce_dr,
 const Com::CVector3D& c, 
 const Com::CVector3D& v,
 const Com::CVector3D& b, 
 const Com::CVector3D& n, 
 double k)
{
  const Com::CVector3D& v1 = c+v;
  const double h = Dot(b-v1,n);
  dce_du = -h*k*n;
  dce_dr = -h*k*Com::CMatrix3(v).MatVec(n);
}

void Get2ndDer_ContactEnergy
(Com::CMatrix3& dce_dudu, Com::CMatrix3& dce_dudr, Com::CMatrix3& dce_drdr,
 const Com::CVector3D& c, 
 const Com::CVector3D& v, 
 const Com::CVector3D& cont_base, const Com::CVector3D& n, double cont_stiff)
{
  const Com::CVector3D& v1 = c+v;
  const double h = Com::Dot(cont_base-v1,n);
  Com::CVector3D dhdr = Com::CMatrix3(n).MatVec(v);  
  Com::CMatrix3  m3 = -cont_stiff*h*Com::CMatrix3(n).MatMat( Com::CMatrix3(v) );
  Com::CMatrix3 m4; m4.SetOuterProduct(dhdr,dhdr); m4 *= cont_stiff;  
  dce_dudu = Com::CMatrix3(n,n); dce_dudu*=cont_stiff;
  dce_drdr = m3+m4;  
  dce_dudr = Com::CMatrix3(n,dhdr); dce_dudr *= -cont_stiff;
}


// Contact for simplest Rigid body model
void TestContact_RigidPointSet()
{
  const Com::CVector3D c( rand2(-1,1), rand2(-1,1), rand2(-1,1) );    
  const Com::CVector3D v( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  Com::CVector3D b( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  Com::CVector3D n( rand2(-1,1), rand2(-1,1), rand2(-1,1) );  
  const double k = rand2(0.01,100);  
  const double ce = GetContactEnergy(c,v,b,n,k);
  Com::CVector3D dce_du, dce_dr;
  Get1stDer_ContactEnergy(dce_du,dce_dr, c,v,b,n,k);
  Com::CMatrix3 dce_dudu, dce_dudr, dce_drdr;
  Get2ndDer_ContactEnergy(dce_dudu, dce_dudr, dce_drdr, c,v,b,n,k);    
  
  double eps = 1.0e-4;  
  Com::CVector3D e( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  e *= eps;
  
  {
    const Com::CVector3D& c1 = c + e;
    double ce1 = GetContactEnergy(c1,v,b,n,k);
    std::cout << "dce_du" << std::endl;
    std::cout << (ce1-ce)/eps << " " << Dot(dce_du,e)/eps << std::endl; 
    Com::CVector3D dce_du1, dce_dr1;    
    Get1stDer_ContactEnergy(dce_du1,dce_dr1, c1,v,b,n,k);    
    Com::CVector3D d0 = (dce_du1-dce_du)*=1.0/eps;
    Com::CVector3D d1 = dce_dudu.MatVec(e); d1*=1.0/eps;
    std::cout << "dce_dudu" << std::endl;
    std::cout << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    std::cout << "dce_drdu" << std::endl;
    Com::CVector3D e0 = (dce_dr1-dce_dr)*=1.0/eps;
    Com::CVector3D e1 = dce_dudr.MatVecTrans(e); e1*=1.0/eps;
    std::cout << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << e1.x << " " << e1.y << " " << e1.z << std::endl;        
  }
  {
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e);
    const Com::CVector3D& v1 = rot.MatVec(v);
    double ce1 = GetContactEnergy(c,v1,b,n,k);    
    std::cout << "dce_dr" << std::endl;        
    std::cout << (ce1-ce)/eps << " " << Dot(dce_dr,e)/eps << std::endl; 
    Com::CVector3D dce_du1, dce_dr1;    
    Get1stDer_ContactEnergy(dce_du1,dce_dr1, c,v1,b,n,k);    
    Com::CVector3D d0 = (dce_du1-dce_du)*=1.0/eps;
    Com::CVector3D d1 = dce_dudr.MatVec(e); d1*=1.0/eps;    
    std::cout << "dce_dudr" << std::endl;    
    std::cout << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    std::cout << "dce_drdr" << std::endl;        
    Com::CVector3D e0 = (dce_dr1-dce_dr)*=1.0/eps;
    Com::CVector3D e1 = dce_drdr.MatVec(e); e1*=1.0/eps;
    std::cout << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << e1.x << " " << e1.y << " " << e1.z << std::endl;            
  }  
}

////////////////////////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////

double GetStaticFrictionEnergy
(const Com::CVector3D& c,
 const Com::CVector3D& v,
 const Com::CVector3D& a, 
 const Com::CVector3D& n, 
 const double stiff_f)
{
  Com::CVector3D ap_t = c+v-a;  // relative position from anchor
  ap_t -= n*Com::Dot(n,ap_t);  // tangent vector from anchor to point
  return 0.5*stiff_f*Dot(ap_t,ap_t);
}

void Get1stDer_StaticFrictionEnergy
(Com::CVector3D& fv, Com::CVector3D& fr,
 const Com::CVector3D& c,
 const Com::CVector3D& v,
 const Com::CVector3D& a,
 const Com::CVector3D& n, 
 const double stiff_f)
{
  Com::CVector3D ap_t = c+v-a;  // relative position from anchor
  ap_t -= n*Com::Dot(n,ap_t);  // tangent vector from anchor to point
  fv = stiff_f*ap_t;// calc a spring force between anchor
  fr = Com::CMatrix3(v).MatVec(fv);
}

void Get2ndDer_StaticFrictionEnergy
(Com::CMatrix3& dfv_du, Com::CMatrix3& dfv_dr, Com::CMatrix3& dfr_du, Com::CMatrix3& dfr_dr,
 const Com::CVector3D& c, 
 const Com::CVector3D& v,
 const Com::CVector3D& a,
 const Com::CVector3D& n, 
 double stiff_f)
{
  //  const Com::CVector3D Qa = v;   
  const Com::CMatrix3 wv = Com::CMatrix3(v);
  Com::CVector3D ap_t = c+v-a;  // relative position from anchor  
  ap_t -= n*Com::Dot(n,ap_t);  // tangent vector from anchor to pointw
  Com::CMatrix3 m3 = stiff_f*Com::CMatrix3(ap_t).MatMat( Com::CMatrix3(v) );
  
  dfv_du = stiff_f*(Com::CMatrix3(1)-Com::CMatrix3(n,n));
  //  dfv_du = -stiff_f*Com::CMatrix3(1);  
  dfv_dr = -dfv_du.MatMat(wv);
  dfr_du = wv.MatMat(dfv_du);
  dfr_dr = wv.MatMat(dfv_dr) + m3;
}


void TestStaticFriction_RigidPointSet()
{
  const Com::CVector3D c( rand2(-1,1), rand2(-1,1), rand2(-1,1) );    
  const Com::CVector3D v( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  const Com::CVector3D a( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  Com::CVector3D n( rand2(-1,1), rand2(-1,1), rand2(-1,1) );   n.SetNormalizedVector();
  const double k = rand2(0.01,100);  
  const double fe = GetStaticFrictionEnergy(c,v,a,n,k);
  Com::CVector3D dfe_du, dfe_dr;
  Get1stDer_StaticFrictionEnergy(dfe_du,dfe_dr, c,v,a,n,k);
  Com::CMatrix3 dfe_dudu, dfe_dudr, dfe_drdu, dfe_drdr;
  Get2ndDer_StaticFrictionEnergy(dfe_dudu,dfe_dudr,dfe_drdu,dfe_drdr, c,v,a,n,k);    
  
  const double eps = 1.0e-5;  
  Com::CVector3D e( rand2(-1,1), rand2(-1,1), rand2(-1,1) );
  e *= eps;
  
  {
    const Com::CVector3D c1 = c + e;
    const double fe1 = GetStaticFrictionEnergy(c1,v,a,n,k);
    std::cout << "dfe_du" << std::endl;
    std::cout << (fe1-fe)/eps << " " << Dot(dfe_du,e)/eps << std::endl; 
    Com::CVector3D dfe_du1, dfe_dr1;    
    Get1stDer_StaticFrictionEnergy(dfe_du1,dfe_dr1, c1,v,a,n,k);    
    Com::CVector3D d0 = (dfe_du1-dfe_du)*=1.0/eps;
    Com::CVector3D d1 = dfe_dudu.MatVec(e); d1*=1.0/eps;
    std::cout << "dfe_dudu" << std::endl;
    std::cout << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    std::cout << "dfe_drdu" << std::endl;
    Com::CVector3D e0 = (dfe_dr1-dfe_dr)*=1.0/eps;
    Com::CVector3D e1 = dfe_dudr.MatVecTrans(e); e1*=1.0/eps;
    std::cout << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << e1.x << " " << e1.y << " " << e1.z << std::endl;        
  }
  {
    Com::CMatrix3 rot; rot.SetRotMatrix_Cartesian(e);
    const Com::CVector3D v1 = rot.MatVec(v);
    const double fe1 = GetStaticFrictionEnergy(c,v1,a,n,k);    
    std::cout << "dfe_dr" << std::endl;        
    std::cout << (fe1-fe)/eps << " " << Dot(dfe_dr,e)/eps << std::endl; 
    Com::CVector3D dfe_du1, dfe_dr1;    
    Get1stDer_StaticFrictionEnergy(dfe_du1,dfe_dr1, c,v1,a,n,k);    
    Com::CVector3D d0 = (dfe_du1-dfe_du)*=1.0/eps;
    Com::CVector3D d1 = dfe_dudr.MatVec(e); d1*=1.0/eps;    
    std::cout << "dfe_dudr" << std::endl;    
    std::cout << d0.x << " " << d0.y << " " << d0.z << std::endl;
    std::cout << d1.x << " " << d1.y << " " << d1.z << std::endl;    
    std::cout << "dfe_drdr" << std::endl;        
    Com::CVector3D e0 = (dfe_dr1-dfe_dr)*=1.0/eps;
    Com::CVector3D e1 = dfe_drdr.MatVec(e); e1*=1.0/eps;
    std::cout << e0.x << " " << e0.y << " " << e0.z << std::endl;
    std::cout << e1.x << " " << e1.y << " " << e1.z << std::endl;            
  }  
}



/////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////

double GetSpringEnergy
(const Com::CVector3D& v, // dist from center
 const Com::CVector3D& a) // anchor
{
  Com::CVector3D ap_t = v-a;  // relative position from anchor
  return 0.5*Dot(ap_t,ap_t);
}

void Get1stDer_SpringEnergy
(Com::CVector3D& fv, Com::CVector3D& fr,
 const Com::CVector3D& v,
 const Com::CVector3D& a)
{
  Com::CVector3D ap_t = v-a;  // relative position from anchor
  fv = ap_t;// calc a spring force between anchor
  fr = Com::CMatrix3(v).MatVec(fv);
}

void Get2ndDer_SpringEnergy
(Com::CMatrix3& dfv_du, Com::CMatrix3& dfv_dr, Com::CMatrix3& dfr_du, Com::CMatrix3& dfr_dr,
 const Com::CVector3D& v,
 const Com::CVector3D& a)
{
  const Com::CMatrix3 wv = Com::CMatrix3(v);
  Com::CVector3D ap_t = v-a;  // relative position from anchor  
  Com::CMatrix3 m3 = Com::CMatrix3(ap_t).MatMat( Com::CMatrix3(v) );  
  
  dfv_du = Com::CMatrix3(1);
  dfv_dr = -wv;
  dfr_du = wv;
  dfr_dr = -wv.MatMat(wv) + m3;
}


